#include "insist.h"
#include "mt_Component.h"
#include "mu_ChangeWidget.h"
#include "mu_MainWidget.h"
extern mu_MainWidget*mu_mainWidget;



static void mu_ChangeWidget_xbar16_apply (GtkWidget*w, mu_ChangeWidget*widget)
{
  widget->createThread ((mu_ThreadableFunction)&mu_ChangeWidget::xbar16_run, w);
}
static void mu_ChangeWidget_xbarPort_apply (GtkWidget*w, mu_ChangeWidget*widget)
{
  widget->createThread ((mu_ThreadableFunction)&mu_ChangeWidget::xbarPort_run, w);
}
static void mu_ChangeWidget_hssdcPort_apply (GtkWidget*w, mu_ChangeWidget*widget)
{
  widget->createThread ((mu_ThreadableFunction)&mu_ChangeWidget::hssdcPort_run, w);
}
static void mu_ChangeWidget_shortwaveFiberPort_apply (GtkWidget*w, mu_ChangeWidget*widget)
{
  widget->createThread ((mu_ThreadableFunction)&mu_ChangeWidget::shortwaveFiberPort_run, w);
}


mu_ChangeType mu_ChangeWidget::types [MU_CHANGE_NUM_TYPES] = 
{
  {
    "Timeout", M3_XBAR16_TABLE, "xbar16", GTK_SIGNAL_FUNC (mu_ChangeWidget_xbar16_apply)
  },
  {
    "Control", M3_XBAR_PORT_TABLE, "xbarPort", GTK_SIGNAL_FUNC (mu_ChangeWidget_xbarPort_apply)
  },
  {
    "Control", M3_HSSDC_PORT_TABLE, "hssdcPort", GTK_SIGNAL_FUNC (mu_ChangeWidget_hssdcPort_apply)
  },
  {
    "Control", M3_SHORTWAVE_FIBER_PORT_TABLE, "shortwaveFiberPort", GTK_SIGNAL_FUNC (mu_ChangeWidget_shortwaveFiberPort_apply)
  }
};

static gboolean mu_ChangeWidget_delete (GtkWidget*w, GdkEvent*e, mu_ChangeWidget*widget)
{
  widget->hide ();
  return 1;
}

static gboolean mu_ChangeWidget_destroy (GtkWidget*w, mu_ChangeWidget*widget)
{
  widget->hide ();
  return 0;
}

static void mu_ChangeWidget_closeButton_clicked (GtkWidget*w, mu_ChangeWidget*widget)
{
  widget->hide ();
}

mu_ChangeWidget::mu_ChangeWidget ()
{
  widget = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  insist (widget);
  GtkWidget*windowBox;
  
  gtk_window_set_title (GTK_WINDOW (widget), MU_TITLE " - Control");
  gtk_window_set_policy (GTK_WINDOW (widget), 1, 1, 0);
  gtk_widget_set_usize (widget, 400, 250);

  gtk_signal_connect (GTK_OBJECT (widget), "delete_event", GTK_SIGNAL_FUNC (mu_ChangeWidget_delete), this);
  gtk_signal_connect (GTK_OBJECT (widget), "destroy", GTK_SIGNAL_FUNC (mu_ChangeWidget_destroy), this);

  windowBox = gtk_vbox_new (0, MU_SPACING);
  insist (windowBox);
  gtk_widget_show (windowBox);
  
  gtk_container_add (GTK_CONTAINER (widget), windowBox);

  GtkWidget*tops [MU_CHANGE_NUM_TYPES];
  
  for (int i = 0; i < MU_CHANGE_NUM_TYPES; i++)
  {
    GtkWidget*box,*bottom, *left, *right, *applyBox, *closeBox;
    GtkWidget*label, *button, *frame;

    char s [M3_DISPLAY_STRING_LENGTH];
    
    boxes [i] = gtk_vbox_new (0, 0);
    insist (boxes [i]);
    gtk_box_pack_start (GTK_BOX (windowBox), boxes [i], 1, 1, 0);

    frame = gtk_frame_new (0);
    insist (frame);
    gtk_widget_show (frame);
    gtk_box_pack_start (GTK_BOX (boxes [i]), frame, 1, 1, 0);
	
    tops [i] = gtk_vbox_new (0, MU_SPACING);
    insist (tops [i]);
    gtk_widget_show (tops [i]);
    gtk_container_add (GTK_CONTAINER (frame), tops [i]);

    bottom = gtk_hbox_new (0, MU_SPACING);
    insist (bottom);
    gtk_widget_show (bottom);
    gtk_box_pack_start (GTK_BOX (boxes [i]), bottom, 0, 0, 0);

    left = gtk_vbox_new (0, MU_SPACING);
    insist (left);
    gtk_widget_show (left);
    gtk_box_pack_start (GTK_BOX (bottom), left, 0, 0, 0);

    right = gtk_vbox_new (0, MU_SPACING);
    insist (right);
    gtk_widget_show (right);
    gtk_box_pack_start (GTK_BOX (bottom), right, 0, 0, 0);

    applyBox = gtk_hbox_new (0, MU_SPACING);
    insist (applyBox);
    gtk_widget_show (applyBox);
    gtk_box_pack_start (GTK_BOX (left), applyBox, 0, 0, 0);

    closeBox = gtk_vbox_new (0, MU_SPACING);
    insist (closeBox);
    gtk_widget_show (closeBox);
    gtk_box_pack_start (GTK_BOX (left), closeBox, 0, 0, 0);

    button = gtk_button_new_with_label ("Apply");
    insist (button);
    gtk_widget_show (button);
    gtk_box_pack_start (GTK_BOX (applyBox), button, 0, 0, 0);
    gtk_signal_connect (GTK_OBJECT (button), "clicked", types [i].apply, this);

    label = gtk_label_new (" to: ");
    insist (label);
    gtk_widget_show (label);
    gtk_box_pack_start (GTK_BOX (applyBox), label, 0, 0, 0);

    label = gtk_label_new ("or");
    insist (label);
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
    gtk_widget_show (label);
    gtk_box_pack_start (GTK_BOX (closeBox), label, 1, 1, 0);

    button = gtk_button_new_with_label ("Close");
    insist (button);
    gtk_widget_show (button);
    gtk_box_pack_start (GTK_BOX (closeBox), button, 0, 0, 0);
    gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_ChangeWidget_closeButton_clicked), this);
  
    box = gtk_hbox_new (0, 0);
    insist (boxes [i]);
    gtk_widget_show (box);
    gtk_box_pack_start (GTK_BOX (right), box, 0, 0, 0);

    partRadios [i] = gtk_radio_button_new (0);
    insist (partRadios [i]);
    gtk_widget_show (partRadios [i]);
    gtk_box_pack_start (GTK_BOX (box), partRadios [i], 0, 0, 0);
  
    label = gtk_label_new (sprintf (s, "This %s", types [i].name) ? s : "");
    insist (label);
    gtk_widget_show (label);
    gtk_box_pack_start (GTK_BOX (box), label, 0, 0, 0);

    box = gtk_hbox_new (0, 0);
    insist (box);
    gtk_widget_show (box);
    gtk_box_pack_start (GTK_BOX (right), box, 0, 0, 0);

    slotRadios [i] = gtk_radio_button_new_from_widget (GTK_RADIO_BUTTON (partRadios [i]));
    insist (slotRadios [i]);
    gtk_widget_show (slotRadios [i]);
    gtk_box_pack_start (GTK_BOX (box), slotRadios [i], 0, 0, 0);  

    label = gtk_label_new (sprintf (s, "This slot's %ss", types [i].name) ? s : "");
    insist (label);
    gtk_widget_show (label);
    gtk_box_pack_start (GTK_BOX (box), label, 0, 0, 0);

    box = gtk_hbox_new (0, 0);
    insist (box);
    gtk_widget_show (box);
    gtk_box_pack_start (GTK_BOX (right), box, 0, 0, 0);

    switchRadios [i] = gtk_radio_button_new_from_widget (GTK_RADIO_BUTTON (partRadios [i]));
    insist (switchRadios [i]);
    gtk_widget_show (switchRadios [i]);
    gtk_box_pack_start (GTK_BOX (box), switchRadios [i], 0, 0, 0);  

    label = gtk_label_new (sprintf (s, "This switch's %ss", types [i].name) ? s : "");
    insist (label);
    gtk_widget_show (label);
    gtk_box_pack_start (GTK_BOX (box), label, 0, 0, 0);
  
    box = gtk_hbox_new (0, 0);
    insist (box);
    gtk_widget_show (box);
    gtk_box_pack_start (GTK_BOX (right), box, 0, 0, 0);

    clusterRadios [i] = gtk_radio_button_new_from_widget (GTK_RADIO_BUTTON (switchRadios [i]));
    insist (clusterRadios [i]);
    gtk_widget_show (clusterRadios [i]);
    gtk_box_pack_start (GTK_BOX (box), clusterRadios [i], 0, 0, 0);

    label = gtk_label_new (sprintf (s, "This cluster's %ss", types [i].name) ? s : "");
    insist (label);
    gtk_widget_show (label);
    gtk_box_pack_start (GTK_BOX (box), label, 0, 0, 0);
  }


  int i;
  i = 0;

  labels [i] = gtk_label_new ("Timeout for Xbar");
  insist (labels [i]);
  gtk_widget_show (labels [i]);
  insist (tops [i]);
  gtk_box_pack_start (GTK_BOX (tops [i]), labels [i], 0, 0, 0);

  xbar16_sixteenth = gtk_radio_button_new_with_label (0, "sixteenth second");
  insist (xbar16_sixteenth);
  gtk_widget_show (xbar16_sixteenth);
  gtk_box_pack_start (GTK_BOX (tops [i]), xbar16_sixteenth, 0, 0, 0);

  xbar16_quarter = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (xbar16_sixteenth), "quarter second");
  insist (xbar16_quarter);
  gtk_widget_show (xbar16_quarter);
  gtk_box_pack_start (GTK_BOX (tops [i]), xbar16_quarter, 0, 0, 0);

  xbar16_one = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (xbar16_sixteenth), "one second");
  insist (xbar16_one);
  gtk_widget_show (xbar16_one);
  gtk_box_pack_start (GTK_BOX (tops [i]), xbar16_one, 0, 0, 0);

  xbar16_four = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (xbar16_sixteenth), "four seconds");
  insist (xbar16_four);
  gtk_widget_show (xbar16_four);
  gtk_box_pack_start (GTK_BOX (tops [i]), xbar16_four, 0, 0, 0);


  i++;
  labels [i] = gtk_label_new ("");
  insist (labels [i]);
  gtk_widget_show (labels [i]);
  insist (tops [i]);
  gtk_box_pack_start (GTK_BOX (tops [i]), labels [i], 0, 0, 0);

  xbarPort_slow =  gtk_check_button_new_with_label ("Slow");
  insist (xbarPort_slow);
  gtk_widget_show (xbarPort_slow);
  gtk_box_pack_start (GTK_BOX (tops [i]), xbarPort_slow, 0, 0, 0);

  xbarPort_off =  gtk_check_button_new_with_label ("Off");
  insist (xbarPort_off);
  gtk_widget_show (xbarPort_off);
  gtk_box_pack_start (GTK_BOX (tops [i]), xbarPort_off, 0, 0, 0);

  xbarPort_ha =  gtk_check_button_new_with_label ("HA Off");
  insist (xbarPort_ha);
  gtk_widget_show (xbarPort_ha);
  gtk_box_pack_start (GTK_BOX (tops [i]), xbarPort_ha, 0, 0, 0);

  i++;
  labels [i] = gtk_label_new ("");
  insist (labels [i]);
  gtk_widget_show (labels [i]);
  insist (tops [i]);
  gtk_box_pack_start (GTK_BOX (tops [i]), labels [i], 0, 0, 0);

  hssdc_internal =  gtk_check_button_new_with_label ("Internal Loopback");
  insist (hssdc_internal);
  gtk_widget_show (hssdc_internal);
  gtk_box_pack_start (GTK_BOX (tops [i]), hssdc_internal, 0, 0, 0);

  hssdc_external =  gtk_check_button_new_with_label ("External Loopback");
  insist (hssdc_external);
  gtk_widget_show (hssdc_external);
  gtk_box_pack_start (GTK_BOX (tops [i]), hssdc_external, 0, 0, 0);

  hssdc_reset =  gtk_check_button_new_with_label ("Reset Serdes/San");
  insist (hssdc_reset);
  gtk_widget_show (hssdc_reset);
  gtk_box_pack_start (GTK_BOX (tops [i]), hssdc_reset, 0, 0, 0);

  i++;
  labels [i] = gtk_label_new ("");
  insist (labels [i]);
  gtk_widget_show (labels [i]);
  insist (tops [i]);
  gtk_box_pack_start (GTK_BOX (tops [i]), labels [i], 0, 0, 0);

  shortwave_internal =  gtk_check_button_new_with_label ("Internal Loopback");
  insist (shortwave_internal);
  gtk_widget_show (shortwave_internal);
  gtk_box_pack_start (GTK_BOX (tops [i]), shortwave_internal, 0, 0, 0);

  shortwave_external =  gtk_check_button_new_with_label ("External Loopback");
  insist (shortwave_external);
  gtk_widget_show (shortwave_external);
  gtk_box_pack_start (GTK_BOX (tops [i]), shortwave_external, 0, 0, 0);

  shortwave_reset =  gtk_check_button_new_with_label ("Reset Serdes/San");
  insist (shortwave_reset);
  gtk_widget_show (shortwave_reset);
  gtk_box_pack_start (GTK_BOX (tops [i]), shortwave_reset, 0, 0, 0);

  shortwave_laser =  gtk_check_button_new_with_label ("Laser Off");
  insist (shortwave_laser);
  gtk_widget_show (shortwave_laser);
  gtk_box_pack_start (GTK_BOX (tops [i]), shortwave_laser, 0, 0, 0);

  exception:;
}

int mu_ChangeWidget::lookup (int type)
{
  insist (this);

  for (int i = 0; i < MU_CHANGE_NUM_TYPES; i++)
    if (type == types[i].type)
      return i;
  
  exception: return -1;
}

int mu_ChangeWidget::set (int address, int type, int index)
{
  insist (this);
  char s [M3_DISPLAY_STRING_LENGTH];
  
  type = lookup (type);
  insist (type >= 0 && type < MU_CHANGE_NUM_TYPES);

  this->index = index;
  this->type = type;
  this->address = address;
  
  sprintf (s, "%d.%d.%d.%d\n%s For %s %d", 
	   (unsigned) (((char*)(&address)) [0] & 0xff),
	   (unsigned) (((char*)(&address)) [1] & 0xff),
	   (unsigned) (((char*)(&address)) [2] & 0xff),
	   (unsigned) (((char*)(&address)) [3] & 0xff),
	   types [type].title,
	   types [type].name,
	   index);

  insist (labels [type]);
  gtk_label_set_text (GTK_LABEL (labels [type]), s);
  
  for (int i = 0; i < MU_CHANGE_NUM_TYPES; i++)
    gtk_widget_hide (boxes [i]);
  
  gtk_widget_show (boxes [type]);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (partRadios [type]), 1);
  
  show ();

  exception: return 0;
}


char*mu_ChangeWidget::getChange (char*s, int type, int*cluster)
{
  insist (this);
  insist (s && cluster);
  insist (type >= 0 && type < MU_CHANGE_NUM_TYPES);
  insist (partRadios [type] && slotRadios [type] && switchRadios [type] && clusterRadios [type]);

  *cluster = 0;
  
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (partRadios [type])))
    strcpy (s, "this");
  else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (slotRadios [type])))
    strcpy (s, "slot");
  else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (switchRadios [type])))
    strcpy (s, "all");
  else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (clusterRadios [type])))
  {
    *cluster = 1;
    strcpy (s, "all");
  }
  else 
  {
    insist (0);
  }
  return s;
  exception: return 0;  
}

void mu_ChangeWidget::xbar16_run (void*data)
{
  GtkWidget*button = GTK_WIDGET (data);

  insist (this);
  insist (address);
  insist (index > 0);
  insist (type == 0);
  insist (button);
  insist (mu_mainWidget && mu_mainWidget->cluster);
  
  char url [M3_DISPLAY_STRING_LENGTH];
  char v [100];
  char v2 [100];
  
  char*s;
  
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (xbar16_sixteenth)))
    strcpy (v, "sixteenth+second");
  else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (xbar16_quarter)))
    strcpy (v, "quarter+second");
  else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (xbar16_one)))
    strcpy (v, "one+second");
  else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (xbar16_four)))
    strcpy (v, "four+seconds");
  else
  {
    insist (0);
  }
  
  sprintf (url, "/change_xbar16%d", index);
  

  setWatchCursor (1);
  gdk_threads_enter ();
  gtk_widget_set_sensitive (button, 0);
  gdk_threads_leave ();

  int all;
  getChange (v2, type, &all);  

  for (int i = 0; i < mu_mainWidget->cluster->numSwitches; i++)
  {
    if (all || mu_mainWidget->cluster->switches [i]->ipAddress == address)
    {
      s = web.post (mu_mainWidget->cluster->switches [i]->ipAddress, url, 2, "timeout", v, "change", v2);
    }
  }

  setWatchCursor (0);
  gdk_threads_enter ();
  gtk_widget_set_sensitive (button, 1);
  gdk_threads_leave ();
  
  exception:;
} 


void mu_ChangeWidget::xbarPort_run (void*)
{
  GtkWidget*button = GTK_WIDGET (data);

  insist (this);
  insist (address);
  insist (index > 0);
  insist (type == 1);
  insist (button);
  insist (mu_mainWidget && mu_mainWidget->cluster);

  char url [M3_DISPLAY_STRING_LENGTH];
  char v [M3_DISPLAY_STRING_LENGTH];
  char*s;
  
  sprintf (url, "/change_xbarPort%d", index);
 

  setWatchCursor (1);
  gdk_threads_enter ();
  gtk_widget_set_sensitive (button, 0);
  gdk_threads_leave ();

  int all;
  getChange (v, type, &all);  

  for (int i = 0; i < mu_mainWidget->cluster->numSwitches; i++)
  {
    if (all || mu_mainWidget->cluster->switches [i]->ipAddress == address)
    {
      s = web.post (mu_mainWidget->cluster->switches [i]->ipAddress,
		    url, 4, 
		    "slow", gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (xbarPort_slow)) ? "on" : "off",
		    "port_off", gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (xbarPort_off)) ? "on" : "off",
		    "ha_off", gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (xbarPort_ha)) ? "on" : "off",
		    "change", v);
    }
  }
  
  setWatchCursor (0);
  gdk_threads_enter ();
  gtk_widget_set_sensitive (button, 1);
  gdk_threads_leave ();
  
  exception:;
}

void mu_ChangeWidget::hssdcPort_run (void*)
{
  GtkWidget*button = GTK_WIDGET (data);

  insist (this);
  insist (address);
  insist (index > 0);
  insist (type == 2);
  insist (button);
  insist (mu_mainWidget && mu_mainWidget->cluster);
  
  char url [M3_DISPLAY_STRING_LENGTH];
  char v [M3_DISPLAY_STRING_LENGTH];
  char*s;
  
  sprintf (url, "/change_hssdcPort%d", index);


  setWatchCursor (1);
  gdk_threads_enter ();
  gtk_widget_set_sensitive (button, 0);
  gdk_threads_leave ();

  int all;
  getChange (v, type, &all);  

  for (int i = 0; i < mu_mainWidget->cluster->numSwitches; i++)
  {
    if (all || mu_mainWidget->cluster->switches [i]->ipAddress == address)
    {
      s = web.post (mu_mainWidget->cluster->switches [i]->ipAddress, 
		    url, 4, 
		    "internal_loopback", gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (hssdc_internal)) ? "on" : "off",
		    "external_loopback", gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (hssdc_external)) ? "on" : "off",
		    "reset_serdes_san", gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (hssdc_reset)) ? "on" : "off",
		    "change", v);
    }
  }
  

  setWatchCursor (0);
  gdk_threads_enter ();
  gtk_widget_set_sensitive (button, 1);
  gdk_threads_leave ();
  
  exception:;
}

void mu_ChangeWidget::shortwaveFiberPort_run (void*)
{
  GtkWidget*button = GTK_WIDGET (data);

  insist (this);
  insist (address);
  insist (index > 0);
  insist (type == 3);
  insist (button);
  insist (mu_mainWidget && mu_mainWidget->cluster);
  
  char url [M3_DISPLAY_STRING_LENGTH];
  char v [M3_DISPLAY_STRING_LENGTH];
  char*s;
  
  sprintf (url, "/change_shortwaveFiberPort%d", index);
  
  setWatchCursor (1);
  gdk_threads_enter ();
  gtk_widget_set_sensitive (button, 0);
  gdk_threads_leave ();

  int all;
  getChange (v, type, &all);  

  for (int i = 0; i < mu_mainWidget->cluster->numSwitches; i++)
  {
    if (all || mu_mainWidget->cluster->switches [i]->ipAddress == address)
    {
      s = web.post (mu_mainWidget->cluster->switches [i]->ipAddress, 
		    url, 5, 
		    "internal_loopback", gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (shortwave_internal)) ? "on" : "off",
		    "external_loopback", gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (shortwave_external)) ? "on" : "off",
		    "reset_serdes_san", gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (shortwave_reset)) ? "on" : "off",
		    "laser_off", gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (shortwave_laser)) ? "on" : "off",
		    "change", v);
    }
  }

  setWatchCursor (0);
  gdk_threads_enter ();
  gtk_widget_set_sensitive (button, 1);
  gdk_threads_leave ();

  exception:;
}
