#include <ctype.h>
#include "insist.h"
#include "mu_Cluster.h"
#include "mu_MainWidget.h"
#include "mt_Component.h"
#include "m3_TestClient.h"

#include "point.xbm"
#include "point.xpm"
#include "zoom.xbm"
#include "zoom.xpm"
#include "hand.xpm"
#include "hand.xbm"
#include "question.xpm"
#include "question.xbm"
#include "change.xpm"
#include "change.xbm"

#include "myricom.xpm"

#define MU_CLUSTER_FONT "-misc-*-medium-*-*-*-*-*-*-*-*-*-*-*"

#define MU_CLUSTER_LINECARD_PORT_COLOR "darkgrey"

#define MU_CLUSTER_PORT_UNITS 4
#define MU_CLUSTER_MAX_LINK_WIDTH 4

#define MU_CLUSTER_SELECT_SIZE 6
#define MU_CLUSTER_SELECT_OFFSET 4

#define MU_CLUSTER_PORT_CONNECTOR_WIDTH (MU_CLUSTER_PORT_UNITS * 5)
#define MU_CLUSTER_PORT_CONNECTOR_HEIGHT (MU_CLUSTER_PORT_UNITS * 3)

#define MU_CLUSTER_PORT_TOTAL_WIDTH (MU_CLUSTER_PORT_UNITS * 15)
#define MU_CLUSTER_PORT_TOTAL_HEIGHT (MU_CLUSTER_PORT_UNITS * 8)

#define MU_XBAR_SIDE (MU_CLUSTER_PORT_UNITS * 7)

#define MU_CLUSTER_MARGIN 2
#define MU_CLUSTER_NAME_HEIGHT (MU_CLUSTER_PORT_TOTAL_HEIGHT * 2)
#define MU_CLUSTER_NAME_WIDTH (MU_CLUSTER_PORT_TOTAL_WIDTH * 6)
#define MU_CLUSTER_SIZE 5000

#define MU_CLUSTER_HOST_HEIGHT (MU_CLUSTER_PORT_TOTAL_HEIGHT / 2)
#define MU_CLUSTER_HOST_WIDTH MU_CLUSTER_PORT_TOTAL_WIDTH

#define MU_CLUSTER_ZOOM_WINDOW_WIDTH 100
#define MU_CLUSTER_ZOOM_WINDOW_HEIGHT 20

#define MU_CLUSTER_OBJECT_COLOR "lightgray"

int mu_Cluster_colors [MU_CLUSTER_NUM_COLORS] = 
{
  0x00008800,
  0x0000cc00,
  0x0000ff00,

  0x00660000,
  0x00990000,
  0x00ff0000,

  0xccff0000,
  0xffff6600,

  0xffcc0000,
  0xff993300,
  
  0xcc663300,
  0xff000000,
  0xcc000000,
};

extern mu_MainWidget*mu_mainWidget;
extern int mu_maxThreads;

static gint mu_Cluster_xbar (GnomeCanvasItem*item, GdkEvent *event, mu_Xbar*xbar)
{
  insist (xbar);
  return xbar->callback (item, event);
  exception: return 0;
}

static gint mu_Cluster_port (GnomeCanvasItem*item, GdkEvent *event, mu_Xbar*xbar)
{
  insist (xbar);
  return xbar->portCallback (item, event);
  exception: return 0;
}

static gint mu_Cluster_linecardPort (GnomeCanvasItem*item, GdkEvent *event, mu_LinecardPort*port)
{
  insist (port);
  return port->callback (item, event);
  exception: return 0;
}

static gint mu_Cluster_canvas (GnomeCanvasItem *item, GdkEvent *event, mu_Cluster*cluster)
{
  insist (cluster);
  return cluster->callback (item, event);  
  exception: return 0;
}

static gint mu_Cluster_toggled (GtkWidget*w, mu_Cluster*cluster)
{
  insist (cluster);
  return cluster->toggled ();
  exception: return 0;
}

int mu_Cluster::callback (GnomeCanvasItem*item, GdkEvent*event)
{
  static double x, y;
  double bx, by;
  double ax, ay;
  double xx, yy;
  static double cx, cy;
  static double hv, vv;
  static int dragging;
  static int selecting;
  static int moving;
  
  insist (this && item && event && selection);
  insist (zoomToggle);
  
  if (building)
    return 1;
  
  ax = (double) GTK_WIDGET (canvas)->allocation.width / canvas->pixels_per_unit - 3 * MU_CLUSTER_SIZE;
  ay = (double) GTK_WIDGET (canvas)->allocation.height / canvas->pixels_per_unit - 3 * MU_CLUSTER_SIZE;
  

  xx = yy = 0;
  
  if (event->type == GDK_BUTTON_PRESS || event->type == GDK_BUTTON_RELEASE)
  {
    xx = event->button.x;
    yy = event->button.y;
  }
  else if (event->type == GDK_MOTION_NOTIFY)
  {
    xx = event->motion.x;
    yy = event->motion.y;
  }
  else return 0;
  
  bx = (xx + hadjustment->value) / canvas->pixels_per_unit - MU_CLUSTER_SIZE;
  by = (yy + vadjustment->value) / canvas->pixels_per_unit - MU_CLUSTER_SIZE;
  
  if (ax > 0) bx -= ax / 2;
  if (ay > 0) by -= ay / 2;
  
  switch (event->type)
  {
    case GDK_BUTTON_RELEASE:      
      
      gnome_canvas_item_hide (selection);
      dragging = 0;
      moving = 0;
      
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (zoomToggle)))
      {
	zoom (event->button.x, event->button.y, event->button.button == 1 ? 1.5 : 0.75);
	return 1;
      }
      break;
    case GDK_BUTTON_PRESS:
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (pointToggle)) &&
	  event->button.button == 1 && !inside (bx, by))
      {
	selectAll (0);
	x = bx;
	y = by;
	dragging = 1;
	return 1;
      }
      else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (handToggle)) && event->button.button == 1)
      {
	cx = event->button.x;
	cy = event->button.y;
	hv = hadjustment->value;
	vv = vadjustment->value;
	moving = 1;
	return 1;
      }
      break;
    case GDK_MOTION_NOTIFY:
      if (dragging && event->motion.state & GDK_BUTTON1_MASK)
      {
	double x1 = x < bx ? x : bx;
	double x2 = x > bx ? x : bx;
	double y1 = y < by ? y : by;
	double y2 = y > by ? y : by;

	selectAll (0);
	selectAll (x1, y1, x2, y2, 1);

	gnome_canvas_item_set (selection,  "x1", x1, "y1", y1, "x2", x2, "y2", y2, NULL);
	gnome_canvas_item_show (selection);
	return 1;
      }
      else if (moving && event->motion.state & GDK_BUTTON1_MASK)
      {
	gtk_adjustment_set_value (hadjustment, hv + (cx - xx));
	gtk_adjustment_set_value (vadjustment, vv + (cy - yy));
	return 1;
      }
      break;
    default:
      break;
  }
  exception: return 0;
}

gint mu_Cluster_movable (GnomeCanvasItem *item, GdkEvent *event, mu_Movable*movable)
{
  insist (item && movable && event);
  return movable->callback (item, event);
  exception: return 0;
}


int mu_Host::callback (GnomeCanvasItem*item, GdkEvent*event)
{
  insist (this && item && event && cluster);
  
  if (event->type == GDK_BUTTON_PRESS && gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (cluster->questionToggle)))
  {
    if (cluster->routeHost == this)
    {
      mu_Link*link;
      mu_Host*h;

      for (link = (mu_Link*) cluster->links; link; link = (mu_Link*) link->next)
      {
	gnome_canvas_item_set (link->item, "fill_color", "black",  "width_pixels", 1, NULL);
	link->routeCount = 0;
      }
      
      for (h = (mu_Host*) cluster->hosts; h; h = (mu_Host*) h->next)
      {
	h->routeSelected = 0;
	gnome_canvas_item_set (h->box, "fill_color", MU_CLUSTER_OBJECT_COLOR, NULL);
      }
      cluster->routeHost = 0;
      return 1;
    }
    else if (cluster->routeHost)
    {
      routeSelected = !routeSelected;
      cluster->followRoute (cluster->routeHost, this, 1, routeSelected ? 1 : -1);
      gnome_canvas_item_set (box, "fill_color", routeSelected ? "green" : MU_CLUSTER_OBJECT_COLOR, NULL);
      return 1;
    }
    else
    {
      cluster->routeHost = this;
      routeSelected = 1;
      gnome_canvas_item_set (box, "fill_color", "red", NULL);
      return 1;
    }   
  }
  else return mu_Movable::callback (item, event);
  exception: return 0;
}

int mu_Movable::callback (GnomeCanvasItem*item, GdkEvent*event)
{
  static double x, y;
  static int dragging;
  double ix, iy;

  insist (this && item && event && cluster);
  
  ix = event->button.x;
  iy = event->button.y;
  
  switch (event->type)
  {
    case GDK_BUTTON_PRESS:
      if (event->button.button == 1 && gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (cluster->pointToggle)))
      { 
	if (!selected && !(event->button.state & GDK_SHIFT_MASK))
	  cluster->selectAll (0);
	else if (selected && (event->button.state & GDK_SHIFT_MASK))
	{
	  select (0, 0);
	  if (cluster->selected == this)
	    cluster->selected = 0;
	  return 1;
	}
	
	cluster->select (this);
	x = ix;
	y = iy;
	gnome_canvas_item_grab (item, GDK_POINTER_MOTION_MASK | GDK_BUTTON_RELEASE_MASK, 0, event->button.time);
	dragging = 1;
	return 1;
      }
      break;
    case GDK_MOTION_NOTIFY:
      if (dragging && event->motion.state & GDK_BUTTON1_MASK)
      {
	for (int i = 0; i < cluster->numSwitches; i++)
	{
	  if (cluster->switches [i]->selected)
	    cluster->switches [i]->move (ix - x, iy - y);
	}
	for (mu_Host*h = (mu_Host*) cluster->hosts; h; h = (mu_Host*) h->next)
	{
	  if (h->selected)
	    h->move (ix - x, iy - y);
	}
	
	x = ix;
	y = iy;
	return 1;
      }
      break;
    case GDK_BUTTON_RELEASE:
      gnome_canvas_item_ungrab (item, event->button.time);
      dragging = FALSE;
      return 1;
    default:
      break;
  }    
  
  exception: return 0;
}

static void free_freeables (mu_Freeable*n)
{
  while (n)
  {
    mu_Freeable*t = n;
    n = n->next;
    delete t;
  }
}

mu_Freeable::~mu_Freeable ()
{
}

mu_Host::~mu_Host ()
{
}

mu_Switch::~mu_Switch ()
{
}

mu_Cluster::~mu_Cluster ()
{
  free_freeables (links);
  free_freeables (hosts);
  pthread_mutex_destroy (&mutex);
}

mu_Movable::~mu_Movable ()
{
  insist (this);
  
  gnome_canvas_item_hide (GNOME_CANVAS_ITEM (topGroup));
  gtk_object_destroy (GTK_OBJECT (topGroup));
  
  exception:;
}

mu_Switch::mu_Switch (mu_Cluster*cluster)
{
  insist (this && cluster);
  this->cluster = cluster;
  memset (&this->linecardPorts, 0, sizeof (this->linecardPorts));
  memset (&this->xbars, 0, sizeof (this->xbars));
  memset (&this->linecardItems, 0, sizeof (this->linecardItems));
  selected = 0;
  exception:;
}

int mu_Movable::inside (double x1, double y1, double x2, double y2)
{
  insist (this && cluster);

  return cluster->inside (GNOME_CANVAS_ITEM (topGroup), x1, y1, x2, y2);
  
  return 1;
  exception: return 0;
}

int mu_Movable::intersects (double x1, double y1, double x2, double y2)
{
  insist (this && cluster);

  return cluster->intersects (GNOME_CANVAS_ITEM (topGroup), x1, y1, x2, y2);
  
  return 1;
  exception: return 0;
}

int mu_Movable::inside (double x, double y)
{
  insist (this && cluster);

  return cluster->inside (GNOME_CANVAS_ITEM (topGroup), x, y);
  
  return 1;
  exception: return 0;
}

int mu_Movable::select (int selected, int raise)
{
  insist (this);
  
  this->selected = selected;
  
  if (selected)
  {
    if (raise)
    {
      gnome_canvas_item_raise_to_top (GNOME_CANVAS_ITEM (topGroup));
      gnome_canvas_item_raise_to_top (GNOME_CANVAS_ITEM (cluster->linkGroup));
    }
    
    gnome_canvas_item_show (GNOME_CANVAS_ITEM (selectGroup));
  }
  else
  {
    gnome_canvas_item_hide (GNOME_CANVAS_ITEM (selectGroup));
  }
  
  exception: return 0;
}

int mu_Switch::build (double x, double y)
{
  insist (this && cluster && cluster->canvas);
  GnomeCanvasItem*item;
  GdkFont*font;
  char s [64];

  mu_Movable::build (gnome_canvas_root (cluster->canvas), x, y, 8 * MU_CLUSTER_PORT_TOTAL_WIDTH, (backplaneType + 1) * MU_CLUSTER_PORT_TOTAL_HEIGHT);
  insist (topGroup && selectGroup);
  
  outsideGroup = GNOME_CANVAS_GROUP (gnome_canvas_item_new (topGroup,
							    gnome_canvas_group_get_type (),
							    "x", 0.0,
							    "y", 0.0,
							    NULL));
  insist (outsideGroup);
  gnome_canvas_item_show (GNOME_CANVAS_ITEM (outsideGroup));

  for (int i = 0; i <= backplaneType; i++)
  {

    linecardItems [i] = gnome_canvas_item_new (outsideGroup, gnome_canvas_rect_get_type (),
					       "x1", (double) MU_CLUSTER_SELECT_OFFSET + MU_CLUSTER_SELECT_SIZE,
					       "y1", (double) MU_CLUSTER_SELECT_OFFSET + MU_CLUSTER_SELECT_SIZE + i * MU_CLUSTER_PORT_TOTAL_HEIGHT,
					       "x2", (double) MU_CLUSTER_SELECT_OFFSET + MU_CLUSTER_SELECT_SIZE + 8 * MU_CLUSTER_PORT_TOTAL_WIDTH,
					       "y2", (double) MU_CLUSTER_SELECT_OFFSET + MU_CLUSTER_SELECT_SIZE + (i + 1) * MU_CLUSTER_PORT_TOTAL_HEIGHT,
					       "fill_color", MU_CLUSTER_OBJECT_COLOR,
					       "outline_color", "black",
					       "width_pixels", 1,
					       NULL);
    insist (linecardItems [i]);
    gnome_canvas_item_show (linecardItems [i]);

    if (i && linecards [i - 1].type)
    {
      for (int j = 0; j < 8; j++)
      {
	double x1, y1, x2, y2;

	x1 = j * MU_CLUSTER_PORT_TOTAL_WIDTH;
	y1 = MU_CLUSTER_SELECT_OFFSET + MU_CLUSTER_SELECT_SIZE + i * MU_CLUSTER_PORT_TOTAL_HEIGHT;
	x2 = (j + 1) * MU_CLUSTER_PORT_TOTAL_WIDTH;
	y2 = MU_CLUSTER_SELECT_OFFSET + MU_CLUSTER_SELECT_SIZE + (i + 1) * MU_CLUSTER_PORT_TOTAL_HEIGHT;
	
	item = gnome_canvas_item_new (outsideGroup, gnome_canvas_rect_get_type (),
				      "x1", x1 + MU_CLUSTER_PORT_UNITS * 5,
				      "y1", y1 + MU_CLUSTER_PORT_UNITS * 2,
				      "x2", x2 - MU_CLUSTER_PORT_UNITS * 5,
				      "y2", y2 - MU_CLUSTER_PORT_UNITS * 2,
				      "fill_color", MU_CLUSTER_LINECARD_PORT_COLOR,
				      "outline_color", "black",
				      "width_pixels", 1,
				      NULL);
	insist (item);
	gnome_canvas_item_show (item);

	linecardPorts [i - 1] [M3_MAX_LINE_CARD_PORT - 1 - j].portItem = item;
	linecardPorts [i - 1] [M3_MAX_LINE_CARD_PORT - 1 - j].parent = this;
	gtk_signal_connect (GTK_OBJECT (item), "event", (GtkSignalFunc) mu_Cluster_linecardPort, 
			    &linecardPorts [i - 1] [M3_MAX_LINE_CARD_PORT - 1 - j]);

	item = gnome_canvas_item_new (outsideGroup, gnome_canvas_rect_get_type (),
				      "x1", x2 - MU_CLUSTER_PORT_UNITS * 4,
				      "y1", y2 - MU_CLUSTER_PORT_UNITS * 3,
				      "x2", x2 - MU_CLUSTER_PORT_UNITS,
				      "y2", y2 - MU_CLUSTER_PORT_UNITS * 2,
				      "fill_color", MU_CLUSTER_OBJECT_COLOR,
				      "outline_color", "black",
				      "width_pixels", 1,
				      NULL);
	insist (item);
	gnome_canvas_item_show (item);
	linecardPorts [i - 1] [M3_MAX_LINE_CARD_PORT - 1 - j].ledItem = item;
	
	sprintf (s, "%d", M3_MAX_LINE_CARD_PORT - 1 - j + (linecards [i - 1].xbar.index ? 8 : 0));
	
	item = gnome_canvas_item_new (outsideGroup, gnome_canvas_text_get_type (),
				      "text", s,
				      "anchor", GTK_ANCHOR_NORTH_WEST,
				      "font", MU_CLUSTER_FONT,
				      "x", (double) x2 - 4 * MU_CLUSTER_PORT_UNITS,
				      "y", (double) y1 + 2 * MU_CLUSTER_PORT_UNITS,
				      "justification", GTK_JUSTIFY_LEFT,
				      "fill_color", "black",
				      "clip", 1, 
				      "clip_width", (double) MU_CLUSTER_PORT_UNITS * 4,
				      "clip_height", (double) MU_CLUSTER_PORT_UNITS * 2, 
				      NULL);

	gnome_canvas_item_show (item);
      }
    }
  }

  sprintf (s, "%d.%d.%d.%d (%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x)",
	   (unsigned) (((char*)(&ipAddress)) [0] & 0xff),
	   (unsigned) (((char*)(&ipAddress)) [1] & 0xff),
	   (unsigned) (((char*)(&ipAddress)) [2] & 0xff),
	   (unsigned) (((char*)(&ipAddress)) [3] & 0xff),
	   ethernetAddress [0] & 0xff,
	   ethernetAddress [1] & 0xff,
	   ethernetAddress [2] & 0xff, 
	   ethernetAddress [3] & 0xff,
	   ethernetAddress [4] & 0xff,
	   ethernetAddress [5] & 0xff);
  
  item = gnome_canvas_item_new (topGroup, gnome_canvas_text_get_type (),
				"text", s,
				"anchor", GTK_ANCHOR_NORTH_WEST,
				"font", MU_CLUSTER_FONT,
				"x", (double) MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET,
				"y", (double) MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET,
				"justification", GTK_JUSTIFY_LEFT,
				"fill_color", "black",
				"clip", 1, 
				"clip_width", (double) 8 * MU_CLUSTER_PORT_TOTAL_WIDTH,
				"clip_height", (double) MU_CLUSTER_PORT_TOTAL_HEIGHT,
				NULL);

  gnome_canvas_item_show (item);


  insideGroup = GNOME_CANVAS_GROUP (gnome_canvas_item_new (topGroup,
							   gnome_canvas_group_get_type (),
							   "x", 0.0,
							   "y", 0.0,
							   NULL));
  insist (insideGroup);
  gnome_canvas_item_hide (GNOME_CANVAS_ITEM (insideGroup));

  for (int i = 0; i < backplaneType; i++)
  {
    if (linecards [i].type && linecards [i].xbar.index)
    {
      insist (linecards [i].xbar.index > 0 && linecards [i].xbar.index <= M3_MAX_LINE_CARD);
      xbars [linecards [i].xbar.index - 1].parent = this;
      
      if (!xbars [linecards [i].xbar.index - 1].build (insideGroup,
						       MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET + i * (8 * MU_CLUSTER_PORT_TOTAL_WIDTH) / backplaneType,
						       MU_CLUSTER_PORT_CONNECTOR_HEIGHT +
						       MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET + (backplaneType - 1) * MU_CLUSTER_PORT_TOTAL_HEIGHT))
	return 0;
    }
  }

  for (int i = 0; i < numBackplaneXbars; i++)
  {
    insist (backplaneXbars [i].index > 0 && backplaneXbars [i].index <= M3_MAX_LINE_CARD + M3_MAX_BACKPLANE_XBAR);
    xbars [backplaneXbars [i].index - 1]. parent = this;
	  
    if (!xbars [backplaneXbars [i].index - 1].build (insideGroup,
						     MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET + i * (8 * MU_CLUSTER_PORT_TOTAL_WIDTH) / (backplaneType / 2),
						     MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET + MU_CLUSTER_PORT_TOTAL_HEIGHT))
      return 0;
  } 
  
  exception: return 0;
}

GdkCursor*mu_Cluster::makeCursor (GdkWindow*window, char*data, char*maskData, int width, int height)
{
  GdkCursor*cursor;
  GdkBitmap*bitmap,*mask;
  GdkColor black, white;

  insist (this && window && data && maskData && width > 0 && height > 0);
  
  gdk_color_black (gdk_window_get_colormap (window), &white);
  gdk_color_white (gdk_window_get_colormap (window), &black);

  bitmap = gdk_bitmap_create_from_data (window, data, width, height);
  insist (bitmap);
  mask = gdk_bitmap_create_from_data (window, maskData, width, height);
  insist (mask);
  
  cursor = gdk_cursor_new_from_pixmap (bitmap, mask, &white, &black, 1, 0);
  insist (cursor);

  gdk_bitmap_unref (bitmap);
  gdk_bitmap_unref (mask);

  return cursor;
  exception: return 0;
}

GtkWidget*mu_Cluster::makeToggleButton (GtkWidget*toolbar, char**data, char*description, GSList*group, GdkWindow*window, GtkStyle*style)
{
  GtkWidget*gtkpixmap;
  GdkPixmap*pixmap;
  GdkBitmap*mask;
  GtkWidget*toggle;
  
  insist (this && toolbar && data && description  && window && style);
  
  pixmap = gdk_pixmap_create_from_xpm_d (window, &mask, &style->bg [GTK_STATE_NORMAL], data);
  insist (pixmap);  

  gtkpixmap = gtk_pixmap_new (pixmap, mask);
  insist (gtkpixmap);
  gtk_widget_show (gtkpixmap);
  
  toggle = gtk_radio_button_new (group);
  insist (toggle);
  gtk_toggle_button_set_mode (GTK_TOGGLE_BUTTON (toggle), 0);
  gtk_widget_show (toggle);
  gtk_container_add (GTK_CONTAINER (toggle), gtkpixmap);
  gtk_toolbar_append_widget (GTK_TOOLBAR (toolbar), toggle, description, NULL);

  gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
		      GTK_SIGNAL_FUNC (mu_Cluster_toggled),
		      this);
  
  gdk_pixmap_unref (pixmap);
  if (mask)
    gdk_bitmap_unref (mask);
  return toggle;
  exception: return 0;
}

mu_Cluster::mu_Cluster (GtkWidget*mainWidget)
{
  GtkWidget*box, *button, *frame, *scrolledWindow;
  GtkStyle*style;
  GtkWidget*toolbar;
  GSList*group;
  
  insist (mainWidget && mainWidget->window);
  
  pthread_mutex_init (&mutex, 0);

  widget = gtk_vbox_new (0, MU_SPACING);
  insist (widget);  
  
  style = gtk_widget_get_style (mainWidget);
  insist (style);
  
  box = gtk_hbox_new (0, 0);
  insist (box);
  gtk_widget_show (box);
  gtk_box_pack_start (GTK_BOX (widget), box, 0, 0, 0); 

  toolbar = gtk_toolbar_new (GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_BOTH);
  insist (toolbar);

  gtk_widget_show (toolbar);
  gtk_container_set_border_width (GTK_CONTAINER (toolbar), MU_SPACING);
  gtk_box_pack_start (GTK_BOX (box), toolbar, 0, 0, MU_SPACING); 

  GtkWidget*gtkpixmap;
  GdkPixmap*pixmap;
  GdkBitmap*mask;
  GdkColor white;
    
  gdk_color_white (gdk_window_get_colormap (mainWidget->window), &white);

  pixmap = gdk_pixmap_create_from_xpm_d (mainWidget->window, &mask, &white, myricom_xpm);
  insist (pixmap);  

  gtkpixmap = gtk_pixmap_new (pixmap, mask);
  insist (gtkpixmap);
  gtk_widget_show (gtkpixmap);
  gtk_box_pack_end (GTK_BOX (box), gtkpixmap, 0, 0, MU_SPACING);

  pointToggle = makeToggleButton (toolbar, point_xpm, "point", 0, mainWidget->window, style);
  insist (pointToggle);
  
  group = gtk_radio_button_group (GTK_RADIO_BUTTON (pointToggle));
  insist (group);
  
  zoomToggle = makeToggleButton (toolbar, zoom_xpm, "zoom in/out", 
				 gtk_radio_button_group (GTK_RADIO_BUTTON (pointToggle)), mainWidget->window, style);

  insist (zoomToggle);

  handToggle = makeToggleButton (toolbar, hand_xpm, "scroll", 
				 gtk_radio_button_group (GTK_RADIO_BUTTON (zoomToggle)), mainWidget->window, style);

  insist (handToggle);

  questionToggle = makeToggleButton (toolbar, question_xpm, "info",
				     gtk_radio_button_group (GTK_RADIO_BUTTON (handToggle)), mainWidget->window, style);

  changeToggle = makeToggleButton (toolbar, change_xpm, "control",
				     gtk_radio_button_group (GTK_RADIO_BUTTON (questionToggle)), mainWidget->window, style);

  insist (handToggle);
  
  hadjustment = (GtkAdjustment*) gtk_adjustment_new (0, 0, 3 * MU_CLUSTER_SIZE, 1, 1, 1);
  vadjustment = (GtkAdjustment*) gtk_adjustment_new (0, 0, 3 * MU_CLUSTER_SIZE, 1, 1, 1);
  
  pointCursor = makeCursor (mainWidget->window, (char*) point_bits, (char*) point_bits, point_width, point_height);
  insist (pointCursor);
  zoomCursor = makeCursor (mainWidget->window, (char*) zoom_bits, (char*) zoom_bits, zoom_width, zoom_height);
  insist (zoomCursor);
  handCursor = makeCursor (mainWidget->window, (char*) hand_bits, (char*) hand_bits, hand_width, hand_height);
  insist (handCursor);

  questionCursor = makeCursor (mainWidget->window, (char*) question_bits, (char*) question_bits, question_width, question_height);
  insist (questionCursor);

  changeCursor = makeCursor (mainWidget->window, (char*) change_bits, (char*) change_bits, change_width, change_height);
  insist (changeCursor);
  
  canvas = (GnomeCanvas*) gnome_canvas_new ();
  insist (canvas);

  scrolledWindow = gtk_scrolled_window_new (hadjustment, vadjustment);
  insist (scrolledWindow);
  
  gtk_box_pack_start (GTK_BOX (widget), scrolledWindow, 1, 1, MU_SPACING);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledWindow), GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);
  gnome_canvas_set_scroll_region (canvas, -MU_CLUSTER_SIZE, -MU_CLUSTER_SIZE, 2 * MU_CLUSTER_SIZE, 2 * MU_CLUSTER_SIZE);
  gtk_container_add (GTK_CONTAINER (scrolledWindow), GTK_WIDGET (canvas));
  gtk_widget_show (scrolledWindow);
  gtk_widget_set_usize (scrolledWindow, 700, 700);

  gtk_widget_show (GTK_WIDGET (canvas));
  
  linkGroup = GNOME_CANVAS_GROUP (gnome_canvas_item_new (gnome_canvas_root (canvas),
							 gnome_canvas_group_get_type (),
							 "x", 0.0,
							 "y", 0.0,
							 NULL));
  insist (linkGroup);  

  bounds = gnome_canvas_item_new (gnome_canvas_root (canvas), gnome_canvas_rect_get_type (),
				  "x1", (double) 0,
				  "y1", (double) 0,
				  "x2", (double) MU_CLUSTER_SIZE,
				  "y2", (double) MU_CLUSTER_SIZE,
				  "outline_color", "black",
				  "fill_color", "white",
				  "width_pixels", 1,
				  NULL);
  insist (bounds);
  gnome_canvas_item_show (bounds);

  selection = gnome_canvas_item_new (gnome_canvas_root (canvas), gnome_canvas_rect_get_type (),
				     "x1", 0.0,
				     "y1", 0.0,
				     "x2", 1.0,
				     "y2", 1.0,
				     "outline_color", "darkgrey",
				     "width_pixels", 1,
				     NULL);
  gnome_canvas_item_hide (selection);

  gnome_canvas_set_pixels_per_unit (canvas, 0.36);
  gtk_adjustment_set_value (hadjustment, 2400);
  gtk_adjustment_set_value (vadjustment, 2400);

  gtk_signal_connect (GTK_OBJECT (canvas), "event", (GtkSignalFunc) mu_Cluster_canvas, this);
  
  selected = 0;
  switches = 0;
  hosts = 0;
  links = 0;
  numSwitches = 0;
  lastEventType = lastEventIndex = 0;
  building = 0;
  routeHost = 0;
  
  exception:;
}

int mu_Cluster::eradicate ()
{
  insist (this);
  free_freeables (hosts);
  if (switches)
  {
    for (int i = 0; i < numSwitches; i++)
    {
      if (switches [i])
	delete switches [i];
    }
    delete (switches);
  }
  numSwitches = 0;
  switches = 0;

  free_freeables (links);

  if (linkGroup)
    gtk_object_destroy (GTK_OBJECT (linkGroup));
  
  linkGroup = GNOME_CANVAS_GROUP (gnome_canvas_item_new (gnome_canvas_root (canvas),
							 gnome_canvas_group_get_type (),
							 "x", 0.0,
							 "y", 0.0,
							 NULL));
  insist (linkGroup);

  links = hosts = 0;

  return 1;
  exception: return 0;  
}

int mu_Cluster::build (mt_Graph*graph, m3_XbarTable*table, int*addresses, int numAddresses, char*directory)
{
  insist (this && canvas && graph && table && addresses && numAddresses);
  building = 1;
  
  gdk_threads_enter ();
  eradicate ();
  gdk_threads_leave ();

  populate (addresses, numAddresses, directory);
  
  m3_Component::printFormat ("building graph...");
  
  gdk_threads_enter ();
  build ();
  gdk_threads_leave ();

  m3_Component::printFormat ("connecting graph...");

  gdk_threads_enter ();
  connect (graph, table);
  gnome_canvas_item_raise_to_top (GNOME_CANVAS_ITEM (linkGroup));
  //tree ();
  gdk_threads_leave ();
  building = 0;
  
  exception: return 0;
}

int mu_Cluster::populate (int*addresses, int numAddresses, char*directory)
{
  insist (this && canvas && addresses && numAddresses);
  insist (numSwitches == 0);
  char switchname [M3_DISPLAY_STRING_LENGTH + 1];

  switches = new mu_Switch* [numAddresses];
  insist (switches);

  int numHelpers;
  numHelpers = mu_maxThreads;
  if (numAddresses < numHelpers)
    numHelpers = numAddresses;

  {
    mu_SwitchHelper helpers [numHelpers];
  
    for (numSwitches = 0; numSwitches < numAddresses; numSwitches++)
    {
      switches [numSwitches] = new mu_Switch (this);
      insist (switches [numSwitches] && addresses [numSwitches]);
      sprintf (switchname, "%d.%d.%d.%d",
	       (unsigned) (((char*)(&addresses [numSwitches])) [0] & 0xff),
	       (unsigned) (((char*)(&addresses [numSwitches])) [1] & 0xff),
	       (unsigned) (((char*)(&addresses [numSwitches])) [2] & 0xff),
	       (unsigned) (((char*)(&addresses [numSwitches])) [3] & 0xff));

      helpers [numSwitches % numHelpers].join ();

      helpers [numSwitches % numHelpers]._switch = switches [numSwitches];
      helpers [numSwitches % numHelpers].filename [0] = 0;
      helpers [numSwitches % numHelpers].address = 0;

      if (directory)
      {
	sprintf (helpers [numSwitches % numHelpers].filename, "%s/%s.html", directory, switchname);
	mt_Component::printFormat ("reading %s", helpers [numSwitches % numHelpers].filename);
	helpers [numSwitches % numHelpers].createThread ((mu_ThreadableFunction) &mu_SwitchHelper::run, 0);
      }
      else
      {
	mt_Component::printFormat ("examining %s", switchname);
	helpers [numSwitches % numHelpers].address = addresses [numSwitches];      
	helpers [numSwitches % numHelpers].createThread ((mu_ThreadableFunction) &mu_SwitchHelper::run, 0);
      }
    }

    for (int i = 0; i < numHelpers; i++)
      helpers [i].join ();
  }
  
  return 1;
  exception: return 0;
}

int mu_Cluster::connect (mt_Graph*graph, m3_XbarTable*table)
{
  insist (this && graph && table);

  for (int i = 0; i < graph->getNumSwitches (); i++)
  {
    m3_XbarTableEntry*e1,*e2;
    mu_Switch*s1,*s2;
    mt_Node*n1, *n2;
    m3_DescriptionXbar*x1,*x2;

    n1 = graph->getSwitch (i);
    insist (n1);
    e1 = table->lookup (n1);
    insist (e1);
    
    s1 = getSwitch (e1->ethernetAddress);
    insist (s1);    
    insist (e1->xbar > 0 && e1->xbar <= s1->getNumXbars ());

    x1 = s1->getXbar (e1->xbar);
    insist (x1);    
    insist (x1->index > 0 && x1->index <= M3_MAX_LINE_CARD + M3_MAX_BACKPLANE_XBAR);

    for (int p = 0; p < 16; p++)
    {
      int pp = p + e1->correction;

      if ((n2 = n1->getNode (p)))
      {
	int o, oo;
	insist (p >= 0 && p < 16);
	o = n1->getOpposite (p);
	insist (o >= 0 && o < 16);

	if (n2->isHost ())
	{
	  mu_Host*h = new mu_Host (&hosts, n2, this);
	  insist (h);

	  mu_Link*link;
	  link = new mu_Link (&links);
	  insist (link);
	  s1->xbars [x1->index - 1].ports [pp].link = h->link = link;

	  link->port = &s1->xbars [x1->index - 1].ports [pp];
	  link->port->port = &x1->ports [pp];
	  link->other = h;
	  link->other->port = 0;

	  continue;
	}
	
	e2 = table->lookup (n2);
	insist (e2);
	s2 = getSwitch (e2->ethernetAddress);
	insist (s2);

	insist (e2->xbar > 0 && e2->xbar <= s2->getNumXbars ());

	x2 = s2->getXbar (e2->xbar);
	insist (x2);

	insist (x2->index > 0 && x2->index <= M3_MAX_LINE_CARD + M3_MAX_BACKPLANE_XBAR);

	oo = o + e2->correction;
	insist (pp >= 0 && pp < 16);
	insist (oo >= 0 && oo < 16);	
	
	if (s1->xbars [x1->index - 1].ports [pp].link)
	{
	  insist (s1->xbars [x1->index - 1].ports [pp].link->other);
	  insist (s1->xbars [x1->index - 1].ports [pp].link->port);
	  insist (s2->xbars [x2->index - 1].ports [oo].link); 
	  insist (s2->xbars [x2->index - 1].ports [oo].link->port);
	  insist (s2->xbars [x2->index - 1].ports [oo].link->other);
	  continue;
	}
	insist (!s2->xbars [x2->index - 1].ports [oo].link);
	
	mu_Link*link;
	link = new mu_Link (&links);
	insist (link);
	s2->xbars [x2->index - 1].ports [oo].link = s1->xbars [x1->index - 1].ports [pp].link = link;
	
	link->port = &s1->xbars [x1->index - 1].ports [pp];
	link->port->port = &x1->ports [pp];

	link->other = &s2->xbars [x2->index - 1].ports [oo];
	link->other->port = &x2->ports [oo];

	if (!link->build (s1 == s2 ? s1->insideGroup: linkGroup))
	  return 0;
      }
    }
  }

  for (int i = 0; i < numSwitches; i++)
    if (!switches [i]->connect ())
      return 0;
  
  return 1;
  exception: return 0;
}

int mu_Switch::connect ()
{
  insist (this);
  
  for (int i = 0; i < M3_MAX_LINE_CARD; i++)
  {
    if (linecards [i].type)
    {
      for (int j = 0; j < M3_MAX_LINE_CARD_PORT; j++)
      {
	int p;
	m3_DescriptionXbar*xbar;
	m3_DescriptionPort*port = linecards [i].ports [j].port;

	insist (port);
	
	linecardPorts [i] [j].port = port;
	xbar = (m3_DescriptionXbar*) port->parent;
	insist (xbar);
	
	p = ((port->index - 1) % 16);
	insist (p >= 0 && p < 16);

	insist (linecardPorts [i] [j].ledItem);

	if (xbars [xbar->index - 1].ports[p].link)
	{
	  gnome_canvas_item_set (linecardPorts [i][j].ledItem,  "fill_color", linecards [i].ports [j].portDown ? "red" : "green", NULL);

	  insist (xbars [xbar->index - 1].ports[p].link->port);
	  insist (xbars [xbar->index - 1].ports[p].link->other);
	  
	  insist (xbars [xbar->index - 1].ports[p].link->port->port);
	  if (!xbars [xbar->index - 1].ports[p].link->other->port)
	  {
	    double x1, y1, x2, y2;
	    mu_Host*h;
	    
	    gnome_canvas_item_get_bounds (linecardPorts [i][j].ledItem, &x1, &y1, &x2, &y2);
	    gnome_canvas_item_i2w (linecardPorts [i][j].ledItem, &x2, &y2);

	    h = (mu_Host*) xbars [xbar->index - 1].ports[p].link->other;
	    h->build (cluster->linkGroup, x2 + MU_CLUSTER_PORT_TOTAL_HEIGHT / 4 , y2 + MU_CLUSTER_PORT_TOTAL_HEIGHT / 4);
	    linecardPorts [i][j].host = h;
	    
	    if (!xbars [xbar->index - 1].ports[p].link->build (cluster->linkGroup))
	      return 0;
	  }
	}   
      }
    }
  }
  return 1;
  exception: return 0;
}

mu_Switch*mu_Cluster::getSwitch (char*ethernetAddress)
{
  insist (this && ethernetAddress);

  for (int i = 0; i < numSwitches; i++)
  {
    insist (switches [i]);
    if (!memcmp (ethernetAddress, switches [i]->ethernetAddress, 6))
      return switches [i];
  }
  exception: return 0;
}

mu_Switch*mu_Cluster::getSwitchFromIpAddress (char*ipAddress)
{
  insist (this && ipAddress);

  for (int i = 0; i < numSwitches; i++)
  {
    insist (switches [i]);
    if (!memcmp (ipAddress, &switches [i]->ipAddress, 4))
      return switches [i];
  }

  exception: return 0;
}

int mu_Cluster::build ()
{
  insist (this && switches);
  for (int i = 0; i < numSwitches; i++)
  {
    insist (switches [i]);
    switches [i]->build (MU_CLUSTER_SIZE / 2 + i * 100, MU_CLUSTER_SIZE / 2 + i * 100);
    switches [i]->show ();
  }
  return 1;
  exception: return 0;
}

int mu_Cluster::zoom (double x, double y, double zoom)
{
  insist (this && canvas && zoom > 0.0);
  double ppu;  

  if (zoom > 1.0)
  {
    double ax, ay;

    ax = x - ((double) GTK_WIDGET (canvas)->allocation.width / 2);
    ay = y - ((double) GTK_WIDGET (canvas)->allocation.height / 2);    

    gtk_adjustment_set_value (hadjustment, (float) hadjustment->value + ax);
    gtk_adjustment_set_value (vadjustment, (float) vadjustment->value + ay);
  }
  
  ppu = canvas->pixels_per_unit * zoom;
  gnome_canvas_set_pixels_per_unit (canvas, ppu);

  return 1;
  exception: return 0;		      
}

int mu_Cluster::inside (double x, double y, double x1, double y1, double x2, double y2)
{
  insist (this);
  insist (x2 >= x1 && y2 >= y1);
  
  return x >= x1 && x <= x2 && y >= y1 && y <= y2;

  exception: return 0;
}

int mu_Cluster::inside (GnomeCanvasItem*item, double x1, double y1, double x2, double y2)
{
  insist (this && item);
  insist (x2 >= x1 && y2 >= y1);
  
  double ix1, ix2, iy1, iy2;
  
  gnome_canvas_item_get_bounds (item, &ix1, &iy1, &ix2, &iy2);
  
  return 
    inside (ix1, iy1, x1, y1, x2, y2) &&
    inside (ix1, iy2, x1, y1, x2, y2) &&
    inside (ix2, iy1, x1, y1, x2, y2) &&
    inside (ix2, iy2, x1, y1, x2, y2);
  
  exception: return 0;
}

int mu_Cluster::inside (GnomeCanvasItem*item, double x, double y)
{
  insist (this && item);
  
  double x1, x2, y1, y2;
  gnome_canvas_item_get_bounds (item, &x1, &y1, &x2, &y2);
  return inside (x, y, x1, y1, x2, y2);
  
  exception: return 0;
}

int mu_Cluster::intersects (GnomeCanvasItem*item, double x1, double y1, double x2, double y2)
{
  insist (this && item);
  
  double ix1, ix2, iy1, iy2;
  gnome_canvas_item_get_bounds (item, &ix1, &iy1, &ix2, &iy2);

  return ! (x2 < ix1 || x1 > ix2 || y2 < iy1 || y1 > iy2);
  
  exception: return 0;
}

int mu_Cluster::inside (double x, double y)
{
  insist (this);
  
  for (int i = 0; i < numSwitches; i++)
    if (switches [i]->inside (x, y))
      return 1;

  for (mu_Host*h = (mu_Host*) hosts; h; h = (mu_Host*) h->next)
    if (h->inside (x, y))
      return 1;

  exception: return 0;
}

mu_Host*mu_Cluster::getHost (int index)
{
  insist (this && index >= 0);

  for (mu_Host*h = (mu_Host*) hosts; h; h = (mu_Host*) h->next)
    if (h->host && h->host->getTypeIndex () == index)
      return h;
  exception: return 0;
}

int mu_Cluster::selectAll (int selected)
{
  insist (this);
  
  for (int i = 0; i < numSwitches; i++)
    switches [i]->select (selected, 1);
  for (mu_Host*h = (mu_Host*) hosts; h; h = (mu_Host*) h->next)
    h->select (selected, 1);
  
  return 1;
  exception: return 0;
}

int mu_Cluster::selectAll (double x1, double y1, double x2, double y2, int selected)
{
  insist (this);
  
  for (int i = 0; i < numSwitches; i++)
    if (switches [i]->inside (x1, y1, x2, y2))
      switches [i]->select (selected, 0);  

  for (mu_Host*h = (mu_Host*) hosts; h; h = (mu_Host*) h->next)
    if (h->inside (x1, y1, x2, y2))
      h->select (selected, 0);

  return 1;
  exception: return 0;
}

int mu_Cluster::select (mu_Movable*m)
{
  insist (this);
  
  if ((this->selected = m))
    m->select (1, 1);
  return 1;
  exception: return 0;
}

int mu_Cluster::toggled ()
{
  insist (this);
  insist (zoomToggle && pointToggle && zoomCursor && pointCursor && handCursor && handToggle);
  insist (questionCursor && questionToggle);
  insist (changeCursor && changeToggle);
  
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (pointToggle)))
    gdk_window_set_cursor (widget->window, pointCursor);
  else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (zoomToggle)))
    gdk_window_set_cursor (widget->window, zoomCursor);
  else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (handToggle)))
    gdk_window_set_cursor (widget->window, handCursor);
  else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (questionToggle)))
    gdk_window_set_cursor (widget->window, questionCursor);
  else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (changeToggle)))
    gdk_window_set_cursor (widget->window, changeCursor);
  
  return 1;
  exception: return 0;
}

mu_Freeable::mu_Freeable (mu_Freeable**head)
{
  insist (head);
  next = *head;
  *head = this;
  exception:;
}

mu_Link::mu_Link (mu_Freeable**head) : mu_Freeable (head)
{
  item = 0;
  port = other = 0;

  points = gnome_canvas_points_new (2);
  insist (points);
  points->num_points = 2;
  routeCount = 0;
  
  exception:;
}

mu_Link::~mu_Link ()
{
  if (points)
    gnome_canvas_points_free (points);
}

int mu_Link::build (GnomeCanvasGroup*group)
{
  insist (this && group);
  insist (port && other && points);
  insist (port->port || other->port);
  insist (!item);
  
  //if (!port->port || !other->port || port->port->parent != other->port->parent)
  {
    if (!port->getCenter (&points->coords [0], &points->coords [1]) ||
	!other->getCenter (&points->coords [2], &points->coords [3]))
      return 0;
    
    insist (mu_mainWidget);
    
    mu_Cluster*cluster;
    cluster = mu_mainWidget->cluster;
    insist (cluster);
    
    if (cluster->linkGroup != group)
    {
      gnome_canvas_item_w2i (GNOME_CANVAS_ITEM (group), &points->coords [0], &points->coords [1]);
      gnome_canvas_item_w2i (GNOME_CANVAS_ITEM (group), &points->coords [2], &points->coords [3]);
    }
    
    item = gnome_canvas_item_new (group,
				  gnome_canvas_line_get_type (),
				  "points", points,
				  "fill_color", "black",
				  "width_pixels", 1,
				  NULL);
    insist (item);
    gnome_canvas_item_show (item);
  }
  
  return 1;
  exception: return 0;
}

mu_Host::mu_Host (mu_Freeable**head, mt_Node*node, mu_Cluster*cluster) : mu_Freeable (head)
{
  insist (head && node && node->isHost () && cluster);
  host = node;
  this->cluster = cluster;
  routeSelected = 0;
  
  exception:;
}


int mu_Host::build (GnomeCanvasGroup*group, double x, double y)
{
  GnomeCanvasItem*item;
  
  insist (group);
  
  mu_Movable::build (group, x, y, MU_CLUSTER_HOST_WIDTH, MU_CLUSTER_HOST_HEIGHT);
  insist (topGroup && selectGroup);
  insist (host && host->getName ());
  
  gnome_canvas_item_raise_to_top (GNOME_CANVAS_ITEM (topGroup));

  item = gnome_canvas_item_new (topGroup, gnome_canvas_text_get_type (),
				"text", host->getName (),
				"anchor", GTK_ANCHOR_NORTH_WEST,
				"font", MU_CLUSTER_FONT,
				"x", (double) MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET,
				"y", (double) MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET,
				"justification", GTK_JUSTIFY_LEFT,
				"fill_color", "black",
				"clip", 1, 
				"clip_width", (double) MU_CLUSTER_HOST_WIDTH,
				"clip_height", (double) MU_CLUSTER_HOST_HEIGHT,
				NULL);
  insist (item);
  gnome_canvas_item_show (item);

  exception: return 0;
}

int mu_Movable::build (GnomeCanvasGroup*group, double x, double y, int width, int height)
{
  insist (width > 0 && height > 0);
  insist (group);
  
  GnomeCanvasItem*item;
  double xcorner, ycorner;
  xcorner = width + MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET;
  ycorner = height + MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET;

  topGroup = GNOME_CANVAS_GROUP (gnome_canvas_item_new (group,
							gnome_canvas_group_get_type (),
							"x", x,
							"y", y,
							NULL));

  insist (topGroup);
  
  box = gnome_canvas_item_new (topGroup, gnome_canvas_rect_get_type (),
			       "x1", (double) (MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET),
			       "y1", (double) MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET,
			       "x2", (double) width + MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET,
			       "y2", (double) height + MU_CLUSTER_SELECT_SIZE + MU_CLUSTER_SELECT_OFFSET,
			       "fill_color", MU_CLUSTER_OBJECT_COLOR,
			       "outline_color", "black",
			       "width_pixels", 1,
			       NULL);
  insist (box);
  gnome_canvas_item_show (box);

  selectGroup = GNOME_CANVAS_GROUP (gnome_canvas_item_new (topGroup,
							   gnome_canvas_group_get_type (),
							   "x", 0,
							   "y", 0,
							   NULL));
  insist (selectGroup);
  gnome_canvas_item_hide (GNOME_CANVAS_ITEM (selectGroup));

  item = gnome_canvas_item_new (selectGroup, gnome_canvas_rect_get_type (),
				"x1", 0.0,
				"y1", 0.0,
				"x2", (double) MU_CLUSTER_SELECT_SIZE,
				"y2", (double)  MU_CLUSTER_SELECT_SIZE,
				"fill_color", "black",
				"outline_color", "black",
				"width_pixels", 1,
				NULL);
  insist (item);
  gnome_canvas_item_show (item);

  item = gnome_canvas_item_new (selectGroup, gnome_canvas_rect_get_type (),
				"x1", (double) xcorner + MU_CLUSTER_SELECT_OFFSET,
				"y1", 0.0,
				"x2", (double) xcorner + MU_CLUSTER_SELECT_OFFSET + MU_CLUSTER_SELECT_SIZE,
				"y2", (double)  MU_CLUSTER_SELECT_SIZE,
				"fill_color", "black",
				"outline_color", "black",
				"width_pixels", 1,
				NULL);
  insist (item);
  gnome_canvas_item_show (item);

  item = gnome_canvas_item_new (selectGroup, gnome_canvas_rect_get_type (),
				"x1", 0.0,
				"y1", (double) ycorner + MU_CLUSTER_SELECT_OFFSET,
				"x2", (double) MU_CLUSTER_SELECT_SIZE,
				"y2", (double) ycorner + MU_CLUSTER_SELECT_OFFSET + MU_CLUSTER_SELECT_SIZE,
				"fill_color", "black",
				"outline_color", "black",
				"width_pixels", 1,
				NULL);
  insist (item);
  gnome_canvas_item_show (item);

  item = gnome_canvas_item_new (selectGroup, gnome_canvas_rect_get_type (),
				"x1", (double) xcorner + MU_CLUSTER_SELECT_OFFSET,
				"y1", (double) ycorner + MU_CLUSTER_SELECT_OFFSET,
				"x2", (double) xcorner + MU_CLUSTER_SELECT_OFFSET + MU_CLUSTER_SELECT_SIZE,
				"y2", (double) ycorner + MU_CLUSTER_SELECT_OFFSET + MU_CLUSTER_SELECT_SIZE,
				"fill_color", "black",
				"outline_color", "black",
				"width_pixels", 1,
				NULL);
  insist (item);
  gnome_canvas_item_show (item);

  gtk_signal_connect (GTK_OBJECT (topGroup), "event", (GtkSignalFunc) mu_Cluster_movable, this);

  return 1;
  exception: return 0;
}

int mu_Switch::move (double dx, double dy)
{
  insist (this);
  gnome_canvas_item_move (GNOME_CANVAS_ITEM (topGroup), dx, dy);

  for (int i = 0; i < M3_MAX_LINE_CARD; i++)
  {
    for (int j = 0; j < M3_MAX_LINE_CARD_PORT; j++)
    {
      if (linecardPorts [i][j].host && !linecardPorts [i][j].host->selected)
	linecardPorts [i][j].host->move (dx, dy);
    }
  }
  
  for (int i = 0; i < M3_MAX_BACKPLANE_XBAR + M3_MAX_LINE_CARD; i++)
  {
    for (int j = 0; j < 16; j++)
    {
      if (xbars [i].ports[j].link)
      {
	insist (xbars [i].ports[j].link->port && xbars [i].ports[j].link->other);
	
	if (xbars [i].ports[j].link->port->getSwitch () == xbars [i].ports[j].link->other->getSwitch ())
	  continue;

	if (xbars [i].ports[j].link->port->getSwitch () == this)
	  xbars [i].ports[j].link->move (dx, dy, 0, 0);
	if (xbars [i].ports[j].link->other->getSwitch () == this)
	  xbars [i].ports[j].link->move (0, 0, dx, dy);
      }
    }
  }
  
  return 1;
  exception: return 0;  
}

int mu_Link::move (double dx1, double dy1, double dx2, double dy2)
{
  insist (this);
  points->coords [0] += dx1;
  points->coords [1] += dy1;
  points->coords [2] += dx2;
  points->coords [3] += dy2;
  
  gnome_canvas_item_set (item,  "points", points, NULL);
  return 1;
  exception: return 0;
}

int mu_Host::move (double dx, double dy)
{
  insist (this && topGroup);
  gnome_canvas_item_move (GNOME_CANVAS_ITEM (topGroup), dx, dy);
  insist (link);
  insist (link->other == this);
  
  return link->move (0, 0, dx, dy);
  exception: return 0;
}

mu_Port::mu_Port ()
{
  link = 0;
  port = 0;
}

mu_Movable::mu_Movable ()
{
  cluster = 0;
  box = 0;
  topGroup = selectGroup = 0;
  selected = 0;
}

mu_LinecardPort::mu_LinecardPort ()
{
  port = 0;
  parent = 0;
  host = 0;
  portItem = 0;
  ledItem = 0;
}

int mu_Port::getCenter (double*x1, double*y1)
{
  insist (this);
  double x2, y2;

  if (!getBounds (x1, y1, &x2, &y2))
    return 0;
  
  *x1 += (x2 - *x1) / 2;
  *y1 += (y2 - *y1) / 2;
  
  return 1;
  exception: return 0;
}

int mu_Port::getBounds (double*x1, double*y1, double*x2, double*y2)
{
  insist (this && x1 && y1 && x2 && y2);
  
  if (!port)
  {
    insist (((mu_Host*)this)->box);
    gnome_canvas_item_get_bounds (((mu_Host*)this)->box, x1, y1, x2, y2);
    gnome_canvas_item_i2w (((mu_Host*)this)->box, x1, y1);
    gnome_canvas_item_i2w (((mu_Host*)this)->box, x2, y2);
    return 1;
  }
  insist (port->parent);
  
  mu_LinecardPort*p;

  if (p = getLinecardPort ())
  {  
    gnome_canvas_item_get_bounds (p->portItem, x1, y1, x2, y2);
    gnome_canvas_item_i2w (p->portItem, x1, y1);
    gnome_canvas_item_i2w (p->portItem, x2, y2);
  }
  else
  {
    mu_Switch*s = getSwitch ();
    insist (s);
    
    insist (port && port->index);
    insist (port->portNumber > 0 && port->portNumber <= 16);
    insist (&s->xbars [(port->index - 1) / 16].ports [port->portNumber - 1] == this);
    insist (s->xbars [(port->index - 1) / 16].portSquares [port->portNumber - 1]);
    
    gnome_canvas_item_get_bounds (s->xbars [(port->index - 1) / 16].portSquares [port->portNumber - 1], x1, y1, x2, y2);
    gnome_canvas_item_i2w (s->xbars [(port->index - 1) / 16].portSquares [port->portNumber - 1], x1, y1);
    gnome_canvas_item_i2w (s->xbars [(port->index - 1) / 16].portSquares [port->portNumber - 1], x2, y2);
  }
  
  return 1;
  exception: return 0;
}

mu_LinecardPort*mu_Port::getLinecardPort ()
{
  mu_Switch*s;
  int i, j;
  m3_DescriptionPort*p;

  insist (this && port && port->parent);
  p = port->parent->type == M3_LINECARD_TABLE ? port : port->port;

  if (!p || p->parent->type != M3_LINECARD_TABLE)
    return 0;

  insist (p && p->parent);

  insist (p->parent->parent && !p->parent->parent->type);
  
  s = (mu_Switch*)p->parent->parent;
  insist (s);
  
  i = ((m3_DescriptionLinecard*) p->parent)->slot - 1;
  j = (p->index - 1) % M3_MAX_LINE_CARD_PORT; //vast kludge

  insist (i >= 0 && i < M3_MAX_LINE_CARD);
  insist (j >= 0 && j < M3_MAX_LINE_CARD_PORT);
 
  return &s->linecardPorts [i][j];
  
  exception: return 0;
}


mu_Switch*mu_Port::getSwitch ()
{
  m3_DescriptionPart*p;  
  insist (this);

  for (p = port; p && p->parent; p = p->parent);

  return (mu_Switch*) p;

  exception: return 0;
}

mu_Xbar::mu_Xbar ()
{
  topGroup = pingGroup = 0;
  box = pingText = 0;
  parent = 0;
  memset (portSquares, 0, sizeof (portSquares));
  memset (pingSquares, 0, sizeof (pingSquares));
  memset (ports, 0, sizeof (ports));
}

int mu_Xbar::build (GnomeCanvasGroup*group, double x, double y)
{
  insist (this && group);
  GnomeCanvasItem*item;
  
  topGroup = GNOME_CANVAS_GROUP (gnome_canvas_item_new (group,
							gnome_canvas_group_get_type (),
							"x", x,
							"y", y,
							NULL));
  insist (topGroup);
  gnome_canvas_item_show (GNOME_CANVAS_ITEM (topGroup));
  
  box = gnome_canvas_item_new (topGroup, gnome_canvas_rect_get_type (),
			       "x1", (double) 0,
			       "y1", (double) 0,
			       "x2", (double) MU_XBAR_SIDE,
			       "y2", (double) MU_XBAR_SIDE,
			       "outline_color", "black",
			       "fill_color", MU_CLUSTER_OBJECT_COLOR,
			       "width_pixels", 1,
			       NULL);
  
  insist (box);
  gnome_canvas_item_show (box);

  for (int i = 0; i < 16; i++)
  {
    portSquares [i] =  gnome_canvas_item_new (topGroup, gnome_canvas_rect_get_type (),
					      "x1", (double) (i % 8) * MU_XBAR_SIDE / 8,
					      "y1", (double) (i < 8 ? 0 : 7) * MU_XBAR_SIDE / 8,
					      "x2", (double) ((i % 8) + 1) * MU_XBAR_SIDE / 8,
					      "y2", (double) (i < 8 ? 1 : 8) * MU_XBAR_SIDE / 8,
					      "outline_color", "black",
					      "fill_color", MU_CLUSTER_OBJECT_COLOR,
					      "width_pixels", 1,
					      NULL);
    insist (portSquares [i]);
    gnome_canvas_item_show (portSquares [i]);
    gtk_signal_connect (GTK_OBJECT (portSquares [i]), "event", (GtkSignalFunc) mu_Cluster_port, this);
  }

  pingGroup = GNOME_CANVAS_GROUP (gnome_canvas_item_new (topGroup,
							 gnome_canvas_group_get_type (),
							 "x", (double) MU_XBAR_SIDE / 8,
							 "y", (double) MU_XBAR_SIDE / 8,
							 NULL));
  insist (pingGroup);
  gnome_canvas_item_hide (GNOME_CANVAS_ITEM (pingGroup));
  
  double m;
  m = ((double) MU_XBAR_SIDE - (double) MU_XBAR_SIDE / 4.0) / 16.0;
  
  for (int i = 0; i < 16; i++)
  {
    for (int j = 0; j < 16; j++)
    {
      pingSquares [i] [j] =  gnome_canvas_item_new (pingGroup, gnome_canvas_rect_get_type (),
						    "x1", (double) i * m,
						    "y1", (double) j * m,
						    "x2", (double) (i + 1) * m,
						    "y2", (double) (j + 1) * m,
						    "outline_color", "black",
						    "width_pixels", 1,
						    NULL);
      insist (pingSquares [i][j]);
      gnome_canvas_item_show (pingSquares [i][j]);
    }
  }

  gtk_signal_connect (GTK_OBJECT (topGroup), "event", (GtkSignalFunc) mu_Cluster_xbar, this);
  
  return 1;
  exception: return 0;
}

int mu_Switch::showInside (int state)
{
  insist (this && insideGroup && outsideGroup);

  if (state)
  {
    gnome_canvas_item_hide (GNOME_CANVAS_ITEM (outsideGroup));
    gnome_canvas_item_show (GNOME_CANVAS_ITEM (insideGroup));
    gnome_canvas_item_set (box,  "fill_color", "white", NULL);
  }
  else
  {
    gnome_canvas_item_hide (GNOME_CANVAS_ITEM (insideGroup));
    gnome_canvas_item_show (GNOME_CANVAS_ITEM (outsideGroup));
    gnome_canvas_item_set (box,  "fill_color", MU_CLUSTER_OBJECT_COLOR, NULL);
  }
  return moveLinecardLinks (state);
  exception: return 0;
  
}

int mu_Cluster::showInside (int state)
{
  insist (this);
  
  for (int i = 0; i < numSwitches; i++)
    switches [i]->showInside (state);

  return 1;
  exception: return 0; 
}


int mu_Cluster::showXpoints (int state)
{
  insist (this);
  
  for (int i = 0; i < numSwitches; i++)
  {
    for (int j = 0; j < M3_MAX_LINE_CARD + M3_MAX_BACKPLANE_XBAR; j++)
    {
      if (switches [i]->xbars [j].pingGroup)
      {
	if (state)
	  gnome_canvas_item_show (GNOME_CANVAS_ITEM (switches [i]->xbars [j].pingGroup));
	else
	  gnome_canvas_item_hide (GNOME_CANVAS_ITEM (switches [i]->xbars [j].pingGroup));
      }
    }
  }
  return 1;
  exception: return 0; 
}


int mu_Switch::moveLinecardLinks (int inside)
{
  for (int i = 0; i < backplaneType; i++)
  {
    if (linecards [i].type)
    {
      for (int j = 0; j < 8; j++)
      {
	int p;
	m3_DescriptionXbar*xbar;
	m3_DescriptionPort*port = linecards [i].ports [j].port;

	if (!port)
	  continue;
	
	xbar = (m3_DescriptionXbar*) port->parent;
	insist (xbar);
	
	p = ((port->index - 1) % 16);
	insist (p >= 0 && p < 16);

	mu_Link*link = xbars [xbar->index - 1].ports[p].link;
	
	if (link && link->other != link->port)
	{
	  double x1, y1, x2, y2;
	  GnomeCanvasItem*item = inside ? xbars [xbar->index - 1].portSquares [p] : linecardPorts [i][j].portItem;
	  insist (item);

	  gnome_canvas_item_get_bounds (item, &x1, &y1, &x2, &y2);
	  gnome_canvas_item_i2w (item, &x1, &y1);
	  gnome_canvas_item_i2w (item, &x2, &y2);
	  x1 += (x2 - x1) / 2;
	  y1 += (y2 - y1) / 2;
	  if (&xbars [xbar->index - 1].ports[p] == link->other)
	    link->moveOther(x1, y1);
	  else
	    link->movePort (x1, y1);
	}
      }
    }
  }
  return 1;
  exception: return 0;
}

int mu_Link::movePort (double x, double y)
{ 
  insist (this);
  points->coords [0] = x;
  points->coords [1] = y;
  gnome_canvas_item_set (item, "points", points, NULL);
  return 1;
  exception: return 0;
}

int mu_Link::moveOther (double x, double y)
{ 
  insist (this);
  points->coords [2] = x;
  points->coords [3] = y;
  gnome_canvas_item_set (item, "points", points, NULL);
  return 1;
  exception: return 0;
}


int mu_Cluster::showHosts (int state)
{
  insist (this);
  for (int i = 0; i < numSwitches; i++)
    switches [i]->showHosts (state);
  return 1;
  exception: return 0;
}

int mu_Cluster::showCables (int state)
{
  insist (this);
  for (int i = 0; i < numSwitches; i++)
    switches [i]->showCables (state);
  return 1;
  exception: return 0;
}

int mu_Switch::showHosts (int state)
{
  insist (this);
  
  for (int i = 0; i < M3_MAX_BACKPLANE_XBAR + M3_MAX_LINE_CARD; i++)
  {
    for (int j = 0; j < 16; j++)
    {
      mu_Link*link = xbars [i].ports [j].link;
      if (link && link->port && link->other && !link->other->port)
      {
	if (state)
	{
	  gnome_canvas_item_show (link->item);
	  gnome_canvas_item_show (GNOME_CANVAS_ITEM (((mu_Host*)link->other)->topGroup));
	}
	else
	{
	  gnome_canvas_item_hide (link->item);
	  gnome_canvas_item_hide (GNOME_CANVAS_ITEM (((mu_Host*)link->other)->topGroup));
	}
      }
    }
  }
  return 1;
  exception: return 0;
}

int mu_Switch::showCables (int state)
{
  insist (this);
  
  for (int i = 0; i < M3_MAX_BACKPLANE_XBAR + M3_MAX_LINE_CARD; i++)
  {
    for (int j = 0; j < 16; j++)
    {
      mu_Link*link = xbars [i].ports [j].link;
      if (link && link->port && link->port->port && link->other && link->other->port &&
	  link->port->getSwitch () != link->other->getSwitch ())
      {
	if (state)
	  gnome_canvas_item_show (link->item);
	else
	  gnome_canvas_item_hide (link->item);
      }
    }
  }
  return 1;
  exception: return 0;
}

int mu_Cluster::eatCounters (int testingXbars, int numCounters, pi_Counter*counters)
{
  m3_XbarTable*table;
  mt_Graph*graph;
  
  insist (this && counters);
  insist (mu_mainWidget && mu_mainWidget->exploreWidget);

  table = mu_mainWidget->exploreWidget->xbarTable;
  graph = &mu_mainWidget->exploreWidget->graph;
  insist (table && graph);

  if (!testingXbars) 
  {
    for (int i = 0; i < numCounters; i++)
    {
      mu_Host*h = getHost (i);
      insist (h);
      if (!(h->eatCounter (&counters [i])))
	return 0;
    }
    return 1;
  }
  
  insist (numCounters == graph->getNumSwitches ());
  
  for (int i = 0; i < numCounters; i++)
  {
    mu_Switch*s;
    m3_XbarTableEntry*e = table->lookup (graph->getSwitch (i));
    insist (e && e->xbar > 0 && e->xbar <= M3_MAX_LINE_CARD + M3_MAX_BACKPLANE_XBAR);
    s = getSwitch (e->ethernetAddress);
    insist (s);
    if (!(s->eatCounter (e->xbar, e->correction, &counters [i])))
      return 0;
  }
  
  return 1;
  exception: return 0;
}


int mu_Switch::eatCounter (int xbar, int correction, pi_Counter*counter)
{
  insist (this && counter);
  insist (xbar > 0 && xbar <= M3_MAX_LINE_CARD + M3_MAX_BACKPLANE_XBAR);
  

  for (int i = 0; i < 16; i++)
  {
    int ii = i + correction;
    
    if (ii < 0 || ii >= 16)
      continue;

    for (int j = 0; j < 16; j++)
    {
      int jj = j + correction;

      if (jj < 0 || jj >= 16)
	continue;

      insist (xbars [xbar - 1].pingSquares [ii][jj]);
      gnome_canvas_item_set (xbars [xbar - 1].pingSquares [ii][jj],  "fill_color", !counter->sent [i][j] ? MU_CLUSTER_OBJECT_COLOR : (counter->sent [i][j] == counter->received [i][j] ? "green" : (counter->received [i][j] ? "yellow" : "red")), NULL);
    }
  }

  int numReceived, numSent;
  
  for (int i = 0; i < 16; i++)
  {
    numSent = numReceived = 0;
    int ii = i + correction;
    if (ii < 0 || ii >= 16)
      continue;
    
    for (int j = 0; j < 16; j++)
    {
      numSent += counter->sent [i][j] + counter->sent [j][i];
      numReceived += counter->received [i][j] + counter->received [j][i];
    }
    {
      insist (xbars [xbar - 1].portSquares [ii]);
      gnome_canvas_item_set (xbars [xbar - 1].portSquares [ii],  "fill_color", !numSent ? MU_CLUSTER_OBJECT_COLOR : (numSent == numReceived ? "green" : (numReceived ? "yellow" : "red")), NULL);
    }
  }
  return 1;
  exception: return 0;
}


int mu_Host::eatCounter (pi_Counter*counter)
{
  insist (this && counter);
  gnome_canvas_item_set (box,  "fill_color", counter->sent [0] [0]  == 0 ? MU_CLUSTER_OBJECT_COLOR : (counter->sent [0] [0] == counter->received [0] [0]? "green" : (counter->received [0] [0]? "yellow" : "red")), NULL);
  return 1;
  exception: return 0;
}

int mu_Cluster::eatEvent (char ipAddress[4], int type, mu_int64 count, int partType, int index, int state)
{
  insist (this);
  m3_DescriptionXbar*xbar;
  m3_DescriptionLinecard*linecard;
  int xbarIndex;
  int portIndex;
  
  mu_Switch*s;
  s = getSwitchFromIpAddress (ipAddress);
  insist (s);

  if (state)
  {
    if (lastEventIndex && !eatEvent (lastEventIpAddress, lastEventType, 0, lastEventPartType, lastEventIndex, 0))
      return 0;
  
    lastEventIndex = index;
    lastEventType = type;
    lastEventPartType = partType;
    memcpy (lastEventIpAddress, ipAddress, 4);
  }
  switch (type)
  {
    case M3_ucResetTrap:
    case M3_ucFirmwareFaultTrap:
    case M3_ucOvertemperatureShutdownFaultTrap:
    case M3_ucDeadTrap:
    case M3_slotLostCommunicationTrap:
    case M3_slotReestablishCommunicationTrap:
    case M3_slotChangedTrap:
      insist (partType == M3_SLOT_TABLE);
      
      if ((linecard = s->getLinecardFromSlot (index)))
	s->show (linecard->index, state);
      
      if ((xbar = s->getXbarFromSlot (index)))
	s->xbars [xbar->index - 1].show (state);
      
      return 1;
      
    case M3_fanBadTrap:
    case M3_fanTrayMissingTrap:
    case M3_fanTrayReinstalledTrap:
    case M3_acBadTrap:
    case M3_acGoodTrap:
    case M3_dcBadTrap:
    case M3_dcGoodTrap:
    case M3_bigUcTrap:
      return s->show (state);
      
    case M3_badCrcTrap:
    case M3_invalidRouteTrap:
    case M3_illegalSymbolTrap:
    case M3_missedBeatTrap:
    case M3_deadPortTrap:
    case M3_portDownOnIllegalTrap:
    case M3_portDownOnMissedBeatTrap:
    case M3_portUpTrap:
    case M3_receiveTimeoutTrap:
    case M3_transmitTimeoutTrap:
    case M3_portSelftestTrap:

      xbarIndex = (index - 1) / 16;
      portIndex = (index - 1) % 16;

      if (xbarIndex >= 0 && xbarIndex < M3_MAX_BACKPLANE_XBAR + M3_MAX_LINE_CARD)
      {
	s->xbars [xbarIndex].show (portIndex, state);
	if (s->xbars [xbarIndex].ports [portIndex].link)
	  s->xbars [xbarIndex].ports [portIndex].link->show (state);
      }
      
      return 1;
      
    case M3_sanFaultTrap:
    case M3_serdesFaultTrap:
    case M3_notInSyncTrap:
    case M3_overrunTrap:
    case M3_signalLostTrap:
    case M3_portDownForLanPortTrap:
    case M3_noCableTrap:
      
      if ((linecard = s->getLinecardFromPart (partType, index, &portIndex)))
      {
	insist (portIndex >= 0 && portIndex < M3_MAX_LINE_CARD_PORT);
	insist (linecard->index > 0 && linecard->index < M3_MAX_LINE_CARD);
	s->show (linecard->index - 1, portIndex, state);
	mu_Link*link;
	if ((link = s->linecardPorts [linecard->index - 1] [portIndex].getLink ()))
	  link->show (state);
      }
    default:;
  }
  
  return 1;
  exception: return 0;
}

int mu_Switch::show (int state)
{
  insist (this);
  insist (box);
  if (state)
    gnome_canvas_item_set (box, "outline_color", "red", "width_pixels", 3,  NULL);
  else
    gnome_canvas_item_set (box, "outline_color", "black", "width_pixels", 1,  NULL);
  return 1;
  exception: return 0;
}

int mu_Switch::show (int linecard, int state)
{
  insist (this);
  insist (linecard >= 0 && linecard < backplaneType);
  insist (linecardItems [linecard]);
  
  if (state)
    gnome_canvas_item_set (linecardItems [linecard], "outline_color", "red", "width_pixels", 3,  NULL);
  else
    gnome_canvas_item_set (linecardItems [linecard], "outline_color", "black", "width_pixels", 1,  NULL);
  return 1;
  exception: return 0;
}

int mu_Switch::show (int linecard, int port, int state)
{
  insist (this);
  insist (linecard >= 0 && linecard < backplaneType);
  insist (port >= 0 && port < 16);
  insist (linecardPorts [linecard] [port].portItem);
  
  if (state)
    gnome_canvas_item_set (linecardPorts [linecard] [port].portItem, "fill_color", "red", NULL);
  else
    gnome_canvas_item_set (linecardPorts [linecard] [port].portItem, "fill_color", MU_CLUSTER_LINECARD_PORT_COLOR, NULL);
  return 1;
  exception: return 0;
}

int mu_Xbar::show (int state)
{
  insist (this);
  insist (box);

  if (state)
    gnome_canvas_item_set (box, "outline_color", "red", "width_pixels", 3,  NULL);
  else
    gnome_canvas_item_set (box, "outline_color", "black", "width_pixels", 1,  NULL);
  
  return 1;
  exception: return 0;
}

int mu_Xbar::show (int port, int state)
{
  insist (this);
  insist (port >= 0 && port < 16);
  insist (portSquares [port]);
  
  if (state)
    gnome_canvas_item_set (portSquares [port], "fill_color", "red",  NULL);
  else
    gnome_canvas_item_set (portSquares [port], "fill_color", MU_CLUSTER_OBJECT_COLOR,  NULL);
  
  return 1;
  exception: return 0;
}

int mu_Link::show (int state)
{
  insist (this && item);
	  
  if (state)
    gnome_canvas_item_set (item, "fill_color", "red",  "width_pixels", 3, NULL);
  else
    gnome_canvas_item_set (item, "fill_color", "black",  "width_pixels", 1, NULL);
  return 1;
  exception: return 0;
}

mu_Link*mu_LinecardPort::getLink ()
{
  insist (this && parent);

  if (port)
  {    
    int p;
    m3_DescriptionXbar*xbar = (m3_DescriptionXbar*) port->parent;
    insist (xbar);
	
    p = ((port->index - 1) % 16);
    insist (p >= 0 && p < 16);
    return parent->xbars [xbar->index - 1].ports[p].link;
  }
  exception: return 0;
}

int mu_Cluster::eatPacketCounters (int ipAddress, int good, int numCounts, mu_int64*counters, mu_int64*totalTotal, mu_int64*total, mu_int64*maximum, mu_int64*minimum)
{
  insist (this && ipAddress && counters && totalTotal && total && maximum && minimum);
  mu_Switch*s;
  int r;

  s = getSwitchFromIpAddress ((char*)&ipAddress);

  pthread_mutex_lock (&mutex);
  r = s && s->eatPacketCounters (good, numCounts, counters, totalTotal, total, maximum, minimum);
  pthread_mutex_unlock (&mutex);

  return r;

  exception: return 0;
}


int mu_Switch::eatPacketCounters (int good, int numCounts, mu_int64*counters, mu_int64*totalTotal, mu_int64*total, mu_int64*maximum, mu_int64*minimum)
{
  int c = 0;
  mu_int64 last;
  
  insist (this && counters && totalTotal && total && maximum && minimum);

  for (int i = 0; i < M3_MAX_LINE_CARD + M3_MAX_BACKPLANE_XBAR; i++)
  {
    for (int j = 0; c < numCounts && j < 16; j++, c++)
    {
      if (good)
      {
	last = xbars [i].ports [j].good [0] = xbars [i].ports [j].good [1];
	xbars [i].ports [j].good [1] = counters [c];
      }
      else
      {
	last = xbars [i].ports [j].bad [0] = xbars [i].ports [j].bad [1];
	xbars [i].ports [j].bad [1] = counters [c];
      }
      *totalTotal += counters [c];
	
      if (counters [c] > last)
      {
	*total += counters [c] - last;
	if (counters [c] - last > *maximum)
	  *maximum = counters [c] - last;
	if (!*minimum || counters [c] - last < *minimum)
	  *minimum = counters [c] - last;
      }
    }
  }
  return 1;
  exception: return 0;
}

int mu_Cluster::drawPacketCounters (int good, int maximum, int minimum)
{
  insist (this);

  int minGood, maxGood, minBad, maxBad;
  
  if (!getRanges (&minGood, &maxGood, &minBad, &maxBad))
    return 0;
  
  for (int i = 0; i < numSwitches; i++)
    switches [i]->drawPacketCounters (good, good ? maxGood : maxBad, good ? minGood : minBad);
  
  return 1;
  exception: return 0;

}


int mu_Cluster::updateLegend (char*caption, double maximum, double minimum, int precision)
{
  double values [MU_CLUSTER_NUM_COLORS];
  double step = (maximum - minimum) / (double) MU_CLUSTER_NUM_COLORS;
  
  insist (mu_mainWidget && mu_mainWidget->legendWidget);
  
  for (int i = 0; i < MU_CLUSTER_NUM_COLORS; i++)
    values [i] = minimum + step * (double) (i + 1);

  mu_mainWidget->legendWidget->set (caption, values, precision);
  
  return 1;
  exception: return 0;
}

int mu_Switch::drawPacketCounters (int good, int maximum, int minimum)
{
  insist (this && cluster);
  
  cluster->updateLegend ((char*) (good ? "Good Packets / sec" : "Bad Packets / sec"), maximum, minimum, 0);
  
  for (int i = 0; i < M3_MAX_LINE_CARD + M3_MAX_BACKPLANE_XBAR; i++)
  {
    for (int j = 0; j < 16; j++)
    {
      mu_Link*link = xbars [i].ports [j].link;
      
      if (link && link->port == &xbars [i].ports [j])
      {
	int v = good ? 
	  (link->port->good [1] - link->port->good [0] + (link->other ? link->other->good [1] - link->other->good [0] : 0)) :
	  (link->port->bad [1] - link->port->bad [0] + (link->other ? link->other->bad [1] - link->other->bad [0] : 0));
	
	if (v <= 0 || maximum <= 0)
	{	  
	  gnome_canvas_item_set (xbars [i].ports [j].link->item, "fill_color", "black",  "width_pixels", 1, NULL);
	}
	else
	{
	  int r = (int) (maximum != minimum ? ((MU_CLUSTER_NUM_COLORS - 1) * ((double) v - (double) minimum) / (double) ((maximum - minimum))) : MU_CLUSTER_NUM_COLORS);
	  
	  if (r >= MU_CLUSTER_NUM_COLORS)
	    r = MU_CLUSTER_NUM_COLORS - 1;
	  if (r < 0.0)
	    r = 0;
	  
	  insist (r >= 0 && r < MU_CLUSTER_NUM_COLORS);

	  gnome_canvas_item_set (xbars [i].ports [j].link->item, "fill_color_rgba", mu_Cluster_colors [r],  "width_pixels", 1 + r * MU_CLUSTER_MAX_LINK_WIDTH / MU_CLUSTER_NUM_COLORS, NULL);
	}
      }
    }
  }
  return 1;
  exception: return 0;
}

int mu_Cluster::getRanges (int*minGood, int*maxGood, int*minBad, int*maxBad)
{
  insist (this);
  insist (minGood && maxGood && minBad && maxBad);
  *minGood = *maxGood = *minBad = *maxBad = 0;
  
  for (int k = 0; k < numSwitches; k++)
  {    
    for (int i = 0; i < M3_MAX_LINE_CARD + M3_MAX_BACKPLANE_XBAR; i++)
    { 
      for (int j = 0; j < 16; j++)
      {
	if (!switches [k]->xbars [i].ports [j].link)
	  continue;

	int d = switches [k]->xbars [i].ports [j].good [1] - switches [k]->xbars [i].ports [j].good [0];
	
	if (*minGood == 0 || *minGood > d)
	  *minGood = d;
	if (*maxGood == 0 || *maxGood < d)
	  *maxGood = d;
      
	d = switches [k]->xbars [i].ports [j].bad [1] - switches [k]->xbars [i].ports [j].bad [0];
 
	if (*minBad == 0 || *minBad > d)
	  *minBad = d;
	if (*maxBad == 0 || *maxBad < d)
	  *maxBad = d;
      }
    }
  }
  return 1;
  exception: return 0;
}

int mu_Xbar::callback (GnomeCanvasItem*item, GdkEvent*event)
{
  insist (item && event);
  char s1 [M3_DISPLAY_STRING_LENGTH];
  char s2 [M3_DISPLAY_STRING_LENGTH];
  
  insist (this && item && event);  
  insist (mu_mainWidget && mu_mainWidget->webWidget && mu_mainWidget->changeWidget);

  if (event->type == GDK_BUTTON_PRESS && event->button.button == 1)
  {
    insist (parent);
    int index;
    index = this - parent->xbars + 1;

    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (parent->cluster->questionToggle)))
    {
      sprintf (s1, "%d.%d.%d.%d",
	       (unsigned) (((char*)(&parent->ipAddress)) [0] & 0xff),
	       (unsigned) (((char*)(&parent->ipAddress)) [1] & 0xff),
	       (unsigned) (((char*)(&parent->ipAddress)) [2] & 0xff),
	       (unsigned) (((char*)(&parent->ipAddress)) [3] & 0xff));
      
      sprintf (s2, "/xbar16%d", index);
      mu_mainWidget->webWidget->get (s1, 1, s2);
      return 1;
    }
    else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (parent->cluster->changeToggle)))
    {
      mu_mainWidget->changeWidget->set (parent->ipAddress, M3_XBAR16_TABLE, index);
      return 1;
    }
  }
  exception: return 0;
}

m3_DescriptionPort*mu_LinecardPort::getPort ()
{
  int lineCardIndex, portIndex;
  
  insist (this);
  insist (parent);
  
  portIndex =  (this - &parent->linecardPorts[0][0]) % M3_MAX_LINE_CARD_PORT;
  lineCardIndex = (this - &parent->linecardPorts [0][0]) / M3_MAX_LINE_CARD_PORT;
  insist (lineCardIndex >= 0 && lineCardIndex < M3_MAX_LINE_CARD);
  
  return &parent->linecards [lineCardIndex].ports [portIndex];
  
  exception: return 0;
}

int mu_LinecardPort::callback (GnomeCanvasItem*item, GdkEvent*event)
{
  m3_DescriptionPort*port;

  insist (item && event);
  char s1 [M3_DISPLAY_STRING_LENGTH];
  char s2 [M3_DISPLAY_STRING_LENGTH];
  char s3 [M3_DISPLAY_STRING_LENGTH];
  char s4 [M3_DISPLAY_STRING_LENGTH];
  
  insist (this && item && event); 
  insist (mu_mainWidget && mu_mainWidget->webWidget && mu_mainWidget->changeWidget);
  
  port = getPort ();
  
  insist (port);
  insist (port->type > 0 && port->type <= M3_NUM_TABLES);
  

  if (event->type == GDK_BUTTON_PRESS && event->button.button == 1)
  {
    insist (parent);
    
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (parent->cluster->questionToggle)))
    {
      strcpy (s2, m3_table_names [port->type - 1]);
      *s2 = tolower (*s2);
      
      sprintf (s1, "%d.%d.%d.%d",
	       (unsigned) (((char*)(&parent->ipAddress)) [0] & 0xff),
	       (unsigned) (((char*)(&parent->ipAddress)) [1] & 0xff),
	       (unsigned) (((char*)(&parent->ipAddress)) [2] & 0xff),
	       (unsigned) (((char*)(&parent->ipAddress)) [3] & 0xff));
      
      sprintf (s3, "/%s%d", s2, port->index);
    
      if (port->port)
      {
	sprintf (s4, "/xbarPort%d", port->port->index);
	mu_mainWidget->webWidget->get (s1, 2, s3, s4);
      }
      else
	mu_mainWidget->webWidget->get (s1, 1, s3);

      return 1;
    }
    else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (parent->cluster->changeToggle)))
    {
      mu_mainWidget->changeWidget->set (parent->ipAddress, port->type, port->index);
      return 1;
    }
    
  }
  exception: return 0;
}

int mu_Xbar::portCallback (GnomeCanvasItem*item, GdkEvent*event)
{
  insist (item && event);
  char s1 [M3_DISPLAY_STRING_LENGTH];
  char s2 [M3_DISPLAY_STRING_LENGTH];
  
  insist (this && item && event);  
  insist (mu_mainWidget && mu_mainWidget->webWidget && mu_mainWidget->changeWidget);

  if (event->type == GDK_BUTTON_PRESS && event->button.button == 1)
  {
    insist (parent);
    int p, index;
    for (p = 0; p < 16 && portSquares [p] != item; p++);
    insist (p >= 0 && p < 16);
    index = 1 + (this - parent->xbars) * 16 + p;
    
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (parent->cluster->questionToggle)))
    {
      sprintf (s1, "%d.%d.%d.%d",
	       (unsigned) (((char*)(&parent->ipAddress)) [0] & 0xff),
	       (unsigned) (((char*)(&parent->ipAddress)) [1] & 0xff),
	       (unsigned) (((char*)(&parent->ipAddress)) [2] & 0xff),
	       (unsigned) (((char*)(&parent->ipAddress)) [3] & 0xff));
      
      sprintf (s2, "/xbarPort%d", index);
      
      mu_mainWidget->webWidget->get (s1, 1, s2);
      return 1;
    }
    else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (parent->cluster->changeToggle)))
    {
      mu_mainWidget->changeWidget->set (parent->ipAddress, M3_XBAR_PORT_TABLE, index);
      return 1;
    }
  }
  exception: return 0;
}

int mu_Cluster::followRoute (mu_Host*from, mu_Host*to, int draw, int count)
{
  mt_Route*r;
  mu_Link*link;
  mu_Port*port;
  mu_Switch*s;
  char*hops;
  int length;
  
  insist (this && from && to && from != to);
  insist (mu_mainWidget && mu_mainWidget->routesWidget);

  r = mu_mainWidget->routesWidget->getRoute (from->host->getTypeIndex (), to->host->getTypeIndex ());
  insist (r);
  
  hops = r->getHops ();
  insist (hops);
  
  length = r->getLength ();
  insist (length >= 0);

  port = from;
  
  for (int i = 0; i <= length; i++)
  {
    insist (port);
    link = port->link;
    insist (link);

    link->routeCount += count;
    insist (link->routeCount >= 0);

    if (draw)
    {
      if (link->routeCount)
	gnome_canvas_item_set (link->item, "fill_color", "blue",  "width_pixels", 3, NULL);
      else
	gnome_canvas_item_set (link->item, "fill_color", "black",  "width_pixels", 1, NULL);
    }
    
    port = link->port == port ? link->other : link->port;

    if (i < length)
      port += hops [i]; //kludge
  }
  return 1;
  exception: return 0;
}


int mu_Cluster::clearLinkColors ()
{
  insist (this);
  mu_Link*link;
  
  for (link = (mu_Link*) links; link; link = (mu_Link*) link->next)
  {
    gnome_canvas_item_set (link->item, "fill_color", "black",  "width_pixels", 1, NULL);
    link->routeCount = 0;
  }
  return 1;
  exception: return 0;
}

int mu_Cluster::drawRoutes ()
{
  insist (this);
  mu_Host*h1, *h2;
  mu_Link*link;
  int smallest, biggest;
  
  clearLinkColors ();
  
  for (h1 = (mu_Host*) hosts; h1; h1 = (mu_Host*) h1->next)
    for (h2 = (mu_Host*) hosts; h2; h2 = (mu_Host*) h2->next)
      if (h1 != h2 && !followRoute (h1, h2, 0, 1))
	return 0;
  
  smallest = biggest = -1;
  
  for (link = (mu_Link*) links; link; link = (mu_Link*) link->next)
  {
    if (link->routeCount > biggest || biggest < 0)
      biggest = link->routeCount;
    if (link->routeCount < smallest || smallest < 0)
      smallest = link->routeCount;
  }

  updateLegend ((char*) ("Route Counts"), biggest, smallest, 0);

  for (link = (mu_Link*) links; link; link = (mu_Link*) link->next)
  {
    if (link->routeCount)
    {
      int r = (int) (biggest != smallest ? ((MU_CLUSTER_NUM_COLORS - 1) * ((double) link->routeCount - (double) smallest) / (double) ((biggest - smallest))) : MU_CLUSTER_NUM_COLORS);
    
      if (r >= MU_CLUSTER_NUM_COLORS)
	r = MU_CLUSTER_NUM_COLORS - 1;
      if (r < 0)
	r = 0;
      
      insist (r >= 0 && r < MU_CLUSTER_NUM_COLORS);
    
      gnome_canvas_item_set (link->item, "fill_color_rgba", mu_Cluster_colors [r],  "width_pixels", 1 + r * MU_CLUSTER_MAX_LINK_WIDTH / MU_CLUSTER_NUM_COLORS, NULL);
    }
  } 
  exception: return 0;
}


int mu_Cluster::showCrcs (int good)
{
  insist (this);
  mu_Host*h1, *h2;
  mu_Link*link;
  unsigned long long smallest, biggest;  
  smallest = biggest = 0;
  
  clearLinkColors ();

  for (link = (mu_Link*) links; link; link = (mu_Link*) link->next)
  {
    unsigned long long c;
    insist (link->port && link->port->port);
    insist (link->other);
    
    c = good ?
      link->port->port->goodCrcs [1] + (link->other->port ? link->other->port->goodCrcs [1] : 0) :
      link->port->port->badCrcs [1] + (link->other->port ? link->other->port->badCrcs [1] : 0);

    if (c > biggest || biggest == 0)
      biggest = c;
    if (c < smallest || smallest == 0)
      smallest = c;
  }

  updateLegend ((char*) (good ? "Good Packets" : "Bad Packets"), biggest, smallest, 0);

  for (link = (mu_Link*) links; link; link = (mu_Link*) link->next)
  {
    unsigned long long c;
    insist (link->port && link->port->port);
    insist (link->other);

    c = good ?
      link->port->port->goodCrcs [1] + (link->other->port ? link->other->port->goodCrcs [1] : 0) :
      link->port->port->badCrcs [1] + (link->other->port ? link->other->port->badCrcs [1] : 0);

    if (c)
    {
      int r = (int) (biggest != smallest ? ((MU_CLUSTER_NUM_COLORS - 1) * (double) (c - smallest )/ (double) ((biggest - smallest))) : MU_CLUSTER_NUM_COLORS);
    
      if (r >= MU_CLUSTER_NUM_COLORS)
	r = MU_CLUSTER_NUM_COLORS - 1;
      if (r < 0)
	r = 0;
      insist (r >= 0 && r < MU_CLUSTER_NUM_COLORS);
    
      gnome_canvas_item_set (link->item, "fill_color_rgba", mu_Cluster_colors [r],  "width_pixels", 1 + r * MU_CLUSTER_MAX_LINK_WIDTH / MU_CLUSTER_NUM_COLORS, NULL);
    }
  } 
  exception: return 0;
}

class mu_ClusterLevel
{
  public:
  int numSwitches;
  int totalSwitches;
  mu_Switch**switches;
};


int mu_Cluster::tree ()
{
  int numLevels;
  mu_ClusterLevel*levels;
  mu_Switch*s;
  mt_Queue q;
  
  insist (this && numSwitches && switches);

  setWatchCursor (1);

  for (int i = 0; i < numSwitches; i++)
    switches [i]->depth = -1;

  if (!selected || selected->isHost ())
    s = switches [0];
  else
    s = (mu_Switch*) selected;
  insist (s);
  
  numLevels = 0;
  q.put (s);
  s->depth = 0;
  
  while ((s = (mu_Switch*) q.get ()))
  {
    insist (s->depth >= 0);

    if (s->depth + 1 > numLevels)
      numLevels = s->depth + 1;
    
    for (int i = 0; i < M3_MAX_BACKPLANE_XBAR + M3_MAX_LINE_CARD; i++)
    {
      for (int j = 0; j < 16; j++)
      {
	if (s->xbars [i].ports [j].link)
	{
	  mu_Switch*s2;
	  mu_Port*p =  s->xbars [i].ports [j].link->port == &s->xbars [i].ports [j] ? 
	    s->xbars [i].ports [j].link->other : s->xbars [i].ports [j].link->port;
	  insist (p);
	  
	  if ((s2 = p->getSwitch ()) && s2->depth < 0)
	  {
	    s2->depth = s->depth + 1;
	    q.put (s2);
	  }
	}
      }      
    }
  }
  
  insist (numLevels > 0);
  levels = (mu_ClusterLevel*) calloc (sizeof (mu_ClusterLevel), numLevels);
  insist (levels);
  
  for (int i = 0; i < numSwitches; i++)
  {
    s = switches [i];
    insist (s);
    insist (s->depth >= 0 && s->depth < numLevels);
    levels [s->depth].totalSwitches++;
  }
  
  for (int i = 0; i < numLevels; i++)
  {
    insist (levels [i].totalSwitches);
    levels [i].switches = (mu_Switch**) calloc (sizeof (mu_Switch*), levels [i].totalSwitches);
    insist (levels [i].switches);
  }

  for (int i = 0; i < numSwitches; i++)
  {
    s = switches [i];
    insist (s);
    insist (s->depth >= 0 && s->depth < numLevels);
    insist (levels [s->depth].numSwitches < levels [s->depth].totalSwitches);
    levels [s->depth].switches [levels [s->depth].numSwitches++] = s;
  }

  double ax, ay, bx, by;
  
  ax = (double) GTK_WIDGET (canvas)->allocation.width / canvas->pixels_per_unit - 3 * MU_CLUSTER_SIZE;
  ay = (double) GTK_WIDGET (canvas)->allocation.height / canvas->pixels_per_unit - 3 * MU_CLUSTER_SIZE;
  
  bx = hadjustment->value / canvas->pixels_per_unit - MU_CLUSTER_SIZE;
  by = vadjustment->value / canvas->pixels_per_unit - MU_CLUSTER_SIZE;
  
  if (ax > 0) bx -= ax / 2;
  if (ay > 0) by -= ay / 2;

  double h, w;
  h = ((double) GTK_WIDGET (canvas)->allocation.height / canvas->pixels_per_unit - (MU_CLUSTER_PORT_TOTAL_HEIGHT * 17))/ (double) (numLevels - 1);

  int level;
  level = 0;
  
  for (int i = 0; i < numLevels; i++)
  {
    insist (levels [i].numSwitches);
    double x, y;
    y = by + h * (numLevels - 1 - i);
    level++;
    
    w = (double) GTK_WIDGET (canvas)->allocation.width / canvas->pixels_per_unit / (double) levels [i].numSwitches;
    
    for (int j = 0; j < levels [i].numSwitches; j++)
    {
      x =  bx + j * w;
      levels [i].switches [j]->position (x, y);
    }
  }
  
  toggled ();
  
  for (int i = 0; i < numLevels; i++)
    free (levels [i].switches);
  free (levels);

  return 1;
  exception: return 0;
}

int mu_Movable::position (double x, double y)
{
  double x1, y1, x2, y2;
  gnome_canvas_item_get_bounds (box, &x1, &y1, &x2, &y2);
  gnome_canvas_item_i2w (box, &x1, &y1);

  return move (x - x1, y - y1);
}

int mu_Movable::isHost ()
{
  return 0;
}

int mu_Host::isHost ()
{
  return 1;
}


int mu_Cluster::select (char*name)
{
  insist (this && name);
  mu_Movable*m;

  selectAll (0);

  return (m = find (name)) && select (m);
  exception: return 0;
}

mu_Movable*mu_Cluster::find (char*name)
{
  insist (this && name);
  char lower1 [M3_DISPLAY_STRING_LENGTH];
  char lower2 [M3_DISPLAY_STRING_LENGTH];
  lower1 [M3_DISPLAY_STRING_LENGTH - 1] = 0;
  lower2 [M3_DISPLAY_STRING_LENGTH - 1] = 0;
  
  strncpy (lower1, name, M3_DISPLAY_STRING_LENGTH - 1);
  for (char*s = lower1; *s; s++)
    *s = tolower (*s);
  
  for (mu_Host*h = (mu_Host*) hosts; h; h = (mu_Host*) h->next)
  {
    mt_Address*a;
    a = h->host->getAddress ();
    insist (a);

    insist (h->host && h->host->getName ());
    strncpy (lower2, h->host->getName (), M3_DISPLAY_STRING_LENGTH - 1);
    for (char*s = lower2; *s; s++)
      *s = tolower (*s);
    
    if (!strcmp (lower1, lower2) || !strcmp (lower1, a->toString ()))
      return h;
  }
    
  for (int i = 0; i < numSwitches; i++)
  {
    sprintf (lower2, "%d.%d.%d.%d",
	     (unsigned) (((char*)(&switches [i]->ipAddress)) [0] & 0xff),
	     (unsigned) (((char*)(&switches [i]->ipAddress)) [1] & 0xff),
	     (unsigned) (((char*)(&switches [i]->ipAddress)) [2] & 0xff),
	     (unsigned) (((char*)(&switches [i]->ipAddress)) [3] & 0xff));

    if (!strcmp (lower1, lower2))
      return switches [i];
      
    sprintf (lower2, "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
	     switches [i]->ethernetAddress [0] & 0xff,
	     switches [i]->ethernetAddress [1] & 0xff,
	     switches [i]->ethernetAddress [2] & 0xff, 
	     switches [i]->ethernetAddress [3] & 0xff,
	     switches [i]->ethernetAddress [4] & 0xff,
	     switches [i]->ethernetAddress [5] & 0xff);

    if (!strcmp (lower1, lower2))
      return switches [i];
  }
  exception: return 0;
}

int mu_Cluster::refresh ()
{
  insist (this);
  
  setWatchCursor (1);

  int numHelpers;
  numHelpers = mu_maxThreads;
  if (numSwitches < numHelpers)
    numHelpers = numSwitches;

  {
    mu_SwitchHelper helpers [numHelpers];
    
    for (int i = 0; i < numSwitches; i++)
    {
      helpers [i % numHelpers].join ();
      helpers [i % numHelpers]._switch = switches [i];
      helpers [i % numHelpers].filename [0] = 0;
      helpers [i % numHelpers].address = switches [i]->ipAddress;

      mt_Component::printFormat ("examining %d.%d.%d.%d",
				 (unsigned) (((char*)(&switches [i]->ipAddress)) [0] & 0xff),
				 (unsigned) (((char*)(&switches [i]->ipAddress)) [1] & 0xff),
				 (unsigned) (((char*)(&switches [i]->ipAddress)) [2] & 0xff),
				 (unsigned) (((char*)(&switches [i]->ipAddress)) [3] & 0xff));

      helpers [i % numHelpers].createThread ((mu_ThreadableFunction) &mu_SwitchHelper::run, 0);
    }

    for (int i = 0; i < numHelpers; i++)
      helpers [i].join ();
  }

  toggled ();
  
  exception: return 0;
}

int mu_Cluster::showTemperatures (int state)
{
  insist (this);
  int a, b;
  
  a = b = 0;
  
  return state ? showTemperatures (&a, &b, COMPUTE) && showTemperatures (&a, &b, COLOR) : showTemperatures (0, 0, CLEAR);

  exception: return 0;
}

int mu_Cluster::showTemperatures (int*smallest, int*biggest, int mode)
{
  insist (this);
  insist (mode == CLEAR || biggest && smallest);
  
  if (mode == COLOR)
  {
    double c1, c2;
    c1 = ((double)*biggest - 32.0) * 5.0 / 9.0;
    c2 = ((double)*smallest - 32.0) * 5.0 / 9.0;
    
    updateLegend ("Temperature (C)", c1, c2, 1);
  }
  
  
  for (int i = 0; i < numSwitches; i++)
  {
    int t = 0;
 
    insist (switches [i]);
  
    if (mode != CLEAR)
    {    
      insist (biggest && smallest);
      
      for (int j = 0; j < M3_DESCRIPTION_NUM_LINECARD_TEMPERATURES; j++)
	if (switches [i]->bigUcTemperatures [j] > t)
	  t = switches [i]->bigUcTemperatures [j];

      for (int j = 0; j < switches[i]->numBackplaneTemperatures; j++)
	if (switches [i]->backplaneTemperatures [j] > t)
	  t = switches [i]->backplaneTemperatures [j];

      if (switches [i]->fanCardTemperature > t)
	t = switches [i]->fanCardTemperature;
    }
    
    showTemperature (switches [i]->linecardItems [0], t, smallest, biggest, mode);
    
    for (int j = 0; j < M3_MAX_LINE_CARD; j++)
    {
      if (switches [i]->linecards [j].type)
      {
	t = 0;
	for (int k = 0; k < M3_DESCRIPTION_NUM_LINECARD_TEMPERATURES; k++)
	  if (switches [i]->linecards [j].temperatures [k] > t)
	    t = switches [i]->linecards [j].temperatures [k];

	showTemperature (switches [i]->linecardItems [j + 1], t, smallest, biggest, mode);
      }
    }
  }
  
  return 1;
  exception: return 0;
  
}

int mu_Cluster::showTemperature (GnomeCanvasItem*item, int temperature, int*smallest, int*biggest, int mode)
{
  insist (this && item);
  
  if (mode == COMPUTE)
  {
    insist (smallest && biggest);
    
    if (!*smallest || temperature < *smallest)
      *smallest = temperature;	
    if (!*biggest || temperature > *biggest)
      *biggest = temperature;
  }
  else if (mode == CLEAR)
  {
    gnome_canvas_item_set (item, "fill_color", MU_CLUSTER_OBJECT_COLOR, NULL);
  }
  else
  {
    insist (smallest && biggest);
    insist (*smallest <= *biggest);
    
    int r;
    r  = (int) (*biggest != *smallest ? ((MU_CLUSTER_NUM_COLORS - 1) * ((double)temperature - (double) *smallest) / (double) ((*biggest - *smallest))) : MU_CLUSTER_NUM_COLORS);

    if (r >= MU_CLUSTER_NUM_COLORS)
      r = MU_CLUSTER_NUM_COLORS - 1;
    if (r < 0)
      r = 0;
    insist (r >= 0 && r < MU_CLUSTER_NUM_COLORS);

    gnome_canvas_item_set (item, "fill_color_rgba", mu_Cluster_colors [r], NULL);
  }
  return 1;
  exception: return 0;
}

int mu_Cluster::clearCounters ()
{
  m3_TestClient client;
  int r;
  char*s;
  
  setWatchCursor (1);
  
  for (int i = 0; i < numSwitches; i++)
  {
    if (!(s = web.post (switches [i]->ipAddress, "/change_uc1", 2, "reboot", "on", "change", "all")))
    {
      toggled ();
      return 0;
    }

    while (s = web.token ())
    {
      if (!strcmp (s, "no"))
      {
	if ((s = web.token ()) && !strcmp (s, "problem"))
	  break;
      }
    }
    
    if (!s || !client.open (switches [i]->ipAddress) || !client.reboot ())
    {
      toggled ();
      return 0;
    }
  }
  toggled ();
  return 1;
  exception: return 0;
}

int mu_Cluster::stopSnmpTraps ()
{
  m3_TestClient client;
  int r;
  char*s;
  
  setWatchCursor (1);
  
  for (int i = 0; i < numSwitches; i++)
  {
    if (!(s = web.post (switches [i]->ipAddress, "/change_bigUc1", 2, "hostTrapCount", "10000000", "trapTimeout", "100000")))
      return 0;

    while (s = web.token ())
      if (!strcmp (s, "no") && (s = web.token ()) && !strcmp (s, "problem"))
	break;

    if (!(s = web.post (switches [i]->ipAddress, "/change_trapMask1", 1, "blah", "blah")))
      return 0;

    while (s = web.token ())
      if (!strcmp (s, "no") && (s = web.token ()) && !strcmp (s, "problem"))
	break;

  }
  toggled ();
  return 1;
  exception: return 0;
}

mu_SwitchHelper::mu_SwitchHelper ()
{
  *filename = 0;
  _switch = 0;
  address = 0;
}


void mu_SwitchHelper::run (void*)
{
  insist (this);
  insist (address || *filename);
  insist (_switch);
  
  if (address)
  {
    int tries = 0;
    do
    {
      if (tries)
      {
	mt_Component::printFormat ("retrying %d.%d.%d.%d",
				   (unsigned) (((char*)(&address)) [0] & 0xff),
				   (unsigned) (((char*)(&address)) [1] & 0xff),
				   (unsigned) (((char*)(&address)) [2] & 0xff),
				   (unsigned) (((char*)(&address)) [3] & 0xff));
	sleep (5);
      }
      
      _switch->examine (&web, address);
    }
    while (tries++ < 3 && _switch->backplaneType != 2 && _switch->numBackplaneXbars != _switch->backplaneType / 2);
  }
  else
    _switch->examineFile (&web, filename);

  exception:;
}

int mu_Cluster::fromFile (FILE*fp)
{
  insist (this);
  insist (fp);

  {
    mt_Tokenizer tokens (fp, "(), \n\r\t", M3_DISPLAY_STRING_LENGTH, ";#");
    char*s;
    int width, height;
    double pixels_per_unit, h, v;
    mu_Movable*m;
    
    setWatchCursor (1);
    
    if (!(s = tokens.getNext ()) || strcmp (s, "height") ||
	!(s = tokens.getNext ()) || !tokens.isInteger () ||
	(!(height = tokens.getInteger ()) || height < 0))
      goto error;
    

    if (!(s = tokens.getNext ()) || strcmp (s, "width") ||
	!(s = tokens.getNext ()) || !tokens.isInteger () ||
	(!(width = tokens.getInteger ()) || width < 0))
      goto error;
    
    if (!(s = tokens.getNext ()) || strcmp (s, "pixels_per_unit") ||
	!(s = tokens.getNext ()) || sscanf (s, "%lf", &pixels_per_unit) != 1 ||
	pixels_per_unit <= 0.0)
      goto error;
    
    if (!(s = tokens.getNext ()) || strcmp (s, "hadjustment") ||
	!(s = tokens.getNext ()) || sscanf (s, "%lf", &h) != 1 ||
	h <= 0.0)
      goto error;
    
    if (!(s = tokens.getNext ()) || strcmp (s, "vadjustment") ||
	!(s = tokens.getNext ()) || sscanf (s, "%lf", &v) != 1 ||
	v <= 0.0)
      goto error;
    
    while ((s = tokens.getNext ()))
    {
      double x, y;
      
      m = find (s);

      if (!(s = tokens.getNext ()) || sscanf (s, "%lf", &x) != 1 ||
	  !(s = tokens.getNext ()) || sscanf (s, "%lf", &y) != 1)
	goto error;

      if (m)
	m->position (x, y);
    }
    
    gtk_widget_set_usize (widget, width, height);
    gnome_canvas_set_pixels_per_unit (canvas, pixels_per_unit);
    gtk_adjustment_set_value (hadjustment, (float) h);
    gtk_adjustment_set_value (vadjustment, (float) v);

    toggled ();    
    return 1;
    
    error:
    
    printf ("line: %d. Syntax error near the text \"%s\".\n",
	    tokens.getLineNumber (),
	    tokens.getWord ());
    fflush (stdout);
    toggled ();

    return 0;
    
  }

  exception: return 0;
}


int mu_Cluster::toFile (FILE*fp)
{
  insist (this && widget);
  insist (fp);
  
  fprintf (fp, "height %d\n", widget->allocation.height);
  fprintf (fp, "width %d\n", widget->allocation.width);
  
  fprintf (fp, "pixels_per_unit %lf\n", (double) canvas->pixels_per_unit);
  fprintf (fp, "hadjustment %lf\n", (double) hadjustment->value);
  fprintf (fp, "vadjustment %lf\n", (double) vadjustment->value);
  
  for (int i = 0; i < numSwitches; i++)
  {
    double x1, y1, x2, y2;
    
    gnome_canvas_item_get_bounds (GNOME_CANVAS_ITEM (switches [i]->topGroup), &x1, &y1, &x2, &y2);

    fprintf (fp, "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x ",
	     switches [i]->ethernetAddress [0] & 0xff,
	     switches [i]->ethernetAddress [1] & 0xff,
	     switches [i]->ethernetAddress [2] & 0xff, 
	     switches [i]->ethernetAddress [3] & 0xff,
	     switches [i]->ethernetAddress [4] & 0xff,
	     switches [i]->ethernetAddress [5] & 0xff);

    fprintf (fp, "(%lf, %lf)\n", x1, y1);
  }

  for (mu_Host*h = (mu_Host*) hosts; h; h = (mu_Host*) h->next)
  {
    double x1, y1, x2, y2;
    insist (h->host && h->host->getName ());
    gnome_canvas_item_get_bounds (GNOME_CANVAS_ITEM (h->topGroup), &x1, &y1, &x2, &y2);
    fprintf (fp, "%s (%lf, %lf)\n",  h->host->getName (), x1, y1);
  }
  return 1;
  exception: return 0;
  
}
