#ifndef mu_Cluster_h
#define mu_Cluster_h

#include "m3_defines.h"
#include "pi_Ping.h"
#include "mu_Widget.h"
#include "m3_WebClient.h"
#include "m3_Description.h"
#include "m3_XbarTable.h"
#include "mt_Queueable.h"
#include "mu_Threadable.h"

#define MU_CLUSTER_NUM_COLORS 13


class mu_Cluster;
class mu_Link;
class mu_Switch;
class mu_LinecardPort;

class mu_Freeable
{
  public:
  mu_Freeable*next;
  mu_Freeable (mu_Freeable**head);
  virtual~mu_Freeable ();
};

class mu_Port
{
  public:
  m3_DescriptionPort*port;
  mu_Link*link;
  mu_int64 good [2];
  mu_int64 bad [2];
  int getCenter (double*x, double*y);
  int getBounds (double*x1, double*y1, double*x2, double*y2);
  mu_LinecardPort*getLinecardPort ();
  mu_Switch*getSwitch ();
  mu_Port ();
};

class mu_Host;

class mu_LinecardPort
{
  public:
  m3_DescriptionPort*port;
  mu_Switch*parent;
  mu_Host*host;
  GnomeCanvasItem*portItem,*ledItem;
  mu_Link*getLink ();
  mu_LinecardPort ();
  int callback (GnomeCanvasItem*item, GdkEvent*event);
  m3_DescriptionPort*getPort ();
};


class mu_Movable
{
  public:
  mu_Movable ();
  virtual ~mu_Movable ();
  mu_Cluster*cluster;
  GnomeCanvasItem*box;
  int selected;
  GnomeCanvasGroup*topGroup,*selectGroup;
  int build (GnomeCanvasGroup*group, double x, double y, int width, int height);
  virtual int callback (GnomeCanvasItem *item, GdkEvent *event);
  int select (int selected, int raise);
  int inside (double x1, double y1, double x2, double y2);
  int inside (double x, double y);
  int intersects (double x1, double y1, double x2, double y2);
  virtual int move (double dx, double dy) = 0;
  int position (double x, double y);
  virtual int isHost ();
};

class mu_Host : public mu_Port, public mu_Freeable, public mu_Movable
{
  public:
  mt_Node*host;
  GnomeCanvasItem*group;
  int routeSelected;
  mu_Host (mu_Freeable**head, mt_Node*node, mu_Cluster*cluster);
  virtual ~mu_Host ();
  int build (GnomeCanvasGroup*group, double x, double y);
  int move (double dx, double dy);
  int eatCounter (pi_Counter*counter);
  virtual int callback (GnomeCanvasItem *item, GdkEvent *event);
  virtual int isHost ();
};

class mu_Xbar
{
  public:
  GnomeCanvasGroup*topGroup, *pingGroup;
  GnomeCanvasItem*box;
  GnomeCanvasItem*portSquares [16];
  GnomeCanvasItem*pingSquares [16] [16];
  GnomeCanvasItem*pingText;
  mu_Port ports [16];
  mu_Switch*parent;
  int build (GnomeCanvasGroup*group, double x, double y);
  int show (int state);
  int show (int port, int state);
  int callback (GnomeCanvasItem*item, GdkEvent*event);  
  int portCallback (GnomeCanvasItem*item, GdkEvent*event);  
  mu_Xbar ();
};

class mu_Switch : public m3_Description, public mu_Movable, public mt_Queueable
{
  private:
  int moveLinecardLinks (int inside);
  
  public:
  int depth;
  GnomeCanvasItem*linecardItems [M3_MAX_LINE_CARD + 1];
  GnomeCanvasGroup*outsideGroup, *insideGroup;
  mu_LinecardPort linecardPorts [M3_MAX_LINE_CARD + 1] [M3_MAX_LINE_CARD_PORT];
  mu_Xbar xbars [M3_MAX_BACKPLANE_XBAR + M3_MAX_LINE_CARD];

  mu_Switch (mu_Cluster*cluster);
  virtual ~mu_Switch ();
  int build (double x, double y);
  int connect ();
  int move (double dx, double dy);
  int show ()
  {
    gnome_canvas_item_show (GNOME_CANVAS_ITEM (topGroup));
  }
  int showInside (int state);
  int showHosts (int state);
  int showCables (int state);
  int eatCounter (int xbar, int correction, pi_Counter*counter);
  int eatPacketCounters (int good, int numCounts, mu_int64*counters, mu_int64*totalTotal, mu_int64*total, mu_int64*maximum, mu_int64*minimum);
  int show (int state);
  int show (int linecard, int state);
  int show (int linecard, int port, int state);
  int drawPacketCounters (int good, int maximum, int minimum);
};

class mu_Link : public mu_Freeable
{
  private:
  GnomeCanvasPoints*points;
  mu_Cluster*getCluster ();
  
  public:
  GnomeCanvasItem*item;
  int routeCount;
  
  mu_Link (mu_Freeable**head);
  virtual~mu_Link ();
  mu_Port*port;
  mu_Port*other;
  int build (GnomeCanvasGroup*group);
  int move (double dx1, double dy1, double dx2, double dy2);
  int movePort (double x, double y);
  int moveOther (double x, double y);
  int show (int state);
};


class mu_Cluster : public mu_Widget
{
  private:
  GtkAdjustment*hadjustment,*vadjustment;
  m3_WebClient web;
  int building;
  pthread_mutex_t mutex;
 
  mu_Movable*find (char*name);
  int eradicate ();
  GnomeCanvasItem*selection;
  GnomeCanvasItem*outOfBounds,*bounds;
  GdkCursor*pointCursor, *zoomCursor,*handCursor,*questionCursor;
  GdkCursor*changeCursor;
  GdkCursor*makeCursor (GdkWindow*window, char*data, char*maskData, int width, int height);
  GtkWidget*makeToggleButton (GtkWidget*toolbar, char**data, char*description, GSList*group, GdkWindow*window, GtkStyle*style);
  int connect (mt_Graph*graph, m3_XbarTable*table);
  int populate (int*addresses, int numAddresses, char*directory);
  mu_Switch*getSwitchFromIpAddress (char*ipAddress);
  int build ();
  int lastEventType;
  int lastEventIndex;
  int lastEventPartType;
  char lastEventIpAddress [4];
  int getRanges (int*minGood, int*maxGood, int*minBad, int*maxBad);
  mu_Host*getHost (int index);
  int showTemperatures (int*smallest, int*biggest, int mode);
  int showTemperature (GnomeCanvasItem*item, int temperature, int*smallest, int*biggest, int mode);
  
  public:
  enum
  {
    CLEAR = 0,
    COMPUTE = 1,
    COLOR = 2
  };
  
  mu_Movable*selected;
  int numSwitches;
  mu_Switch**switches;
  GnomeCanvasGroup*linkGroup;
  ~mu_Cluster ();
  GnomeCanvas*canvas;
  GtkWidget*pointToggle,*zoomToggle,*handToggle,*questionToggle;
  GtkWidget*changeToggle;
  
  mu_Cluster (GtkWidget*mainWidget);
  mu_Switch*getSwitch (char*ethernetAddress);
  mu_Host*routeHost;
  mu_Freeable*links;
  mu_Freeable*hosts;
  
  int toFile (FILE*fp);
  int fromFile (FILE*fp);

  int build (mt_Graph*graph, m3_XbarTable*table, int*addresses, int numAddresses, char*directory);
  int zoom (double x, double y, double zoom);
  int selectAll (int selected);
  int selectAll (double x1, double y1, double x2, double y2, int selected);
  int select (mu_Movable*m);
  int callback (GnomeCanvasItem*item, GdkEvent*event);
  int toggled ();
  int inside (double x, double y);
  int inside (double x, double y, double x1, double y1, double x2, double y2);
  int inside (GnomeCanvasItem*item, double x1, double y1, double x2, double y2);
  int inside (GnomeCanvasItem*item, double x, double y);
  int intersects (GnomeCanvasItem*item, double x1, double y1, double x2, double y2);
  int showInside (int state);
  int showHosts (int state);
  int showCables (int state);
  int showXpoints (int state);
  int eatCounters (int testingXbars, int numCounters, pi_Counter*counters);
  int eatEvent (char ipAddress[4], int type, mu_int64 counter, int partType, int index, int state);
  int eatPacketCounters (int ipAddress, int good, int numCounts, mu_int64*counters, mu_int64*totalTotal, mu_int64*total, mu_int64*maximum, mu_int64*minimum);
  int drawPacketCounters (int good, int maximum, int minimum);
  void getSelection (GtkSelectionData*selection_data);
  int setSelection (char*s);
  int followRoute (mu_Host*from, mu_Host*to, int draw, int count);
  int drawRoutes ();
  int clearLinkColors ();
  int showCrcs (int good);
  int tree ();
  int refresh ();
  int clearCounters ();
  int stopSnmpTraps ();
  int select (char*name);
  int updateLegend (char*caption, double maximum, double minimum, int precision);
  int showTemperatures (int state);
};

class mu_SwitchHelper : public mu_Threadable
{
  public:

  mu_Switch*_switch;
  char filename [M3_DISPLAY_STRING_LENGTH + 1];
  int address;
  m3_WebClient web;
  
  mu_SwitchHelper ();
  void run (void*data);
};

#endif /*mu_Cluster_h*/


