#include <sys/types.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/time.h>

#include "mu_EventWidget.h"
#include "mu_MainWidget.h"
#include "insist.h"

extern mu_MainWidget*mu_mainWidget;
extern int mu_eventMask [2];
extern int mu_eventHistory;
extern int mu_logEvents;
extern char*mu_workingDirectory;

#define MU_EVENT_TEXT_LENGTH 50
#define MU_EVENT_HOSTNAME_LENGTH 256
#define MU_EVENT_NUM_COLUMNS 6

static int get_trap_type (char*s)
{
  for (int i = 0; i < M3_NUM_TRAPS; i++)
    if (!strcmp (m3_trap_names [i], s))
      return i + 1;
  return 0;
}

static int get_part_type (char*s)
{
  for (int i = 0; i < M3_NUM_TABLES; i++)
    if (!strcmp (m3_table_names [i], s))
      return i + 1;
  return 0;
}

static gboolean mu_EventWidget_delete (GtkWidget*w, GdkEvent*e, mu_EventWidget*widget)
{
  widget->hide ();
  return 1;
}

static gboolean mu_EventWidget_destroy (GtkWidget*w, mu_EventWidget*widget)
{
  widget->hide ();
  return 0;
}

static gint mu_EventWidget_toggled (GtkWidget*w, mu_EventWidget*eventWidget)
{
  insist (eventWidget);
  return eventWidget->toggled ();
  exception: return 0;
}

static void mu_EventWidget_selected (GtkWidget*w, gint row, gint column, GdkEventButton*event, mu_EventWidget*eventWidget)
{
  insist (eventWidget);
  eventWidget->selected (row, column);
  exception:;
}

int mu_EventWidget::addEvent (char ipAddress[4], int type, int low, int high, int partType, int index)
{
  time_t tp = time (0);
  char texts [MU_EVENT_NUM_COLUMNS] [MU_EVENT_TEXT_LENGTH];
  gchar*t[MU_EVENT_NUM_COLUMNS];
  mu_int64 a = ((mu_int64) low) | ((mu_int64) high << 32);
  
  for (int i = 0; i < MU_EVENT_NUM_COLUMNS; i++)
    t[i] = texts [i];
  
  insist (ipAddress);
  insist (type >= 0 && type < M3_NUM_TRAPS);  
  insist (partType > 0 && partType <= M3_NUM_TABLES);
  
  strftime (texts [0], MU_EVENT_TEXT_LENGTH, "%H:%M:%S", localtime (&tp));

  sprintf (texts [1], "%d.%d.%d.%d", 
	   (unsigned) (ipAddress [0] & 0xff),
	   (unsigned) (ipAddress [1] & 0xff),
	   (unsigned) (ipAddress [2] & 0xff),
	   (unsigned) (ipAddress [3] & 0xff));

  sprintf (texts [2], "%s", m3_trap_names [type]);
  sprintf (texts [3], "%llu", a);
  sprintf (texts [4], "%s", m3_table_names [partType - 1]);
  sprintf (texts [5], "%d", index);

  gdk_threads_enter ();
  
  while (GTK_CLIST (list)->rows > mu_eventHistory - 1)
    gtk_clist_remove (GTK_CLIST (list), 0);
  gtk_clist_append (GTK_CLIST (list), t);

  if (fp)
  {
    for (int i = 0; i < MU_EVENT_NUM_COLUMNS; i++)
      fprintf (fp, "%s ", texts [i]);
    fprintf (fp, "\n");
  }
  
  gtk_clist_select_row (GTK_CLIST (list), GTK_CLIST (list)->rows - 1, 1);

  gdk_threads_leave ();

  return 1;
  exception: return 0;
  
}

int mu_EventWidget::run (void*data)
{
  insist (this && fd && port);
  char s [FILENAME_LENGTH];
  
  sprintf (s, "%s/%s", *mu_workingDirectory ? mu_workingDirectory : "", "mute.events");

  fp = mu_logEvents ? fopen (s, "a") : 0;
    
  while (!isStopped ())
  {
    int partType, index, type;
    m3_notify_message_t m;
    sockaddr_in a;
    int length = sizeof (a);
    fd_set mask;
    struct timeval timeout;
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    FD_ZERO (&mask);
    FD_SET (fd, &mask);
    
    if (select (1 + fd, &mask, 0, 0, &timeout) == 1 &&
	recvfrom (fd, (char*) &m, sizeof (m), 0, (struct sockaddr*) &a, (socklen_t*) &length) == sizeof (m))
    {
      length = sizeof (a);
      
      type = htonl (m.type);
      index = htonl (m.index);
      partType = htonl (m.partType);
      
      if (type == M3_ucResetTrap ||
	  type == M3_ucFirmwareFaultTrap || 
	  type == M3_ucOvertemperatureShutdownFaultTrap || 
	  type == M3_ucDeadTrap || 
	  type == M3_slotLostCommunicationTrap || 
	  type == M3_slotReestablishCommunicationTrap || 
	  type == M3_slotChangedTrap)
      {
	partType = M3_SLOT_TABLE;
	index = htonl (m.slot);
      }

      if (!addEvent ((char*)&a.sin_addr.s_addr, htonl (m.type) - 1, htonl (m.low), htonl (m.high), partType, index))
	goto exception;
    }
  }

  if (fp) fclose (fp);
  fp = 0;
  return 1;

  exception:
  if (fp) fclose (fp);
  fp = 0;
  return 0;
}

int mu_EventWidget::toggled ()
{
  if (!setNotify (0))
    return 0;
  
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (listenToggle)))
    createThread ((mu_ThreadableFunction)&mu_EventWidget::run, (void*)0);
  else
    stopThread (0);

  return 1;
  exception: return 0;
}

int mu_EventWidget::setNotify (int onlyIfListening)
{
  int*addresses = 0;
  int numAddresses;
  int types [2];
  
  insist (this && listenToggle);
  insist (mu_mainWidget && mu_mainWidget->exploreWidget);
  insist (fd > 0);

  if (onlyIfListening && !gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (listenToggle)))
    return 1;
  
  if (!(numAddresses = mu_mainWidget->exploreWidget->getAddresses (&addresses)))
    return 0;
  insist (addresses);

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (listenToggle)))
  {
    types [0] = mu_eventMask [0];
    types [1] = mu_eventMask [1];
  }
  else
    types [0] = types [1] = 0;
  
  for (int i = 0; i < numAddresses; i++)
    if (!client.open (addresses [i]) || !client.setNotify (myself, port, types [0], types [1]))
      return 0;

  return 1;
  exception: return 0;
}

static char *mu_eventTitles [MU_EVENT_NUM_COLUMNS] =
{
  "Time",  "IP address", "Event", "Count", "Part", "Index"
};

static int mu_eventWidths [MU_EVENT_NUM_COLUMNS] =
{
  80, 120, 150, 100, 130, 100
};

mu_EventWidget::mu_EventWidget ()
{
  GtkWidget*box, *scrolledWindow, *button, *bottom;
  
  fp = 0;
  
  widget = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  insist (widget);
  
  gtk_window_set_title (GTK_WINDOW (widget), MU_TITLE " - Events");
  gtk_window_set_policy (GTK_WINDOW (widget), 1, 1, 0);
  gtk_widget_set_usize (widget, 700, 700);
  
  box = gtk_vbox_new (0, MU_SPACING);
  insist (box);
  gtk_widget_show (box);
  gtk_container_add (GTK_CONTAINER (widget), box);
  
  scrolledWindow = gtk_scrolled_window_new (0, 0);
  insist (scrolledWindow);
  gtk_widget_show (scrolledWindow);
  
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledWindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_box_pack_start (GTK_BOX (box), scrolledWindow, 1, 1, MU_SPACING);
  
  list = gtk_clist_new_with_titles (MU_EVENT_NUM_COLUMNS, mu_eventTitles);
  insist (list);
  gtk_widget_show (list);
  gtk_container_add (GTK_CONTAINER (scrolledWindow), GTK_WIDGET (list));

  for (int i = 0; i < MU_EVENT_NUM_COLUMNS; i++)
    gtk_clist_set_column_width (GTK_CLIST (list), i, mu_eventWidths [i]);

  gtk_signal_connect (GTK_OBJECT (widget), "delete_event", GTK_SIGNAL_FUNC (mu_EventWidget_delete), this);
  gtk_signal_connect (GTK_OBJECT (widget), "destroy", GTK_SIGNAL_FUNC (mu_EventWidget_destroy), this);

  bottom = gtk_hbox_new (1, MU_SPACING);
  insist (bottom);
  gtk_widget_show (bottom);
  gtk_box_pack_start (GTK_BOX (box), bottom, 0, 0, MU_SPACING);

  listenToggle = gtk_check_button_new_with_label ("Listen");
  insist (listenToggle);
  gtk_widget_show (listenToggle);
  gtk_box_pack_start (GTK_BOX (bottom), listenToggle, 0, 0, MU_SPACING);
  gtk_signal_connect (GTK_OBJECT (listenToggle), "toggled",
		      GTK_SIGNAL_FUNC (mu_EventWidget_toggled),
		      this);

  button = gtk_button_new_with_label ("Close");
  insist (button);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (bottom), button, 0, 0, MU_SPACING);
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_EventWidget_destroy), this);
  
  gtk_signal_connect (GTK_OBJECT (list), "select_row", GTK_SIGNAL_FUNC (mu_EventWidget_selected), this);
  
  open ();
  exception:;
}

int mu_EventWidget::open ()
{
  int length;
  struct sockaddr_in a;
  struct hostent*hp;
  char hostname [MU_EVENT_HOSTNAME_LENGTH + 1];

  insist (this);
  
  if ((fd = socket (AF_INET, SOCK_DGRAM, 0)) <= 0)
  {
    perror ("bind failed");
    return 0;
  }

  memset ((char*)&a, 0, sizeof (a));
  
  a.sin_family = AF_INET;
  a.sin_addr.s_addr = INADDR_ANY;
  a.sin_port = 0;
  
  if (bind (fd, (struct sockaddr*) &a, sizeof (a)) < 0)
  {
    perror ("bind failed");
    return 0;
  }
  
  length = sizeof (a);
  if (getsockname (fd, (struct sockaddr*) &a, (socklen_t*) &length) < 0)
  {
    perror ("getsockname failed");
    return 0;
  }

  if ((port = ntohs (a.sin_port)) <= 0)
  {
    printf ("port is zero\n");
    fflush (stdout);
    return 0;
  }
  
  if (gethostname (hostname, MU_EVENT_HOSTNAME_LENGTH) < 0 || !(hp = gethostbyname (hostname)))
  {
    perror ("couldn't get our ip address\n");
    return 0;
  }
  memcpy (myself, hp->h_addr, 4);
  return 1;
  exception: return 0;
}

mu_EventWidget::~mu_EventWidget ()
{
  if (fd)
    close (fd);
}

int mu_EventWidget::selected (int row, int column)
{
  insist (this && row >= 0);
  insist (list);
  
  int r;
  int fat [4];
  char ipAddress [4];
  int index, type, partType;
  mu_int64 count;
  char*s;

  r = gtk_clist_get_text (GTK_CLIST (list), row, 1, &s);
  insist (r && s);
  r = sscanf (s, "%d.%d.%d.%d", &fat [0], &fat [1], &fat [2], &fat [3]);
  insist (r == 4);
  for (int i = 0; i < 4; i++)
    ipAddress [i] = fat [i];
  
  r = gtk_clist_get_text (GTK_CLIST (list), row, 2, &s);
  insist (r && s);

  type = get_trap_type (s);
  insist (type > 0 && type <= M3_NUM_TRAPS);

  r = gtk_clist_get_text (GTK_CLIST (list), row, 3, &s);
  insist (r && s);
  r = sscanf (s, "%llu", &count);
  insist (r == 1);

  r = gtk_clist_get_text (GTK_CLIST (list), row, 4, &s);
  insist (r && s);
  partType = get_part_type (s);
  insist (partType >= 0 && partType <= M3_NUM_TABLES);
  
  r = gtk_clist_get_text (GTK_CLIST (list), row, 5, &s);
  insist (r && s);
  r = sscanf (s, "%d", &index);
  insist (r == 1);

  if (mu_mainWidget && mu_mainWidget->cluster)
    mu_mainWidget->cluster->eatEvent (ipAddress, type, count, partType, index, 1);

  return 1;
  exception: return 0;
}
