#include "gnome.h"
#include "insist.h"
#include "mt_MapFile.h"
#include "mm_MapperModule.h"
#include "mu_MainWidget.h"
#include "mu_Cluster.h"

extern char*mu_workingDirectory;
extern mu_MainWidget*mu_mainWidget;

static gboolean mu_ExploreWidget_delete (GtkWidget*w, GdkEvent*e, mu_ExploreWidget*widget)
{
  widget->hide ();
  return 1;
}

static gboolean mu_ExploreWidget_destroy (GtkWidget*w, mu_ExploreWidget*widget)
{
  widget->hide ();
  return 0;
}

static void mu_ExploreWidget_close (GtkWidget*w, mu_ExploreWidget*widget)
{
  widget->hide ();
}
static void mu_ExploreWidget_mapButton_clicked (GtkWidget*w, mu_ExploreWidget*exploreWidget)
{
  exploreWidget->createThread ((mu_ThreadableFunction)&mu_ExploreWidget::run, (void*)0);
}

GtkWidget*mu_ExploreWidget::makeDialogBox ()
{
  GtkWidget*box, *button, *table;
  
  insist (this);
  
  box = gtk_vbox_new (0, MU_SPACING);
  insist (box);
  gtk_widget_show (box);

  table = gtk_table_new (8, 11, 0);
  insist (table);
  gtk_widget_show (table);
  gtk_box_pack_start (GTK_BOX (box), table, 0, 0, MU_SPACING);

  runMapperCheckButton = gtk_check_button_new_with_label ("Run Mapper");
  insist (runMapperCheckButton);
  gtk_table_attach (GTK_TABLE (table), runMapperCheckButton, 0, 4, 0, 1, GTK_FILL, GTK_FILL, MU_SPACING, MU_SPACING);
  gtk_widget_show (runMapperCheckButton);  

  configureCheckButton = gtk_check_button_new_with_label ("Configure GM");
  insist (configureCheckButton);
  gtk_table_attach (GTK_TABLE (table), configureCheckButton, 1, 4, 1, 2, GTK_FILL, GTK_FILL, MU_SPACING, MU_SPACING);
  gtk_widget_show (configureCheckButton);

  shortestPathCheckButton = gtk_check_button_new_with_label ("Shortest path");
  insist (shortestPathCheckButton);
  gtk_table_attach (GTK_TABLE (table), shortestPathCheckButton, 1, 4, 2, 3, GTK_FILL, GTK_FILL, MU_SPACING, MU_SPACING);
  gtk_widget_show (shortestPathCheckButton);

  loopCheckButton = gtk_check_button_new_with_label ("Find Loopbacks");
  insist (loopCheckButton);  
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (loopCheckButton), 1);
  gtk_table_attach (GTK_TABLE (table), loopCheckButton, 1, 4, 3, 4, GTK_FILL, GTK_FILL, MU_SPACING, MU_SPACING);
  gtk_widget_show (loopCheckButton);

  writeRoutesCheckButton = gtk_check_button_new_with_label ("Write Routes");
  insist (writeRoutesCheckButton);  
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (writeRoutesCheckButton), 1);
  gtk_table_attach (GTK_TABLE (table), writeRoutesCheckButton, 1, 4, 4, 5, GTK_FILL, GTK_FILL, MU_SPACING, MU_SPACING);
  gtk_widget_show (writeRoutesCheckButton);

  verboseCheckButton = gtk_check_button_new_with_label ("Verbose");
  insist (verboseCheckButton);
  //gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (verboseCheckButton), 1);
  gtk_table_attach (GTK_TABLE (table), verboseCheckButton, 1, 4, 5, 6, GTK_FILL, GTK_FILL, MU_SPACING, MU_SPACING);
  gtk_widget_show (verboseCheckButton);

  findSwitchesCheckbox = gtk_check_button_new_with_label ("Find Switches");
  insist (findSwitchesCheckbox);
  gtk_table_attach (GTK_TABLE (table), findSwitchesCheckbox, 0, 4, 6, 7, GTK_FILL, GTK_FILL, MU_SPACING, MU_SPACING);
  gtk_widget_show (findSwitchesCheckbox);  

  findXbarsCheckbox = gtk_check_button_new_with_label ("Find Xbars");
  insist (findXbarsCheckbox);
  gtk_table_attach (GTK_TABLE (table), findXbarsCheckbox, 0, 4, 7, 8, GTK_FILL, GTK_FILL, MU_SPACING, MU_SPACING);
  gtk_widget_show (findXbarsCheckbox);

  readHtmlCheckbox = gtk_check_button_new_with_label ("Read HTML files (offline)");
  insist (readHtmlCheckbox);
  gtk_table_attach (GTK_TABLE (table), readHtmlCheckbox, 0, 4, 8, 9, GTK_FILL, GTK_FILL, MU_SPACING, MU_SPACING);
  gtk_widget_show (readHtmlCheckbox);

  GtkWidget*b2;
  b2 = gtk_hbox_new (0, MU_SPACING);
  insist (b2);
  gtk_widget_show (b2);
  gtk_box_pack_end (GTK_BOX (box), b2, 0, 0, MU_SPACING);

  buildButton = gtk_button_new_with_label ("Build");
  gtk_signal_connect (GTK_OBJECT (buildButton), "clicked", GTK_SIGNAL_FUNC (mu_ExploreWidget_mapButton_clicked), this);
  gtk_widget_show (buildButton);
  gtk_box_pack_start (GTK_BOX (b2), buildButton, 1, 0, MU_SPACING);
  
  button = gtk_button_new_with_label ("Close");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_ExploreWidget_close), this);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (b2), button, 1, 0, MU_SPACING);

  setStates ();
  
  return box;
  exception: return 0;
}

mu_ExploreWidget::mu_ExploreWidget ()
{
  widget = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  insist (widget);
  
  gtk_window_set_title (GTK_WINDOW (widget), MU_TITLE " - Build Network Graph");
  gtk_window_set_policy (GTK_WINDOW (widget), 1, 1, 0);

  GtkWidget*windowBox;
  windowBox = gtk_hbox_new (0, MU_SPACING);
  insist (windowBox);
  gtk_widget_show (windowBox);
  
  gtk_container_add (GTK_CONTAINER (widget), windowBox);

  gtk_signal_connect (GTK_OBJECT (widget), "delete_event", GTK_SIGNAL_FUNC (mu_ExploreWidget_delete), this);
  gtk_signal_connect (GTK_OBJECT (widget), "destroy", GTK_SIGNAL_FUNC (mu_ExploreWidget_destroy), this);

  GtkWidget*left, *textBox, *scrollbar;

  left = makeDialogBox ();
  insist (left);
  gtk_box_pack_start (GTK_BOX (windowBox), left, 0, 0, MU_SPACING);

  textBox = gtk_hbox_new (0, MU_SPACING);
  insist (textBox);
  gtk_box_pack_start (GTK_BOX (windowBox), textBox, 1, 1, MU_SPACING);

  threadText = gtk_text_new (0, 0);
  gtk_text_set_editable (GTK_TEXT (threadText), 0);
  scrollbar = gtk_vscrollbar_new (GTK_TEXT (threadText)->vadj);
  gtk_box_pack_start (GTK_BOX (textBox), threadText, 1, 1, MU_SPACING);
  gtk_box_pack_start (GTK_BOX (textBox), scrollbar, 0, 0, 0);
  
  gtk_widget_set_usize (textBox, 400, 300);
  
  gtk_widget_show (scrollbar);
  gtk_widget_show (textBox);
  gtk_widget_show (threadText);
  
  calculator = mt_getCalculator ();
  network = mt_getNetwork ();
  mapper = (mt_MapperModule*) mt_getMapperModule ();

  xbarTable = 0;
  insist (calculator && network && mapper);

  exception:;
}

void mu_ExploreWidget::run (void*data)
{
  int r;
  
  insist (this && mapper && network && calculator);
  insist (runMapperCheckButton);
  
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (findSwitchesCheckbox)))
  {
    mt_Component::printFormat ("finding switches");

    rollcall.open ();
    numAddresses = rollcall.findAddresses (&addresses);
    rollcall.close ();
    for (int i = 0; i < numAddresses; i++)
    {
      mt_Component::printFormat ("%d.%d.%d.%d",
				 (unsigned) (((char*)(&addresses [i])) [0] & 0xff),
				 (unsigned) (((char*)(&addresses [i])) [1] & 0xff),
				 (unsigned) (((char*)(&addresses [i])) [2] & 0xff),
				 (unsigned) (((char*)(&addresses [i])) [3] & 0xff));
      
    }
    sf.toFile (switchFile, numAddresses, addresses);
  }
  else
  {
    mt_Component::printFormat ("reading %s", switchFile);
    numAddresses = sf.fromFile (switchFile, &addresses);
  }

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (runMapperCheckButton)))
  {
    mt_Args mapperArgs (mt_Args::MAPPER, 100);
    mt_Args calculatorArgs (mt_Args::CALCULATOR, 100);

    calculatorArgs.addArg ("-spread");
    calculatorArgs.addArg ("-save-routes");

    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (shortestPathCheckButton)))
    {
      calculatorArgs.addArg ("-shortest-path");
      calculatorArgs.addArg ("-margin"); calculatorArgs.addArg ("0");
    }

    mapperArgs.addArg ("-map-once");
    mapperArgs.addArg ("-warnings");
    //mapperArgs.addArg ("-very-verbose");
    mapperArgs.addArg ("-send-many-at-once");
    mapperArgs.addArg ("-send-many-compares");
    mapperArgs.addArg ("-compare-switches");
    mapperArgs.addArg ("-make-hosts");
    mapperArgs.addArg ("-scout-tries"); mapperArgs.addArg ("3");
    mapperArgs.addArg ("-scout-time"); mapperArgs.addArg ("5000");
    mapperArgs.addArg ("-probe-tries"); mapperArgs.addArg ("3");
    mapperArgs.addArg ("-probe-time"); mapperArgs.addArg ("5000");
    mapperArgs.addArg ("-configure-tries"); mapperArgs.addArg ("10");
    mapperArgs.addArg ("-configure-time"); mapperArgs.addArg ("10000");

    if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (configureCheckButton)))
      mapperArgs.addArg ("-never-configure");
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (verboseCheckButton)))
      mapperArgs.addArg ("-verbose");
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (loopCheckButton)))
      mapperArgs.addArg ("-find-loops");
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (writeRoutesCheckButton)))
    {
      mapperArgs.addArg ("-routes-file"); mapperArgs.addArg (routesFile);
    }
    
    mapperArgs.addArg ("-host-file"); mapperArgs.addArg (hostsFile);
    mapperArgs.addArg ("-map-file"); mapperArgs.addArg (mapFile);

    calculator->parseArgs (&calculatorArgs);  
    mapper->parseArgs (&mapperArgs);
    mapper->initialize (network, calculator);
    mapper->start ();
    graph.clone (((mm_MapperModule*)mapper)->getGraph ());
    mapper->cleanup ();
  }
  else
  {
    mt_Component::printFormat ("reading %s", mapFile);
    mt_MapFile mf (mapFile, mt_File::_READ);
    mf.read (&graph);
  }
  
  graph.setRoot (graph.getHost (0));//kludge
  if (xbarTable)
    delete xbarTable;
  xbarTable = 0;
  
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (findXbarsCheckbox)))
  {
    insist (mapper && network);
    mt_Job*job;
    mt_Address address;
    char networkHostname [mt_Network::HOSTNAME_LENGTH + 1];
    int hostType;
    int unit = 0;

    mt_Component::printFormat ("finding xbars");
    mapper->initialize (network, calculator);
    job = ((mm_MapperModule*)mapper)->getJob ();
    insist (job);
    
    r = network->open (job, &address, networkHostname, &hostType, unit, 3);
    insist (r);

    if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (runMapperCheckButton)))
      graph.computeRoutes (graph.getRoot ());

    for (;;)
    {
      int e, length;
      char*p;

      mt_Component::printFormat ("waiting to avoid lingering invalid routes");
      network->setTimer (job, 2 * 1000 * 1000);
      while ((e = network->wait (&p, &length)) != mt_Network::TIMEOUT);

      m3_XbarTable*t = new m3_XbarTable (graph.getNumSwitches ());
      insist (t);
      
      if (!t->fromNetwork (job, &graph, addresses, numAddresses))
      {
	mt_Component::printFormat ("Trying again.");
	delete t;
	continue;
      }

      if (xbarTable && t->equals (xbarTable))
      {
	delete t;
	break;
      }
      mt_Component::printFormat ("XbarTables did not match. Trying again.");
      if (xbarTable)
	delete xbarTable;
      xbarTable = t;
    }
    
    network->close (job);
    xbarTable->toFile (xbarFile);
    
  }
  else
  {
    xbarTable = new m3_XbarTable (graph.getNumSwitches ());
    mt_Component::printFormat ("reading %s", xbarFile);
    xbarTable->fromFile (graph.getRoot (), &graph, xbarFile);
  }
  mt_Component::printFormat ("examining M3 switches...");
  insist (mu_mainWidget && mu_mainWidget->cluster);

  mu_mainWidget->cluster->build (&graph, xbarTable, addresses, numAddresses,
				 gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (readHtmlCheckbox)) ? mu_workingDirectory : 0);
  gdk_threads_enter ();
  mu_mainWidget->checkStates ();
  gdk_threads_leave ();

  mt_Component::printFormat ("done");
  setStates ();
  exception:;
}

int mu_ExploreWidget::setFilename (char*s, char*directory, char*base)
{
  FILE*fp;
  insist (directory && base && *base);
  sprintf (s, "%s/%s", *directory ? directory : "", base);
  
  fp = fopen (s, "r");
  if (fp > 0)
    fclose (fp);

  return fp > 0;

  exception: return 0;
}

int mu_ExploreWidget::setFileState (char*filename, char*directory, char*basename, GtkWidget*checkButton)
{
  insist (this && filename && directory && basename && checkButton);

  if (setFilename (filename, directory, basename))
  {
    gtk_widget_set_sensitive (checkButton, 1);
  }
  else
  {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkButton), 1);
    gtk_widget_set_sensitive (checkButton, 0);
  }
  return 1;
  exception: return 0;
}

int mu_ExploreWidget::checkForWebFiles (int numAddresses, int*addresses)
{
  insist (this && numAddresses > 0 && addresses);
  
  for (int i = 0; i < numAddresses; i++)
  {
    char s1 [M3_DISPLAY_STRING_LENGTH];
    char s2 [M3_DISPLAY_STRING_LENGTH];

    sprintf (s1, "%d.%d.%d.%d.html",
	     (unsigned) (((char*)(&addresses [i])) [0] & 0xff),
	     (unsigned) (((char*)(&addresses [i])) [1] & 0xff),
	     (unsigned) (((char*)(&addresses [i])) [2] & 0xff),
	     (unsigned) (((char*)(&addresses [i])) [3] & 0xff));

    if (!setFilename (s2, mu_workingDirectory, s1))
      return 0;
  }
  return 1;
  exception: return 0;
}

int mu_ExploreWidget::setStates ()
{
  insist (this);
  insist (runMapperCheckButton && configureCheckButton && findSwitchesCheckbox && findXbarsCheckbox);
  
  setFileState (mapFile, mu_workingDirectory, "mute.map", runMapperCheckButton);
  setFileState (switchFile, mu_workingDirectory, "mute.switches", findSwitchesCheckbox);
  setFileState (xbarFile, mu_workingDirectory, "mute.xbars", findXbarsCheckbox);
  setFilename (routesFile, mu_workingDirectory, "mute.routes");
  setFilename (hostsFile, mu_workingDirectory, "mute.hosts");

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (findSwitchesCheckbox), 0);
  gtk_widget_set_sensitive (findSwitchesCheckbox, 0);

  if (GTK_WIDGET_IS_SENSITIVE (findSwitchesCheckbox))
  {
    int numAddresses;
    int*addresses;
    m3_SwitchFile sf;

    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (readHtmlCheckbox), (numAddresses = sf.fromFile (switchFile, &addresses)) && checkForWebFiles (numAddresses, addresses));
  }
      
  exception: return 0;
}


int mu_ExploreWidget::getAddresses (int**addresses)
{
  insist (this && addresses);
  *addresses = this->addresses;
  return numAddresses;
  exception: return 0;
}
