#include "insist.h"
#include "mu_FlashWidget.h"
#include "mu_ExploreWidget.h"

extern char*mu_workingDirectory;
extern char*mu_firmwareDirectory;

#define MU_FLASH_NUM_COLUMNS 3
#define MU_FLASH_TEXT_LENGTH 50

static char *mu_flashTitles [MU_FLASH_NUM_COLUMNS] =
{
  "IP Address         ",  "Ethernet Address      ", "Firmware Version"
};

static gboolean mu_FlashWidget_delete (GtkWidget*w, GdkEvent*e, mu_FlashWidget*widget)
{
  widget->hide ();
  return 1;
}

static gboolean mu_FlashWidget_destroy (GtkWidget*w, mu_FlashWidget*widget)
{
  widget->hide ();
  return 0;
}

static void mu_FlashWidget_read (GtkWidget*w, mu_FlashWidget*flashWidget)
{
  flashWidget->createThread ((mu_ThreadableFunction)&mu_FlashWidget::readSwitches, (void*)0);
}

static void mu_FlashWidget_update (GtkWidget*w, mu_FlashWidget*flashWidget)
{
  flashWidget->createThread ((mu_ThreadableFunction)&mu_FlashWidget::updateFirmware, (void*)0);
}

static void mu_FlashWidget_find (GtkWidget*w, mu_FlashWidget*flashWidget)
{
  flashWidget->createThread ((mu_ThreadableFunction)&mu_FlashWidget::findSwitches, (void*)0);
}

mu_FlashWidget::mu_FlashWidget ()
{
  GtkWidget*box, *scrolledWindow, *scrollbar;
  GtkWidget*textBox,*hbox;
  
  numAddresses = 0;
  addresses = 0;
  versions = 0;
  
  widget = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  insist (widget);  
  
  gtk_window_set_title (GTK_WINDOW (widget), MU_TITLE " - Update Firmware");
  gtk_window_set_policy (GTK_WINDOW (widget), 1, 1, 0);
  gtk_widget_set_usize (widget, 600, 500);

  box = gtk_vbox_new (0, MU_SPACING);
  insist (box);
  gtk_container_add (GTK_CONTAINER (widget), box);
  gtk_widget_show (box);
  
  scrolledWindow = gtk_scrolled_window_new (0, 0);
  insist (scrolledWindow);
  gtk_widget_show (scrolledWindow);

  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledWindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_box_pack_start (GTK_BOX (box), scrolledWindow, 1, 1, MU_SPACING);
  
  list = gtk_clist_new_with_titles (MU_FLASH_NUM_COLUMNS, mu_flashTitles);
  insist (list);
  gtk_widget_show (list);
  gtk_container_add (GTK_CONTAINER (scrolledWindow), GTK_WIDGET (list));

  scrolledWindow = gtk_scrolled_window_new (0, 0);
  insist (scrolledWindow);
  gtk_widget_show (scrolledWindow);
  
  hbox = gtk_hbox_new (1, MU_SPACING);
  insist (hbox);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (box), hbox, 0, 0, MU_SPACING);  

  readButton = gtk_button_new_with_label ("Read Switches");
  insist (readButton);
  gtk_box_pack_start (GTK_BOX (hbox), readButton, 0, 0, MU_SPACING);
  gtk_widget_show (readButton);
  
  findButton = gtk_button_new_with_label ("Find Switches");
  insist (findButton);
  gtk_box_pack_start (GTK_BOX (hbox), findButton, 0, 0, MU_SPACING);
  gtk_widget_show (findButton);
 
  textBox = gtk_hbox_new (0, MU_SPACING);
  insist (textBox);
  gtk_box_pack_start (GTK_BOX (box), textBox, 0, 0, MU_SPACING);
 
  threadText = gtk_text_new (0, 0);
  gtk_text_set_editable (GTK_TEXT (threadText), 0);
  scrollbar = gtk_vscrollbar_new (GTK_TEXT (threadText)->vadj);
  gtk_box_pack_start (GTK_BOX (textBox), threadText, 1, 1, MU_SPACING);
  gtk_box_pack_start (GTK_BOX (textBox), scrollbar, 0, 0, 0);

  gtk_widget_show (scrollbar);
  gtk_widget_show (textBox);
  gtk_widget_show (threadText);

  switchLabel = gtk_label_new ("switch label");
  insist (switchLabel);
  gtk_widget_show (switchLabel);
  gtk_box_pack_start (GTK_BOX (box), switchLabel, 0, 0, MU_SPACING);
  
  flashLabel = gtk_label_new ("flash label");
  insist (flashLabel);
  gtk_widget_show (flashLabel);
  gtk_box_pack_start (GTK_BOX (box), flashLabel, 0, 0, MU_SPACING);

  hbox = gtk_hbox_new (1, MU_SPACING);
  insist (hbox);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (box), hbox, 0, 0, MU_SPACING);

  updateButton = gtk_button_new_with_label ("Update Firmware");
  insist (updateButton);
  gtk_box_pack_start (GTK_BOX (hbox), updateButton, 0, 0, MU_SPACING);
  gtk_widget_show (updateButton);

  closeButton = gtk_button_new_with_label ("Close");
  insist (closeButton);
  gtk_box_pack_start (GTK_BOX (hbox), closeButton, 0, 0, MU_SPACING);
  gtk_widget_show (closeButton);
  gtk_signal_connect (GTK_OBJECT (widget), "delete_event", GTK_SIGNAL_FUNC (mu_FlashWidget_delete), this);
  gtk_signal_connect (GTK_OBJECT (widget), "destroy", GTK_SIGNAL_FUNC (mu_FlashWidget_destroy), this);

  gtk_signal_connect (GTK_OBJECT (closeButton), "clicked", GTK_SIGNAL_FUNC (mu_FlashWidget_destroy), this);
  gtk_signal_connect (GTK_OBJECT (updateButton), "clicked", GTK_SIGNAL_FUNC (mu_FlashWidget_update), this);
  gtk_signal_connect (GTK_OBJECT (findButton), "clicked", GTK_SIGNAL_FUNC (mu_FlashWidget_find), this);
  gtk_signal_connect (GTK_OBJECT (readButton), "clicked", GTK_SIGNAL_FUNC (mu_FlashWidget_read), this);

  updateFilenames ();
  exception:;
}

void mu_FlashWidget::updateFilenames ()
{
  insist (this && readButton && updateButton);
  
  gtk_widget_set_sensitive (readButton, mu_ExploreWidget::setFilename (switchFile, mu_workingDirectory, "mute.switches"));
  gtk_widget_set_sensitive (updateButton, mu_ExploreWidget::setFilename (flashFile, mu_firmwareDirectory, "vxWorks_var.hex"));

  gtk_label_set_text (GTK_LABEL (switchLabel), switchFile);
  gtk_label_set_text (GTK_LABEL (flashLabel), flashFile);

  exception:;  
}

void mu_FlashWidget::show ()
{
  updateFilenames ();
  mu_Widget::show ();
}

int mu_FlashWidget::findSwitches (void*)
{
  insist (this && list);

  mt_Component::printFormat ("finding switches...");

  rollcall.open ();
  numAddresses = rollcall.findAddresses (&addresses);
  rollcall.close ();
  sf.toFile (switchFile, numAddresses, addresses);

  if (getVersions ())
    fillList ();

  mt_Component::printFormat ("done.");
  exception: return 0;  
}

int mu_FlashWidget::readSwitches (void*)
{
  insist (this && list);

  mt_Component::printFormat ("reading %s", switchFile);
  numAddresses = sf.fromFile (switchFile, &addresses);

  if (getVersions ())
    fillList ();

  mt_Component::printFormat ("done.");

  exception: return 0;
}

int mu_FlashWidget::updateFirmware (void*)
{
  char*image = 0;
  int size;
  unsigned start;
  int numUpdated = 0;
  
  
  insist (this);
  insist (numAddresses > 0);
  insist (versions);
  
  if (!(size = hex2flash.getImage (flashFile, &image, &start)))
    return 0;

  insist (image);
  
  if (!hex2flash.getVersionString (image, size, versionString))
  {
    free (image);
    return 0;
  }
  mt_Component::printFormat ("version is %s", versionString);
  
  char*d1,*d2;
  d1 = strchr (versionString, ' ');
  insist (d1);
  
  for (int i = 0; i < numAddresses; i++)
  {
    char hostname [40];
    gdk_threads_enter ();
    gtk_clist_select_row (GTK_CLIST (list), i, 0);
    gdk_threads_leave ();

    d2 = strchr (versions [i].version, ' ');
    insist (d2);

    sprintf (hostname, "%d.%d.%d.%d",
			       (unsigned) (((char*)(&addresses [i])) [0] & 0xff),
			       (unsigned) (((char*)(&addresses [i])) [1] & 0xff),
			       (unsigned) (((char*)(&addresses [i])) [2] & 0xff),
	     (unsigned) (((char*)(&addresses [i])) [3] & 0xff));
    
    mt_Component::printFormat ("checking %s", hostname);

    if (testClient.firmwareDateCompare (d1, d2) <= 0)
    {
      mt_Component::printFormat ("up to date");
      continue;
    }
    numUpdated++;
    mt_Component::printFormat ("not up to date. updating now.");
    if (!hex2flash.program (image, size, start, hostname, 1, 0, 0, 0, 0, 0) ||
	!testClient.open (addresses [i]) || !testClient.reboot ())
      return 0;
  }
  mt_Component::printFormat ("updated %d switch%s", numUpdated, numUpdated == 1 ? "" : "es");
  
  if (numUpdated)
  {
    mt_Component::printFormat ("sleeping 10 seconds to let %d rebooted switch%s come up.", numUpdated, numUpdated == 1 ? "" : "es");
    sleep (10);
    if (getVersions ())
      fillList ();
  }
  mt_Component::printFormat ("done");
  exception: return 0;
}

int mu_FlashWidget::fillList ()
{
  insist (this && list && versions);
  char texts [MU_FLASH_NUM_COLUMNS] [MU_FLASH_TEXT_LENGTH];
  gchar*t [MU_FLASH_NUM_COLUMNS];

  gdk_threads_enter ();

  gtk_clist_clear (GTK_CLIST (list));
  
  for (int i = 0; i < MU_FLASH_NUM_COLUMNS; i++)
    t [i] = texts [i];
  
  for (int i = 0; i < numAddresses; i++)
  {
    sprintf (texts [0], "%d.%d.%d.%d", 
	     (unsigned) (((char*)(&addresses [i])) [0] & 0xff),
	     (unsigned) (((char*)(&addresses [i])) [1] & 0xff),
	     (unsigned) (((char*)(&addresses [i])) [2] & 0xff),
	     (unsigned) (((char*)(&addresses [i])) [3] & 0xff));


    sprintf (texts [1], "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
	     versions [i].ethernetAddress [0] & 0xff,
	     versions [i].ethernetAddress [1] & 0xff,
	     versions [i].ethernetAddress [2] & 0xff, 
	     versions [i].ethernetAddress [3] & 0xff,
	     versions [i].ethernetAddress [4] & 0xff,
	     versions [i].ethernetAddress [5] & 0xff);
    
    sprintf (texts [2], "%s", versions [i].version);
    
    gtk_clist_append (GTK_CLIST (list), t);
  }
  gdk_threads_leave ();
  exception: return 0;
}

int mu_FlashWidget::getVersions ()
{
  insist (this && numAddresses > 0);
  if (versions)
  {
    free (versions);
    versions = 0;
  }
  versions = (mu_FlashVersion*) malloc (sizeof (mu_FlashVersion) * numAddresses);
  insist (versions);
  
  for (int i = 0; i < numAddresses; i++)
  {
    mt_Component::printFormat ("getting version string for %d.%d.%d.%d", 
			       (unsigned) (((char*)(&addresses [i])) [0] & 0xff),
			       (unsigned) (((char*)(&addresses [i])) [1] & 0xff),
			       (unsigned) (((char*)(&addresses [i])) [2] & 0xff),
			       (unsigned) (((char*)(&addresses [i])) [3] & 0xff));

    if (!testClient.open (addresses [i]) ||
	!testClient.getVersionString (versions [i].version) ||
	!testClient.getEthernetAddress (versions [i].ethernetAddress))
      return 0;
  }
  return 1;
  
  exception: return 0;
}
