#include "insist.h"
#include "mu_MainWidget.h"
#include "gnome.h"

extern mu_MainWidget*mu_mainWidget;
extern char*mu_workingDirectory;
extern int mu_Cluster_colors [MU_CLUSTER_NUM_COLORS];

#define MU_FIRST_TOGGLE 4

static void mu_MainWidget_findFindButton_clicked (GtkWidget*w, mu_MainWidget*mainWidget)
{
  mainWidget->find ();
}

static void mu_MainWidget_findCloseButton_clicked (GtkWidget*w, mu_MainWidget*mainWidget)
{
  mainWidget->hideFind ();
}

static gboolean mu_MainWidget_find_delete (GtkWidget*w, GdkEvent*e, mu_MainWidget*mainWidget)
{
  mainWidget->hideFind ();
  return 1;
}

static gboolean mu_MainWidget_find_destroy (GtkWidget*w, mu_MainWidget*mainWidget)
{
  mainWidget->hideFind ();
  return 1;
}

static gboolean mu_MainWidget_delete (GtkWidget*w, GdkEvent*e, mu_MainWidget*widget)
{
  return 0;
}

static gboolean mu_MainWidget_destroy (GtkWidget*w, mu_MainWidget*widget)
{
  gtk_main_quit ();
  exception: return 0;
}

static void mu_MainWidget_build (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->exploreWidget);
  mu_mainWidget->exploreWidget->show ();
  exception:;
}

static void mu_MainWidget_preferences (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->optionsWidget);
  mu_mainWidget->optionsWidget->show ();
  exception:;
}

static void mu_MainWidget_find (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget);
  mu_mainWidget->showFind ();
  exception:;
}

static void mu_MainWidget_tree (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->cluster);
  mu_mainWidget->cluster->tree ();
  exception:;
}

static void mu_MainWidget_restore (gpointer data, guint action, GtkWidget*widget)
{
  FILE*fp;
  char s [M3_DISPLAY_STRING_LENGTH];
  
  insist (mu_mainWidget && mu_mainWidget->cluster);

  sprintf (s, "%s/%s", mu_workingDirectory, "mute.state");
  
  if ((fp = fopen (s, "r")))
  {
    mu_mainWidget->cluster->fromFile (fp);
    fclose (fp);
  }

  exception:;
}

static void mu_MainWidget_save (gpointer data, guint action, GtkWidget*widget)
{
  FILE*fp;
  char s [M3_DISPLAY_STRING_LENGTH];
  
  insist (mu_mainWidget && mu_mainWidget->cluster);

  sprintf (s, "%s/%s", mu_workingDirectory, "mute.state");
  
  if ((fp = fopen (s, "w+")))
  {
    mu_mainWidget->cluster->toFile (fp);
    fclose (fp);
  }  
 
  exception:;
}

static void mu_MainWidget_clearCounters (gpointer data, guint action, GtkWidget*widget)
{
  int r;
  insist (mu_mainWidget && mu_mainWidget->cluster);
  r = mu_mainWidget->cluster->clearCounters ();
  insist (r);
  exception:;
}

static void mu_MainWidget_stopSnmpTraps (gpointer data, guint action, GtkWidget*widget)
{
  int r;
  insist (mu_mainWidget && mu_mainWidget->cluster);
  r = mu_mainWidget->cluster->stopSnmpTraps ();
  insist (r);
  exception:;
}


static void mu_MainWidget_refresh (gpointer data, guint action, GtkWidget*widget);

static void mu_MainWidget_events (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->eventWidget);
  mu_mainWidget->eventWidget->show ();
  exception:;
}

static void mu_MainWidget_ping (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->pingWidget);
  mu_mainWidget->pingWidget->show ();
  exception:;
}
static void mu_MainWidget_firmware (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->flashWidget);
  mu_mainWidget->flashWidget->show ();
  exception:;
}

static void mu_MainWidget_packets (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->packetWidget);
  mu_mainWidget->packetWidget->show ();

  exception:;
}

static void mu_MainWidget_web (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->webWidget);
  mu_mainWidget->webWidget->show ();

  exception:;
}

static void mu_MainWidget_routes (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->routesWidget);
  mu_mainWidget->routesWidget->show ();

  exception:;
}

static void mu_MainWidget_legend (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->legendWidget);
  mu_mainWidget->legendWidget->show ();

  exception:;
}

static void mu_MainWidget_change (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->changeWidget);
  mu_mainWidget->changeWidget->show ();

  exception:;
}

void mu_MainWidget_insides (gpointer data, guint action, GtkWidget*widget);
void mu_MainWidget_hosts (gpointer data, guint action, GtkWidget*widget);
void mu_MainWidget_cables (gpointer data, guint action, GtkWidget*widget);
void mu_MainWidget_xpoints (gpointer data, guint action, GtkWidget*widget);
void mu_MainWidget_goodCrcs (gpointer data, guint action, GtkWidget*widget);
void mu_MainWidget_badCrcs (gpointer data, guint action, GtkWidget*widget);
void mu_MainWidget_temperatures (gpointer data, guint action, GtkWidget*widget);

static GnomeUIInfo mu_MainWidget_fileMenuInfo [] =
{
  GNOMEUIINFO_ITEM_NONE (N_("Build..."), NULL, mu_MainWidget_build),
  GNOMEUIINFO_ITEM_NONE (N_("Save Positions"), NULL, mu_MainWidget_save),
  GNOMEUIINFO_ITEM_NONE (N_("Restore Positions"), NULL, mu_MainWidget_restore),

  GNOMEUIINFO_MENU_PREFERENCES_ITEM (mu_MainWidget_preferences, NULL),
  GNOMEUIINFO_SEPARATOR,
  GNOMEUIINFO_MENU_EXIT_ITEM (mu_MainWidget_destroy, 0),
  GNOMEUIINFO_END
};

static GnomeUIInfo mu_MainWidget_editMenuInfo [] =
{
  GNOMEUIINFO_MENU_FIND_ITEM (mu_MainWidget_find, NULL),
  GNOMEUIINFO_END
};

static GnomeUIInfo mu_MainWidget_viewMenuInfo [] =
{
  GNOMEUIINFO_ITEM_NONE (N_("Tree Up"), NULL, mu_MainWidget_tree),
  GNOMEUIINFO_ITEM_NONE (N_("Update Counters"), NULL, mu_MainWidget_refresh),
  GNOMEUIINFO_ITEM_NONE (N_("Reset Switches"), NULL, mu_MainWidget_clearCounters),
  GNOMEUIINFO_ITEM_NONE (N_("Stop SNMP Traps"), NULL, mu_MainWidget_stopSnmpTraps),
  GNOMEUIINFO_TOGGLEITEM (N_("Show Insides"), NULL, mu_MainWidget_insides, NULL),
  GNOMEUIINFO_TOGGLEITEM (N_("Show Hosts"), NULL, mu_MainWidget_hosts, NULL),
  GNOMEUIINFO_TOGGLEITEM (N_("Show Cables"), NULL, mu_MainWidget_cables, NULL),
  GNOMEUIINFO_TOGGLEITEM (N_("Show Crosspoints"), NULL, mu_MainWidget_xpoints, NULL),
  GNOMEUIINFO_TOGGLEITEM (N_("Show Good Packets"), NULL, mu_MainWidget_goodCrcs, NULL),
  GNOMEUIINFO_TOGGLEITEM (N_("Show Bad Packets"), NULL, mu_MainWidget_badCrcs, NULL),
  GNOMEUIINFO_TOGGLEITEM (N_("Show Temperatures"), NULL, mu_MainWidget_temperatures, NULL),
  GNOMEUIINFO_END
};


void mu_MainWidget_insides (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->cluster);
  mu_mainWidget->cluster->showInside (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 0].widget)->active);
  exception:;
}

void mu_MainWidget_hosts (gpointer data, guint action, GtkWidget*widget)
{
  if (mu_mainWidget && mu_mainWidget->cluster)
    mu_mainWidget->cluster->showHosts (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 1].widget)->active);
  
  exception:;
}

void mu_MainWidget_cables (gpointer data, guint action, GtkWidget*widget)
{
  if (mu_mainWidget && mu_mainWidget->cluster)
    mu_mainWidget->cluster->showCables (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 2].widget)->active);
  exception:;
}
void mu_MainWidget_xpoints (gpointer data, guint action, GtkWidget*widget)
{
  if (mu_mainWidget && mu_mainWidget->cluster)
    mu_mainWidget->cluster->showXpoints (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 3].widget)->active);
  exception:;
}

void mu_MainWidget_goodCrcs (gpointer data, guint action, GtkWidget*widget)
{
  if (mu_mainWidget && mu_mainWidget->cluster)
  {
    if (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 4].widget)->active)
    {
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 5].widget), 0);
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 6].widget), 0);
      mu_mainWidget->cluster->showCrcs (1);
    }
    else
      mu_mainWidget->cluster->clearLinkColors ();
  }
  exception:;
}

void mu_MainWidget_badCrcs (gpointer data, guint action, GtkWidget*widget)
{
  if (mu_mainWidget && mu_mainWidget->cluster)
  {
    if (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 5].widget)->active)
    {
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 4].widget), 0);
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 6].widget), 0);
      mu_mainWidget->cluster->showCrcs (0);
    }
    else
      mu_mainWidget->cluster->clearLinkColors ();
  }
  exception:;
}

void mu_MainWidget_temperatures (gpointer data, guint action, GtkWidget*widget)
{
  if (mu_mainWidget && mu_mainWidget->cluster)
  {
    if (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 6].widget)->active)
    {
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 4].widget), 0);
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 5].widget), 0);
    }
    mu_mainWidget->cluster->showTemperatures (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 6].widget)->active);
  }
  exception:;
}

static GnomeUIInfo mu_MainWidget_windowsMenuInfo [] =
{
  GNOMEUIINFO_ITEM_NONE (N_("Events..."), NULL, mu_MainWidget_events),
  GNOMEUIINFO_ITEM_NONE (N_("Ping..."), NULL, mu_MainWidget_ping),
  GNOMEUIINFO_ITEM_NONE (N_("Firmware..."), NULL, mu_MainWidget_firmware),
  GNOMEUIINFO_ITEM_NONE (N_("Counters..."), NULL, mu_MainWidget_packets),
  GNOMEUIINFO_ITEM_NONE (N_("Info..."), NULL, mu_MainWidget_web),
  GNOMEUIINFO_ITEM_NONE (N_("Routes..."), NULL, mu_MainWidget_routes),
  GNOMEUIINFO_ITEM_NONE (N_("Legend..."), NULL, mu_MainWidget_legend),
  GNOMEUIINFO_ITEM_NONE (N_("Control..."), NULL, mu_MainWidget_change),
  GNOMEUIINFO_END
};

static GnomeUIInfo mu_MainWidget_menubarInfo [] =
{
  GNOMEUIINFO_MENU_FILE_TREE (mu_MainWidget_fileMenuInfo),
  GNOMEUIINFO_MENU_EDIT_TREE (mu_MainWidget_editMenuInfo),
  GNOMEUIINFO_MENU_VIEW_TREE (mu_MainWidget_viewMenuInfo),
  GNOMEUIINFO_SUBTREE (N_("Windows"), mu_MainWidget_windowsMenuInfo),
  GNOMEUIINFO_END
};

mu_MainWidget::mu_MainWidget (char*name)
{
  widget = gnome_app_new (name, MU_TITLE);
  insist (widget);

  gnome_app_create_menus (GNOME_APP (widget), mu_MainWidget_menubarInfo);
  
  gtk_signal_connect (GTK_OBJECT (widget), "delete_event", GTK_SIGNAL_FUNC (mu_MainWidget_delete), this);
  gtk_signal_connect (GTK_OBJECT (widget), "destroy", GTK_SIGNAL_FUNC (mu_MainWidget_destroy), this);

  GtkWidget*box, *canvas, *hbox, *button;
  
  findDialog = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  insist (findDialog);  

  box = gtk_vbox_new (0, MU_SPACING);
  insist (box);
  gtk_container_add (GTK_CONTAINER (findDialog), box);
  gtk_widget_show (box);
  findEntry = gtk_entry_new_with_max_length (M3_DISPLAY_STRING_LENGTH);
  insist (findEntry);
  gtk_widget_show (findEntry);
  gtk_box_pack_start (GTK_BOX (box), findEntry, 0, 0, MU_SPACING);
  gtk_signal_connect (GTK_OBJECT (findEntry), "activate", GTK_SIGNAL_FUNC (mu_MainWidget_findFindButton_clicked), this);

  hbox = gtk_hbox_new (0, MU_SPACING);
  insist (hbox);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (box), hbox, 0, 0, MU_SPACING);

  button = gtk_button_new_with_label ("Find");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_MainWidget_findFindButton_clicked), this);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, 0, 0, MU_SPACING);

  button = gtk_button_new_with_label ("Close");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_MainWidget_findCloseButton_clicked), this);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, 0, 0, MU_SPACING);

  box = gtk_vbox_new (0, MU_SPACING);
  insist (box);
  gtk_widget_show (box);

  gnome_app_set_contents (GNOME_APP (widget), box);
  gtk_window_set_policy (GTK_WINDOW (widget), 1, 1, 0);
  gtk_container_set_border_width (GTK_CONTAINER (box), MU_SPACING);
  gtk_widget_realize (widget);

  gtk_window_set_title (GTK_WINDOW (findDialog), MU_TITLE " - Find");
  gtk_window_set_policy (GTK_WINDOW (findDialog), 1, 1, 0);
  
  exploreWidget = new mu_ExploreWidget ();
  insist (exploreWidget);

  optionsWidget = new mu_OptionsWidget ();
  insist (optionsWidget);

  eventWidget = new mu_EventWidget ();
  insist (eventWidget);

  pingWidget = new mu_PingWidget ();
  insist (pingWidget);

  packetWidget = new mu_PacketWidget ();
  insist (packetWidget);

  flashWidget = new mu_FlashWidget ();
  insist (flashWidget);

  webWidget = new mu_WebWidget ();
  insist (webWidget);

  routesWidget = new mu_RoutesWidget ();
  insist (routesWidget);
  
  legendWidget = new mu_LegendWidget (MU_CLUSTER_NUM_COLORS, mu_Cluster_colors);
  insist (legendWidget);

  changeWidget = new mu_ChangeWidget ();
  insist (changeWidget);
 
  cluster = new mu_Cluster (widget);
  insist (cluster);
  gtk_box_pack_start (GTK_BOX (box), cluster->getWidget (), 1, 1, 4);

  cluster->show ();
  
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 1].widget), 1);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 2].widget), 1);
  exception:;
}

void mu_MainWidget::showFind ()
{
  insist (this && findDialog && findEntry);
  gtk_widget_show (findDialog);
  gtk_widget_grab_focus (findEntry);
  exception:;
}

void mu_MainWidget::hideFind ()
{
  insist (this && findDialog);
  gtk_widget_hide (findDialog);
  exception:;
}

void mu_MainWidget::find ()
{
  insist (this && findEntry && cluster);
  cluster->select (gtk_entry_get_text (GTK_ENTRY (findEntry)));
  exception:;
}

void mu_MainWidget_refresh (gpointer data, guint action, GtkWidget*widget)
{
  insist (mu_mainWidget && mu_mainWidget->cluster);
  mu_mainWidget->cluster->refresh ();

  mu_mainWidget->checkStates ();
  
  exception:;
}

void mu_MainWidget::checkStates ()
{
  insist (this && cluster);

  mu_mainWidget->cluster->showInside (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 0].widget)->active);
  mu_mainWidget->cluster->showHosts (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 1].widget)->active);
  mu_mainWidget->cluster->showCables (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 2].widget)->active);
  mu_mainWidget->cluster->showXpoints (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 3].widget)->active);
  mu_mainWidget->cluster->showTemperatures (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 6].widget)->active);
  
  if (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 4].widget)->active)
    mu_mainWidget->cluster->showCrcs (1);
  else if (GTK_CHECK_MENU_ITEM (mu_MainWidget_viewMenuInfo [MU_FIRST_TOGGLE + 5].widget)->active)
    mu_mainWidget->cluster->showCrcs (0);

  exception:;
}
