#include "gnome.h"
#include "insist.h"
#include "mu_OptionsWidget.h"
#include "mu_MainWidget.h"

extern char*mu_workingDirectory;
extern char*mu_firmwareDirectory;
extern mu_MainWidget*mu_mainWidget;
extern int mu_eventMask [2];
extern int mu_eventHistory;
extern int mu_maxThreads;
extern int mu_logEvents;

#define MU_OPTIONS_MAX_HISTORY_TEXT 30

static int mu_mask_order [M3_NUM_TRAPS] =
{
  10, 11, 14, 30, 12, 13, 18, 7, 8, 9, 16, 15, 17, 32, 27, 28, 31, 19, 20, 24, 21, 22, 25, 26, 29, 5, 3, 4, 23, 6, 1, 2, 0
};

static void mu_OptionsWidget_setAll (GtkWidget*w, mu_OptionsWidget*widget)
{
  insist (widget);
  widget->setEventToggles (1);
  exception:;
}

static void mu_OptionsWidget_clearAll (GtkWidget*w, mu_OptionsWidget*widget)
{
  insist (widget);
  widget->setEventToggles (0);
  exception:;
}

static void mu_OptionsWidget_applyButton_clicked (GtkWidget*w, mu_OptionsWidget*widget)
{
  widget->writeOptions ();
  if (mu_mainWidget && mu_mainWidget->exploreWidget)
    mu_mainWidget->exploreWidget->setStates ();

  if (mu_mainWidget && mu_mainWidget->flashWidget)
    mu_mainWidget->flashWidget->updateFilenames ();

  if (mu_mainWidget && mu_mainWidget->eventWidget)
    mu_mainWidget->eventWidget->setNotify (1);

  mu_logEvents = gtk_toggle_button_get_active ((GtkToggleButton*) widget->logEventsToggle);
}

static void mu_OptionsWidget_closeButton_clicked (GtkWidget*w, mu_OptionsWidget*widget)
{
  mu_OptionsWidget_applyButton_clicked (w, widget);
  widget->hide ();
}

static void mu_OptionsWidget_cancelButton_clicked (GtkWidget*w, mu_OptionsWidget*widget)
{
  widget->readOptions ();
  widget->hide ();
}

static gboolean mu_OptionsWidget_delete (GtkWidget*w, GdkEvent*e, mu_OptionsWidget*widget)
{
  widget->readOptions ();
  widget->hide ();
  return 1;
}

static gboolean mu_OptionsWidget_destroy (GtkWidget*w, mu_OptionsWidget*widget)
{
  widget->readOptions ();
  widget->hide ();
  return 0;
}

static void mu_OptionsWidget_workingDirectoryButton_clicked (GtkWidget*w, mu_OptionsWidget*optionsWidget)
{
  optionsWidget->showWorkingDirectorySelection ();
}

static void mu_OptionsWidget_workingDirectorySelectionOKButton_clicked (GtkWidget*w, mu_OptionsWidget*optionsWidget)
{
  optionsWidget->hideWorkingDirectorySelection ();
  optionsWidget->setWorkingDirectory ();
}

static void mu_OptionsWidget_workingDirectorySelectionCancelButton_clicked (GtkWidget*w, mu_OptionsWidget*optionsWidget)
{
  optionsWidget->hideWorkingDirectorySelection ();
}

static gboolean mu_OptionsWidget_workingDirectorySelection_delete (GtkWidget*w, GdkEvent*e, mu_OptionsWidget*optionsWidget)
{
  optionsWidget->hideWorkingDirectorySelection ();
  return 1;
}

static gboolean mu_OptionsWidget_workingDirectorySelection_destroy (GtkWidget*w, mu_OptionsWidget*optionsWidget)
{
  optionsWidget->hideWorkingDirectorySelection ();
  return 1;
}

void mu_OptionsWidget::setWorkingDirectory ()
{
  insist (this && workingDirectorySelection && workingDirectoryEntry);
  gtk_entry_set_text (GTK_ENTRY (workingDirectoryEntry), gtk_file_selection_get_filename (GTK_FILE_SELECTION (workingDirectorySelection)));
  exception: return;
}

static void mu_OptionsWidget_firmwareDirectoryButton_clicked (GtkWidget*w, mu_OptionsWidget*optionsWidget)
{
  optionsWidget->showFirmwareDirectorySelection ();
}


static void mu_OptionsWidget_firmwareDirectorySelectionOKButton_clicked (GtkWidget*w, mu_OptionsWidget*optionsWidget)
{
  optionsWidget->hideFirmwareDirectorySelection ();
  optionsWidget->setFirmwareDirectory ();
}

static void mu_OptionsWidget_firmwareDirectorySelectionCancelButton_clicked (GtkWidget*w, mu_OptionsWidget*optionsWidget)
{
  optionsWidget->hideFirmwareDirectorySelection ();
}

static gboolean mu_OptionsWidget_firmwareDirectorySelection_delete (GtkWidget*w, GdkEvent*e, mu_OptionsWidget*optionsWidget)
{
  optionsWidget->hideFirmwareDirectorySelection ();
  return 1;
}

static gboolean mu_OptionsWidget_firmwareDirectorySelection_destroy (GtkWidget*w, mu_OptionsWidget*optionsWidget)
{
  optionsWidget->hideFirmwareDirectorySelection ();
  return 1;
}

void mu_OptionsWidget::setFirmwareDirectory ()
{
  insist (this && firmwareDirectorySelection && firmwareDirectoryEntry);
  gtk_entry_set_text (GTK_ENTRY (firmwareDirectoryEntry), gtk_file_selection_get_filename (GTK_FILE_SELECTION (firmwareDirectorySelection)));
  exception: return;
}

mu_OptionsWidget::mu_OptionsWidget ()
{
  GtkWidget*notebook;
  widget = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  insist (widget);  

  gtk_window_set_title (GTK_WINDOW (widget), MU_TITLE " - Preferences");
  gtk_window_set_policy (GTK_WINDOW (widget), 1, 1, 0);
  
  GtkWidget*label, *button, *applyButton, *closeButton, *cancelButton, *table;
  GtkWidget*box;
  
  box = gtk_vbox_new (0, MU_SPACING);
  insist (box);
  gtk_container_add (GTK_CONTAINER (widget), box);
  gtk_widget_show (box);
  
  notebook = gtk_notebook_new ();
  insist (notebook);
  gtk_widget_show (notebook);
  
  gtk_box_pack_start (GTK_BOX (box), notebook, 0, 0, MU_SPACING);

  table = gtk_table_new (4, 3, 0);
  insist (table);
  gtk_widget_show (table);
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), table, gtk_label_new ("Options"));
  
  gtk_signal_connect (GTK_OBJECT (widget), "delete_event", GTK_SIGNAL_FUNC (mu_OptionsWidget_delete), this);
  gtk_signal_connect (GTK_OBJECT (widget), "destroy", GTK_SIGNAL_FUNC (mu_OptionsWidget_destroy), this);

  workingDirectorySelection = gtk_file_selection_new ("Working Directory");
  gtk_signal_connect (GTK_OBJECT (workingDirectorySelection), "delete_event", GTK_SIGNAL_FUNC (mu_OptionsWidget_workingDirectorySelection_delete), this);
  gtk_signal_connect (GTK_OBJECT (workingDirectorySelection), "destroy", GTK_SIGNAL_FUNC (mu_OptionsWidget_workingDirectorySelection_destroy), this);
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (workingDirectorySelection)->ok_button), "clicked",
		      GTK_SIGNAL_FUNC (mu_OptionsWidget_workingDirectorySelectionOKButton_clicked), this);
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (workingDirectorySelection)->cancel_button), "clicked",
		      GTK_SIGNAL_FUNC (mu_OptionsWidget_workingDirectorySelectionCancelButton_clicked), this);
  

  firmwareDirectorySelection = gtk_file_selection_new ("Firmware Directory");
  gtk_signal_connect (GTK_OBJECT (firmwareDirectorySelection), "delete_event", GTK_SIGNAL_FUNC (mu_OptionsWidget_firmwareDirectorySelection_delete), this);
  gtk_signal_connect (GTK_OBJECT (firmwareDirectorySelection), "destroy", GTK_SIGNAL_FUNC (mu_OptionsWidget_firmwareDirectorySelection_destroy), this);
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (firmwareDirectorySelection)->ok_button), "clicked",
		      GTK_SIGNAL_FUNC (mu_OptionsWidget_firmwareDirectorySelectionOKButton_clicked), this);
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (firmwareDirectorySelection)->cancel_button), "clicked",
		      GTK_SIGNAL_FUNC (mu_OptionsWidget_firmwareDirectorySelectionCancelButton_clicked), this);

  label = gtk_label_new ("Working Directory");
  insist (label);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1, (GtkAttachOptions) 0, (GtkAttachOptions) 0, 0, 0);
  gtk_widget_show (label);
  
  workingDirectoryEntry = gtk_entry_new ();
  gtk_widget_show (workingDirectoryEntry);
  gtk_table_attach (GTK_TABLE (table), workingDirectoryEntry, 1, 2, 0, 1, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), (GtkAttachOptions) 0, 0, 0);

  button = gtk_button_new_with_label (" ... ");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_OptionsWidget_workingDirectoryButton_clicked), this);
  gtk_widget_show (button);
  gtk_table_attach (GTK_TABLE (table), button, 2, 3, 0, 1, (GtkAttachOptions) 0, (GtkAttachOptions) 0, 0, 0);

  label = gtk_label_new ("Firmware Directory");
  insist (label);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2, (GtkAttachOptions) 0, (GtkAttachOptions) 0, 0, 0);
  gtk_widget_show (label);
  
  firmwareDirectoryEntry = gtk_entry_new ();
  gtk_widget_show (firmwareDirectoryEntry);

  gtk_table_attach (GTK_TABLE (table), firmwareDirectoryEntry, 1, 2, 1, 2, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), (GtkAttachOptions) 0, 0, 0);

  button = gtk_button_new_with_label (" ... ");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_OptionsWidget_firmwareDirectoryButton_clicked), this);
  gtk_widget_show (button);
  gtk_table_attach (GTK_TABLE (table), button, 2, 3, 1, 2, (GtkAttachOptions) 0, (GtkAttachOptions) 0, 0, 0);

  label = gtk_label_new ("Max Threads:");
  insist (label);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 2, 3, (GtkAttachOptions) 0, (GtkAttachOptions) 0, 0, 0);
  gtk_widget_show (label);
  
  maxThreadsAdjustment = (GtkAdjustment*) gtk_adjustment_new (MU_DEFAULT_THREADS, MU_MIN_THREADS, MU_MAX_THREADS + 1, 1, 1, 1);
  insist (maxThreadsAdjustment);
  maxThreadsScale = gtk_hscale_new (maxThreadsAdjustment);
  insist (maxThreadsScale);
  gtk_scale_set_digits (GTK_SCALE (maxThreadsScale), 0);
  gtk_widget_show (maxThreadsScale);
  gtk_table_attach (GTK_TABLE (table), maxThreadsScale, 1, 3, 2, 3, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), (GtkAttachOptions) 0, 0, 0);

  label = gtk_label_new ("Event History:");
  insist (label);
  gtk_widget_show (label);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 3, 4, (GtkAttachOptions) 0, (GtkAttachOptions) 0, 0, 0);

  historyAdjustment = (GtkAdjustment*) gtk_adjustment_new (MU_DEFAULT_HISTORY, MU_MIN_HISTORY, MU_MAX_HISTORY + 1, 1, 1, 1);
  insist (historyAdjustment);
  historyScale = gtk_hscale_new (historyAdjustment);
  insist (historyScale);
  gtk_scale_set_digits (GTK_SCALE (historyScale), 0);
  gtk_widget_show (historyScale);
  gtk_table_attach (GTK_TABLE (table), historyScale, 1, 3, 3, 4, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), (GtkAttachOptions) 0, 0, 0);

  logEventsToggle = gtk_check_button_new_with_label ("Log Events");
  insist (logEventsToggle);
  gtk_widget_show (logEventsToggle);
  gtk_table_attach (GTK_TABLE (table), logEventsToggle, 0, 3, 4, 5, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), (GtkAttachOptions) 0, 0, 0);
  
  GtkWidget*b2;
  b2 = gtk_hbox_new (1, MU_SPACING);
  insist (b2);
  gtk_widget_show (b2);
  gtk_box_pack_start (GTK_BOX (box), b2, 0, 0, MU_SPACING);

  applyButton = gtk_button_new_with_label ("Apply");
  gtk_signal_connect (GTK_OBJECT (applyButton), "clicked", GTK_SIGNAL_FUNC (mu_OptionsWidget_applyButton_clicked), this);
  gtk_widget_show (applyButton);
  gtk_box_pack_start (GTK_BOX (b2), applyButton, 0, 0, MU_SPACING);

  closeButton = gtk_button_new_with_label ("OK");
  gtk_signal_connect (GTK_OBJECT (closeButton), "clicked", GTK_SIGNAL_FUNC (mu_OptionsWidget_closeButton_clicked), this);
  gtk_widget_show (closeButton);
  gtk_box_pack_start (GTK_BOX (b2), closeButton, 0, 0, MU_SPACING);
  
  cancelButton = gtk_button_new_with_label ("Cancel");
  gtk_signal_connect (GTK_OBJECT (cancelButton), "clicked", GTK_SIGNAL_FUNC (mu_OptionsWidget_cancelButton_clicked), this);
  gtk_widget_show (cancelButton); 

  gtk_box_pack_start (GTK_BOX (b2), cancelButton, 0, 0, MU_SPACING);

  GtkWidget*b3;
  b3 = gtk_vbox_new (0, 0);
  insist (b3);
  gtk_widget_show (b3);
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), b3, gtk_label_new ("Events"));

  for (int i = 0; i < M3_NUM_TRAPS; i++)
  {
    eventToggles [i] = gtk_check_button_new_with_label (m3_trap_names [i]);
    insist (eventToggles [i]);
  }

  GtkWidget*b4;
  b4 = gtk_hbox_new (0, 0);
  insist (b4);
  gtk_widget_show (b4);
  gtk_box_pack_start (GTK_BOX (b3), b4, 0, 0, MU_SPACING);

  GtkWidget*boxes [2];
  
  for (int i = 0; i < 2; i++)
  {
    boxes [i] = gtk_vbox_new (0, 0);
    insist (boxes [i]);
    gtk_widget_show (boxes [i]);
    gtk_box_pack_start (GTK_BOX (b4), boxes [i], 0, 0, 0);
  }

  for (int i = 0; i < M3_NUM_TRAPS; i++)
  {
    insist (mu_mask_order [i] >= 0 && mu_mask_order [i] < M3_NUM_TRAPS);
    insist (eventToggles [mu_mask_order [i]]);
    gtk_box_pack_start (GTK_BOX (boxes [i / ((M3_NUM_TRAPS + 1) / 2)]),  eventToggles [mu_mask_order [i]], 0, 0, 0);
    gtk_widget_show (eventToggles [mu_mask_order [i]]);
  }

  b4 = gtk_hbox_new (0, 0);
  insist (b4);
  gtk_widget_show (b4);
  gtk_box_pack_start (GTK_BOX (b3), b4, 0, 0, MU_SPACING);

  button = gtk_button_new_with_label ("Set all");
  insist (button);
  gtk_box_pack_start (GTK_BOX (b4), button, 0, 0, MU_SPACING);
  gtk_widget_show (button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_OptionsWidget_setAll), this);

  button = gtk_button_new_with_label ("Clear all");
  insist (button);
  gtk_box_pack_start (GTK_BOX (b4), button, 0, 0, MU_SPACING);
  gtk_widget_show (button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_OptionsWidget_clearAll), this);
  
  readOptions ();

  exception:;
}

void mu_OptionsWidget::writeOptions ()
{
  insist (this && workingDirectoryEntry && firmwareDirectoryEntry);
  insist (mu_mainWidget && mu_mainWidget->eventWidget);
  insist (maxThreadsAdjustment);
  
  mu_workingDirectory = gtk_entry_get_text (GTK_ENTRY (workingDirectoryEntry));
  mu_firmwareDirectory = gtk_entry_get_text (GTK_ENTRY (firmwareDirectoryEntry));

  mu_eventMask [0] = mu_eventMask [1] = 0;
  
  for (int i = 0; i < M3_NUM_TRAPS; i++)
  {
    insist (eventToggles [i]);

    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (eventToggles [i])))
      mu_eventMask [i / 32] |= (1 << (i % 32));
  }
  
  mu_eventHistory = (int) historyAdjustment->value;
  mu_maxThreads = (int) maxThreadsAdjustment->value;
  
  exception:;
}

void mu_OptionsWidget::readOptions ()
{
  insist (this && workingDirectoryEntry && firmwareDirectoryEntry && mu_firmwareDirectory && mu_workingDirectory);
  insist (maxThreadsAdjustment && historyAdjustment);

  gtk_entry_set_text (GTK_ENTRY (workingDirectoryEntry), mu_workingDirectory);
  gtk_file_selection_set_filename (GTK_FILE_SELECTION (workingDirectorySelection), mu_workingDirectory);

  gtk_entry_set_text (GTK_ENTRY (firmwareDirectoryEntry), mu_firmwareDirectory);
  gtk_file_selection_set_filename (GTK_FILE_SELECTION (firmwareDirectorySelection), mu_firmwareDirectory);

  for (int i = 0; i < M3_NUM_TRAPS; i++)
  {
    insist (eventToggles [i]);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (eventToggles [i]), mu_eventMask [i / 32] & (1 << (i % 32)));
  }

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (logEventsToggle), mu_logEvents);
  
  gtk_adjustment_set_value (historyAdjustment, mu_eventHistory);
  gtk_adjustment_set_value (maxThreadsAdjustment, mu_maxThreads);
  
  exception:;
}

int mu_OptionsWidget::setEventToggles (int state)
{
  insist (this);
  
  for (int i = 0; i < M3_NUM_TRAPS; i++)
  {
    insist (eventToggles [i]);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (eventToggles [i]), state);
  }
  return 1;
  exception: return 0;
}
