#include "mt_Host.h"
#include "mu_PacketWidget.h"
#include "mu_MainWidget.h"
#include "mu_ExploreWidget.h"
#include "insist.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/time.h>

extern mu_MainWidget*mu_mainWidget;
extern int mu_maxThreads;

static void mu_PacketWidget_start (GtkWidget*w, mu_PacketWidget*widget)
{
  widget->createThread ((mu_ThreadableFunction)&mu_PacketWidget::run, (void*)0);
}

static void mu_PacketWidget_stop (GtkWidget*w, mu_PacketWidget*widget)
{
  widget->stopThread (0);
}

static void mu_PacketWidget_close (GtkWidget*w, mu_PacketWidget*widget)
{
  widget->hide ();
}

static gboolean mu_PacketWidget_delete (GtkWidget*w, GdkEvent*e, mu_PacketWidget*widget)
{
  widget->hide ();
  return 1;
}

static gboolean mu_PacketWidget_destroy (GtkWidget*w, mu_PacketWidget*widget)
{
  widget->hide ();
  return 0;
}

mu_PacketWidget::mu_PacketWidget ()
{
  GtkWidget*box;
  GtkWidget*hboxes [3];
  GtkWidget*vbox, *table;
  GtkWidget*frame, *label, *button;
  GtkAdjustment*adjustment;
  GSList*group;

  widget = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  insist (widget);

  gtk_window_set_title (GTK_WINDOW (widget), MU_TITLE " - Packet Counters");
  gtk_window_set_policy (GTK_WINDOW (widget), 1, 1, 0);

  box = gtk_vbox_new (0, MU_SPACING);
  insist (box);
  gtk_container_add (GTK_CONTAINER (widget), box);
  gtk_widget_show (box);

  for (int i = 0; i < 3; i++)
  {
    hboxes [i] = gtk_hbox_new (1, MU_SPACING);
    insist (hboxes [i]);
    gtk_box_pack_start (GTK_BOX (box), hboxes [i], 1, 0, 0);
    gtk_widget_show (hboxes [i]);  
  }

  frame = gtk_frame_new ("Good Crcs");
  insist (frame);
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (hboxes [0]), frame, 1, 1, 0);

  vbox = gtk_vbox_new (1, MU_SPACING);
  insist (vbox);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (frame), vbox);

  goodTotalLabel = gtk_label_new ("Total Packets");
  //gtk_widget_show (goodTotalLabel);
  gtk_label_set_justify (GTK_LABEL (goodTotalLabel), GTK_JUSTIFY_LEFT);
  gtk_box_pack_start (GTK_BOX (vbox), goodTotalLabel, 1, 1, 0);

  goodTotalRateLabel = gtk_label_new ("Total Packets/sec");
  gtk_widget_show (goodTotalRateLabel);
  gtk_label_set_justify (GTK_LABEL (goodTotalRateLabel), GTK_JUSTIFY_LEFT);
  gtk_box_pack_start (GTK_BOX (vbox), goodTotalRateLabel, 0, 0, 1);

  goodMaxRateLabel = gtk_label_new ("Max Packets/sec");
  gtk_widget_show (goodMaxRateLabel);
  gtk_label_set_justify (GTK_LABEL (goodMaxRateLabel), GTK_JUSTIFY_LEFT);
  gtk_box_pack_start (GTK_BOX (vbox), goodMaxRateLabel, 1, 1, 0);

  group = 0;
  goodRadio = gtk_radio_button_new_with_label (group, "Colors");
  insist (goodRadio);
  gtk_widget_show (goodRadio);
  gtk_box_pack_start (GTK_BOX (vbox), goodRadio, 1, 0, 0);

  frame = gtk_frame_new ("Bad Crcs");
  insist (frame);
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (hboxes [0]), frame, 1, 1, 0);
  vbox = gtk_vbox_new (1, MU_SPACING);
  insist (vbox);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (frame), vbox);

  badTotalLabel = gtk_label_new ("Total Packets");
  //gtk_widget_show (badTotalLabel);
  gtk_label_set_justify (GTK_LABEL (badTotalLabel), GTK_JUSTIFY_LEFT);
  gtk_box_pack_start (GTK_BOX (vbox), badTotalLabel, 1, 1, 0);

  badTotalRateLabel = gtk_label_new ("Total Packets/sec");
  gtk_widget_show (badTotalRateLabel);
  gtk_label_set_justify (GTK_LABEL (badTotalRateLabel), GTK_JUSTIFY_LEFT);
  gtk_box_pack_start (GTK_BOX (vbox), badTotalRateLabel, 1, 1, 0);

  badMaxRateLabel = gtk_label_new ("Max Packets/sec");
  gtk_widget_show (badMaxRateLabel);
  gtk_label_set_justify (GTK_LABEL (badMaxRateLabel), GTK_JUSTIFY_LEFT);
  gtk_box_pack_start (GTK_BOX (vbox), badMaxRateLabel, 1, 1, 0);

  group = gtk_radio_button_group (GTK_RADIO_BUTTON (goodRadio));
  badRadio = gtk_radio_button_new_with_label (group, "Colors");
  insist (badRadio);
  gtk_widget_show (badRadio);
  gtk_box_pack_start (GTK_BOX (vbox), badRadio, 1, 0, 0);

  vbox = gtk_vbox_new (0, MU_SPACING);
  insist (vbox);
  gtk_widget_show (vbox);
  gtk_box_pack_start (GTK_BOX (hboxes [1]), vbox, 1, 1, 0);

  progressLabel = gtk_label_new ("   ");
  insist (progressLabel);
  gtk_widget_show (progressLabel);
  gtk_box_pack_start (GTK_BOX (vbox), progressLabel, 1, 1, 0);

  progressBar = gtk_progress_bar_new ();
  insist (progressBar);
  gtk_widget_show (progressBar);
  gtk_box_pack_start (GTK_BOX (vbox), progressBar, 1, 1, 0);

  startButton = gtk_button_new_with_label ("Start");
  gtk_signal_connect (GTK_OBJECT (startButton), "clicked", GTK_SIGNAL_FUNC (mu_PacketWidget_start), this);
  gtk_widget_show (startButton);
  gtk_box_pack_start (GTK_BOX (hboxes [2]), startButton, 0, 0, MU_SPACING);

  stopButton = gtk_button_new_with_label ("Stop");
  gtk_signal_connect (GTK_OBJECT (stopButton), "clicked", GTK_SIGNAL_FUNC (mu_PacketWidget_stop), this);
  gtk_widget_show (stopButton);
  gtk_box_pack_start (GTK_BOX (hboxes [2]), stopButton, 0, 0, MU_SPACING);
  gtk_widget_set_sensitive (stopButton, 0);

  button = gtk_button_new_with_label ("Close");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_PacketWidget_close), this);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hboxes [2]), button, 0, 0, MU_SPACING);

  exception:;
}


void mu_PacketWidget::run (void*data)
{
  int*addresses = 0;
  m3_XbarTable*table;
  mt_Graph*graph;

  insist (mu_mainWidget && mu_mainWidget->exploreWidget);
  insist (mu_mainWidget && mu_mainWidget->cluster);
  
  table = mu_mainWidget->exploreWidget->xbarTable;
  graph = &mu_mainWidget->exploreWidget->graph;
  insist (table && graph);
  
  if (!(numAddresses = mu_mainWidget->exploreWidget->getAddresses (&addresses)))
    return;

  insist (addresses);

  gdk_threads_enter ();
  gtk_progress_bar_set_bar_style (GTK_PROGRESS_BAR (progressBar), GTK_PROGRESS_DISCRETE);
  if (numAddresses > 1)
    gtk_progress_bar_set_discrete_blocks (GTK_PROGRESS_BAR (progressBar), numAddresses);
  gtk_progress_bar_update (GTK_PROGRESS_BAR (progressBar), 0);
  gtk_widget_set_sensitive (startButton, 0);
  gtk_widget_set_sensitive (stopButton, 1);
  gdk_threads_leave ();

  int e ;
  e = 0;
  
  struct timeval tv, lasttv;
  gettimeofday (&lasttv, 0);

  goodTotalTotal = badTotalTotal = 0;


  int numHelpers;
  numHelpers = mu_maxThreads;

  if (numAddresses < numHelpers)
    numHelpers = numAddresses;
  
  {
    mu_PacketHelper helpers [numHelpers];
  
    goodTotalTotal = badTotalTotal = 0;
    
    while (!isStopped ())
    {
      double d;
      double goodMaxRate, goodTotalRate, badMaxRate, badTotalRate;
      goodTotal = badTotal = goodMax = badMax = badMin = goodMin = 0;
    
      numSeen = 0;
    
      for (int i = 0; i < numAddresses; i++)
      {  
	helpers [i % numHelpers].join ();  
	helpers [i % numHelpers].createThread ((mu_ThreadableFunction) &mu_PacketHelper::run, (void*) addresses [i]);
      }    
      
      for (int i = 0; i < numHelpers; i++)
	helpers [i].join ();
    
      gettimeofday (&tv, 0);
      d = tv.tv_sec - lasttv.tv_sec + (tv.tv_usec - lasttv.tv_usec) / (1000 * 1000.0);
      if (d > 0)
      {
	goodMaxRate = goodMax / d;
	badMaxRate = badMax / d;
	goodTotalRate = goodTotal / d;
	badTotalRate = badTotal / d;

	gdk_threads_enter ();
	setLabel (goodTotalLabel, "Total Packets", (double) goodTotalTotal);
	setLabel (goodTotalRateLabel, "Total Packets/sec", goodTotalRate);
	setLabel (goodMaxRateLabel, "Max Packets/sec", goodMaxRate);
	setLabel (badTotalLabel, "Total Packets", (double) badTotalTotal);
	setLabel (badTotalRateLabel, "Total Packets/sec", badTotalRate);
	setLabel (badMaxRateLabel, "Max Packets/sec", badMaxRate);
	mu_mainWidget->cluster->drawPacketCounters (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (goodRadio)), (int) goodMax, (int) goodMin);
	gdk_threads_leave ();
      }
      lasttv = tv;
      sleep (1);
    }
  }
  
  gdk_threads_enter ();
  gtk_widget_set_sensitive (startButton, 1);
  gtk_widget_set_sensitive (stopButton, 0);
  gdk_threads_leave ();

  exception:;
}

int mu_PacketWidget::setLabel (GtkWidget*label, char*s, double d)
{
  insist (this && label && s);
  char buffer [100];
  insist (strlen (s) < 50);
  sprintf (buffer, "%s: %.0lf", s, d);
  gtk_label_set_text (GTK_LABEL (label), buffer);
  return 1;
  exception: return 0;
}

void mu_PacketWidget::eatCounters (int address, int numCounters, mu_int64*good, mu_int64*bad)
{
  char s [45];

  insist (this && numCounters && good && bad);
  insist (numSeen < numAddresses);
  
  insist (mu_mainWidget && mu_mainWidget->exploreWidget);
  insist (mu_mainWidget && mu_mainWidget->cluster);

  mu_mainWidget->cluster->eatPacketCounters (address, 1, numCounters, good, &goodTotalTotal, &goodTotal, &goodMax, &goodMin);
  mu_mainWidget->cluster->eatPacketCounters (address, 0, numCounters, bad, &badTotalTotal, &badTotal, &badMax, &goodMin);

  sprintf (s, "%d:%d:%d:%d",
	   (unsigned) (((char*)(&address)) [0] & 0xff),
	   (unsigned) (((char*)(&address)) [1] & 0xff),
	   (unsigned) (((char*)(&address)) [2] & 0xff),
	   (unsigned) (((char*)(&address)) [3] & 0xff));

  gdk_threads_enter ();
  gtk_label_set_text (GTK_LABEL (progressLabel), s);
  gtk_progress_bar_update (GTK_PROGRESS_BAR (progressBar), (float) (++numSeen) / numAddresses);
  gdk_threads_leave ();

  exception:;  
}

void mu_PacketHelper::run (void*data)
{
  int address = (int) data;
  m3_int64*good;
  m3_int64*bad;
  int numCounters;
  
  insist (this && address);
  insist (mu_mainWidget && mu_mainWidget->packetWidget);
  
  if (client.open (address) &&
      (numCounters = client.getPacketCounters (&good, &bad)) &&
      swap (numCounters, good, this->good) &&
      swap (numCounters, bad, this->bad))
    mu_mainWidget->packetWidget->eatCounters (address, numCounters, this->good, this->bad);
  
  exception:;
}


int mu_PacketHelper::swap (int count, m3_int64*from,  mu_int64*to)
{
  insist (this && from && to && count >= 0 && count <= M3_MAX_XBAR_PORT);
  
  for (int i = 0; i < count; i++)
  {
    unsigned high, low;
    high = htonl (from [i].high);
    low = htonl (from [i].low);
    
    to [i] = low;
#if 0
#if __BYTE_ORDER == __LITTLE_ENDIAN
    to [i] = ((unsigned long long) low) | ((unsigned long long) high << 32);
#else
    to [i] = ((unsigned long long) high) | ((unsigned long long) low << 32);
#endif
#endif
  }
  return 1;
  exception: return 0;
}
