#include "pi_Ping.h"
#include "mt_Host.h"
#include "mu_PingWidget.h"
#include "mu_MainWidget.h"
#include "mu_ExploreWidget.h"
#include "insist.h"

extern mu_MainWidget*mu_mainWidget;

#define MU_PING_SCALE_WIDTH 200

static void set_adjustment (GtkAdjustment*a, int lower, int upper, int value, int step_increment, int page_increment, int page_size)
{
  a->lower = lower;
  a->upper = upper;
  a->value = value;
  a->step_increment = step_increment;
  a->page_increment = page_increment;
  a->page_size = page_size;
}

static int mu_PingWidget_progress (mu_PingWidget*widget, int node, int inPort, int outPort, int size, int count)
{
  return widget && widget->progress (node, inPort, outPort, size, count);
}

static void mu_PingWidget_start (GtkWidget*w, mu_PingWidget*widget)
{
  widget->createThread ((mu_ThreadableFunction)&mu_PingWidget::run, (void*)0);
}

static void mu_PingWidget_stop (GtkWidget*w, mu_PingWidget*widget)
{
  widget->stopThread (0);
}

static void mu_PingWidget_close (GtkWidget*w, mu_PingWidget*widget)
{
  widget->hide ();
}

static gboolean mu_PingWidget_delete (GtkWidget*w, GdkEvent*e, mu_PingWidget*widget)
{
  widget->hide ();
  return 1;
}

static gboolean mu_PingWidget_destroy (GtkWidget*w, mu_PingWidget*widget)
{
  widget->hide ();
  return 0;
}


mu_PingWidget::mu_PingWidget ()
{
  GtkWidget*box;
  GtkWidget*hboxes [4];
  GtkWidget*vbox,*hbox;
  GtkWidget*frame, *label, *button;
  GtkAdjustment*adjustment;
  GSList*group;
  
  widget = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  insist (widget);  

  gtk_window_set_title (GTK_WINDOW (widget), MU_TITLE " - Ping");
  gtk_window_set_policy (GTK_WINDOW (widget), 1, 1, 0);

  box = gtk_vbox_new (0, MU_SPACING);
  insist (box);
  gtk_container_add (GTK_CONTAINER (widget), box);
  gtk_widget_show (box);

  for (int i = 0; i < 4; i++)
  {
    hboxes [i] = gtk_hbox_new (i == 3, MU_SPACING);
    insist (hboxes [i]);
    gtk_box_pack_start (GTK_BOX (box), hboxes [i], 1, 0, 0);
    gtk_widget_show (hboxes [i]);  
  }

  frame = gtk_frame_new ("Targets");
  insist (frame);
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (hboxes [0]), frame, 1, 0, 0);
  vbox = gtk_vbox_new (0, MU_SPACING);
  insist (vbox);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (frame), vbox);

  group = 0;
  testSwitchesRadio = gtk_radio_button_new_with_label (group, "Xbars");
  insist (testSwitchesRadio);
  gtk_widget_show (testSwitchesRadio);
  gtk_box_pack_start (GTK_BOX (vbox), testSwitchesRadio, 1, 0, 0);

  group = gtk_radio_button_group (GTK_RADIO_BUTTON (testSwitchesRadio));
  testHostsRadio = gtk_radio_button_new_with_label (group, "Hosts");
  insist (testHostsRadio);
  gtk_widget_show (testHostsRadio);
  gtk_box_pack_start (GTK_BOX (vbox), testHostsRadio, 1, 0, 0);

  frame = gtk_frame_new ("Sending");
  insist (frame);
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (hboxes [0]), frame, 1, 0, 0);
  vbox = gtk_vbox_new (0, MU_SPACING);
  insist (vbox);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (frame), vbox);

  hbox = gtk_hbox_new (0, MU_SPACING);
  insist (hbox);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, 1, 0, 0);
  label = gtk_label_new ("Count");
  insist (label);
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (hbox), label, 1, 0, 0);
  
  countScale = gtk_hscale_new (0);
  insist (countScale);
  gtk_widget_set (countScale, "width", MU_PING_SCALE_WIDTH, NULL);
  gtk_scale_set_digits (GTK_SCALE (countScale), 0);
  gtk_widget_show (countScale);
  gtk_box_pack_start (GTK_BOX (hbox), countScale, 1, 0, 0);
  adjustment = gtk_range_get_adjustment (GTK_RANGE (countScale));
  insist (adjustment);
  set_adjustment (adjustment, 1, 2000, 1, 1, 1, 0);

  hbox = gtk_hbox_new (0, MU_SPACING);
  insist (hbox);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, 1, 0, 0);
  label = gtk_label_new ("Timeout");
  insist (label);
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (hbox), label, 1, 0, 0);
  
  timeoutScale = gtk_hscale_new (0);
  insist (timeoutScale);
  gtk_widget_set (timeoutScale, "width", MU_PING_SCALE_WIDTH, NULL);
  gtk_scale_set_digits (GTK_SCALE (timeoutScale), 0);
  gtk_widget_show (timeoutScale);
  gtk_box_pack_start (GTK_BOX (hbox), timeoutScale, 1, 0, 0);
  adjustment = gtk_range_get_adjustment (GTK_RANGE (timeoutScale));
  insist (adjustment);
  set_adjustment (adjustment, 1, 1000*100, 1000, 1, 1, 0);
  
  foreverToggle = gtk_check_button_new_with_label ("Forever");
  gtk_widget_show (foreverToggle);
  gtk_box_pack_start (GTK_BOX (vbox), foreverToggle, 1, 0, 0);

  frame = gtk_frame_new ("Messages");
  insist (frame);
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (hboxes [1]), frame, 1, 0, 0);
  vbox = gtk_vbox_new (0, MU_SPACING);
  insist (vbox);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (frame), vbox);

  group = 0;
  fixedRadio = gtk_radio_button_new_with_label (group, "Fixed Size");
  insist (fixedRadio);
  gtk_widget_show (fixedRadio);
  gtk_box_pack_start (GTK_BOX (vbox), fixedRadio, 1, 0, 0);

  sizeScale = gtk_hscale_new (0);
  insist (sizeScale);
  gtk_widget_set (sizeScale, "width", MU_PING_SCALE_WIDTH, NULL);
  gtk_scale_set_digits (GTK_SCALE (sizeScale), 0);
  gtk_widget_show (sizeScale);
  gtk_box_pack_start (GTK_BOX (vbox), sizeScale, 1, 0, 0);
  adjustment = gtk_range_get_adjustment (GTK_RANGE (sizeScale));
  insist (adjustment);
  set_adjustment (adjustment, 16, 2000, 16, 1, 1, 0);

  group = gtk_radio_button_group (GTK_RADIO_BUTTON (fixedRadio));
  rangeRadio = gtk_radio_button_new_with_label (group, "Variable Size");
  insist (rangeRadio);
  gtk_widget_show (rangeRadio);
  gtk_box_pack_start (GTK_BOX (vbox), rangeRadio, 1, 0, 0);

  hbox = gtk_hbox_new (0, MU_SPACING);
  insist (hbox);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, 1, 0, 0);
  label = gtk_label_new ("Min");
  insist (label);
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (hbox), label, 1, 0, 0);
  
  minScale = gtk_hscale_new (0);
  insist (minScale);
  gtk_widget_set (minScale, "width", MU_PING_SCALE_WIDTH, NULL);
  gtk_scale_set_digits (GTK_SCALE (minScale), 0);
  gtk_widget_show (minScale);
  gtk_box_pack_start (GTK_BOX (hbox), minScale, 1, 0, 0);
  adjustment = gtk_range_get_adjustment (GTK_RANGE (minScale));
  insist (adjustment);
  set_adjustment (adjustment, 16, 2000, 16, 1, 1, 0);

  hbox = gtk_hbox_new (0, MU_SPACING);
  insist (hbox);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, 1, 0, 0);
  label = gtk_label_new ("Max");
  insist (label);
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (hbox), label, 1, 0, 0);
  
  maxScale = gtk_hscale_new (0);
  insist (maxScale);
  gtk_widget_set (maxScale, "width", MU_PING_SCALE_WIDTH, NULL);
  gtk_scale_set_digits (GTK_SCALE (maxScale), 0);
  gtk_widget_show (maxScale);
  gtk_box_pack_start (GTK_BOX (hbox), maxScale, 1, 0, 0);
  adjustment = gtk_range_get_adjustment (GTK_RANGE (maxScale));
  insist (adjustment);
  set_adjustment (adjustment, 16, 2000, 16, 1, 1, 0);

  hbox = gtk_hbox_new (0, MU_SPACING);
  insist (hbox);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, 1, 0, 0);
  label = gtk_label_new ("Step");
  insist (label);
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (hbox), label, 1, 0, 0);
  
  stepScale = gtk_hscale_new (0);
  insist (stepScale);
  gtk_widget_set (stepScale, "width", MU_PING_SCALE_WIDTH, NULL);
  gtk_scale_set_digits (GTK_SCALE (stepScale), 0);
  gtk_widget_show (stepScale);
  gtk_box_pack_start (GTK_BOX (hbox), stepScale, 1, 0, 0);
  adjustment = gtk_range_get_adjustment (GTK_RANGE (stepScale));
  insist (adjustment);
  set_adjustment (adjustment, 16, 2000, 16, 1, 1, 0);

  frame = gtk_frame_new ("Contents");
  insist (frame);
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (hboxes [1]), frame, 1, 0, 0);
  vbox = gtk_vbox_new (0, MU_SPACING);
  insist (vbox);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (frame), vbox);

  group = 0;
  noneRadio = gtk_radio_button_new_with_label (group, "None");
  insist (noneRadio);
  gtk_widget_show (noneRadio);
  gtk_box_pack_start (GTK_BOX (vbox), noneRadio, 0, 0, 0);  

  group = gtk_radio_button_group (GTK_RADIO_BUTTON (noneRadio));  
  sequentialRadio = gtk_radio_button_new_with_label (group, "Sequential");
  insist (sequentialRadio);
  gtk_widget_show (sequentialRadio);
  gtk_box_pack_start (GTK_BOX (vbox), sequentialRadio, 0, 0, 0);

  group = gtk_radio_button_group (GTK_RADIO_BUTTON (sequentialRadio));  
  randomRadio = gtk_radio_button_new_with_label (group, "Random");
  insist (randomRadio);
  gtk_widget_show (randomRadio);
  gtk_box_pack_start (GTK_BOX (vbox), randomRadio, 0, 0, 0);

  vbox = gtk_vbox_new (0, MU_SPACING);
  insist (vbox);
  gtk_widget_show (vbox);
  gtk_box_pack_start (GTK_BOX (hboxes [2]), vbox, 1, 1, 0);

  progressLabel = gtk_label_new ("   ");
  insist (progressLabel);
  gtk_widget_show (progressLabel);
  gtk_box_pack_start (GTK_BOX (vbox), progressLabel, 1, 1, 0);

  progressBar = gtk_progress_bar_new ();
  insist (progressBar);
  gtk_widget_show (progressBar);
  gtk_box_pack_start (GTK_BOX (vbox), progressBar, 1, 1, 0);

  startButton = gtk_button_new_with_label ("Start");
  gtk_signal_connect (GTK_OBJECT (startButton), "clicked", GTK_SIGNAL_FUNC (mu_PingWidget_start), this);
  gtk_widget_show (startButton);
  gtk_box_pack_start (GTK_BOX (hboxes [3]), startButton, 0, 0, MU_SPACING);

  stopButton = gtk_button_new_with_label ("Stop");
  gtk_signal_connect (GTK_OBJECT (stopButton), "clicked", GTK_SIGNAL_FUNC (mu_PingWidget_stop), this);
  gtk_widget_show (stopButton);
  gtk_box_pack_start (GTK_BOX (hboxes [3]), stopButton, 0, 0, MU_SPACING);
  gtk_widget_set_sensitive (stopButton, 0);

  button = gtk_button_new_with_label ("Close");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_PingWidget_close), this);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hboxes [3]), button, 0, 0, MU_SPACING);

  exception:;
}

void mu_PingWidget::run (void*data)
{
  int r;
  pi_Options options;
  mt_Calculator*calculator = mt_getCalculator ();
  mt_Args calculatorArgs (mt_Args::CALCULATOR, 100);
  mt_Args pingArgs (mt_Args::JOB, 100);
  
  ping = 0;
  
  insist (this && mu_mainWidget && mu_mainWidget->exploreWidget && mu_mainWidget->exploreWidget->network);

  pingArgs.addArg ("-unit");
  pingArgs.addArg ("0");
  pingArgs.addArg ("-careful");

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (testHostsRadio)))
  {
    pingArgs.addArg ("-test-hosts");
    testingXbars = 0;
    numNodes = mu_mainWidget->exploreWidget->graph.getNumHosts ();
  }
  else
  {
    pingArgs.addArg ("-test-xbars");
    testingXbars = 1;
    numNodes = mu_mainWidget->exploreWidget->graph.getNumSwitches ();
  }
  pingArgs.addArg ("-count");
  pingArgs.addArg ((int) gtk_range_get_adjustment (GTK_RANGE (countScale))->value);
  
  pingArgs.addArg ("-timeout");
  pingArgs.addArg ((int) gtk_range_get_adjustment (GTK_RANGE (timeoutScale))->value);

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (foreverToggle)))
    pingArgs.addArg ("-forever");
  
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (fixedRadio)))
  {    
    pingArgs.addArg ("-size");
    pingArgs.addArg ((int) gtk_range_get_adjustment (GTK_RANGE (sizeScale))->value);
  }
  else
  {
    pingArgs.addArg ("-min-size");
    pingArgs.addArg ((int) gtk_range_get_adjustment (GTK_RANGE (minScale))->value);    
    pingArgs.addArg ("-max-size");
    pingArgs.addArg ((int) gtk_range_get_adjustment (GTK_RANGE (maxScale))->value);
    pingArgs.addArg ("-increment-size");
    pingArgs.addArg ((int) gtk_range_get_adjustment (GTK_RANGE (stepScale))->value);
  }
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (sequentialRadio)))
    pingArgs.addArg ("-fill-fast");
  else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (randomRadio)))
    pingArgs.addArg ("-fill-slow");

  pingArgs.addArg ("-print-xbars");
  
  if (!options.parseArgs (&pingArgs))
    return;

  gdk_threads_enter ();
  gtk_widget_set_sensitive (startButton, 0);
  gtk_widget_set_sensitive (stopButton, 1);
  gdk_threads_leave ();

  if (!calculator->parseArgs (&calculatorArgs))
    return;
  calculator->clone (&mu_mainWidget->exploreWidget->graph);
  
  ping = new pi_Ping (mu_mainWidget->exploreWidget->graph.getRoot (), mu_mainWidget->exploreWidget->network, calculator, &options);
  insist (ping);
  ping->setProgressFunction ((pi_ProgressFunction) mu_PingWidget_progress, this);

  gdk_threads_enter ();
  gtk_progress_bar_set_bar_style (GTK_PROGRESS_BAR (progressBar), GTK_PROGRESS_DISCRETE);
  if (numNodes > 1)
    gtk_progress_bar_set_discrete_blocks (GTK_PROGRESS_BAR (progressBar), numNodes);
  gtk_progress_bar_update (GTK_PROGRESS_BAR (progressBar), 0);
  gdk_threads_leave ();
  
  ping->start ();
  mu_mainWidget->exploreWidget->network->close (ping);
  refresh ();
  
  
  gdk_threads_enter ();
  gtk_widget_set_sensitive (startButton, 1);
  gtk_widget_set_sensitive (stopButton, 0);
  gdk_threads_leave ();

  delete ping;
  delete calculator;
  
  exception:;
}


int mu_PingWidget::refresh ()
{
  insist (this && ping);
  insist (mu_mainWidget->cluster);

  char s [100];

  if (ping->numBadPhases + ping->numReceived == ping->numSent)
    sprintf (s, "received all %d messages", ping->numSent);
  else
    sprintf (s, "received %d/%d (%d missing)", ping->numReceived + ping->numBadPhases, ping->numSent, ping->numSent - ping->numReceived - ping->numBadPhases);

  insist (mu_mainWidget->cluster);
  gdk_threads_enter ();
  gtk_label_set_text (GTK_LABEL (progressLabel), s);
  mu_mainWidget->cluster->eatCounters (testingXbars, ping->numCounters, ping->counters);       
  gdk_threads_leave ();

  return 1;
  exception: return 0;
}


int mu_PingWidget::progress (int node, int inPort, int outPort, int size, int count)
{
  static int last = 0;
  char s [100];
  
  if (node < 0)
    return refresh () && !isStopped ();
  
  if (node != last)
  {
    insist (mu_mainWidget->exploreWidget);
    insist (mu_mainWidget->exploreWidget->xbarTable);
    insist (mu_mainWidget->cluster);

    mt_Node*n = testingXbars ? mu_mainWidget->exploreWidget->graph.getSwitch (node) : mu_mainWidget->exploreWidget->graph.getHost (node);
    insist (n);
  
    if (testingXbars)
    {
      m3_XbarTableEntry*e = mu_mainWidget->exploreWidget->xbarTable->lookup (n);
      insist (e);
      mu_Switch*sw;
      sw = mu_mainWidget->cluster->getSwitch (e->ethernetAddress);
      insist (sw);
      
      sprintf (s, "%d.%d.%d.%d. xbar %d", 
	       (unsigned) (((char*)(&sw->ipAddress)) [0] & 0xff),
	       (unsigned) (((char*)(&sw->ipAddress)) [1] & 0xff),
	       (unsigned) (((char*)(&sw->ipAddress)) [2] & 0xff),
	       (unsigned) (((char*)(&sw->ipAddress)) [3] & 0xff),
	       e->xbar);
    }
    else
      sprintf (s, "%s", n->getName ());
    
    gdk_threads_enter ();
    gtk_label_set_text (GTK_LABEL (progressLabel), s);
    gtk_progress_bar_update (GTK_PROGRESS_BAR (progressBar), ((float) node + 1) / (float) numNodes);
    gdk_threads_leave ();
    last = node;
  }
  
  return !isStopped ();
  exception: return 0;
}
