#include "insist.h"
#include "mt_Component.h"
#include "mu_RoutesWidget.h"
#include "mu_ExploreWidget.h"
#include "mu_MainWidget.h"

extern char*mu_workingDirectory;
extern mu_MainWidget*mu_mainWidget;

static gint mu_RoutesWidget_toggled (GtkWidget*w, mu_RoutesWidget*routesWidget)
{
  insist (routesWidget);
  return routesWidget->toggled ();
  exception: return 0;
}

static void mu_RoutesWidget_routesFileButton_clicked (GtkWidget*w, mu_RoutesWidget*routesWidget)
{
  routesWidget->showRoutesFileSelection ();
}

static void mu_RoutesWidget_routesFileSelectionOKButton_clicked (GtkWidget*w, mu_RoutesWidget*routesWidget)
{
  routesWidget->hideRoutesFileSelection ();
  routesWidget->setRoutesFile ();
}

static void mu_RoutesWidget_routesFileSelectionCancelButton_clicked (GtkWidget*w, mu_RoutesWidget*routesWidget)
{
  routesWidget->hideRoutesFileSelection ();
}

static gboolean mu_RoutesWidget_routesFileSelection_delete (GtkWidget*w, GdkEvent*e, mu_RoutesWidget*routesWidget)
{
  routesWidget->hideRoutesFileSelection ();
  return 1;
}

static gboolean mu_RoutesWidget_routesFileSelection_destroy (GtkWidget*w, mu_RoutesWidget*routesWidget)
{
  routesWidget->hideRoutesFileSelection ();
  return 1;
}



static gboolean mu_RoutesWidget_delete (GtkWidget*w, GdkEvent*e, mu_RoutesWidget*widget)
{
  widget->hide ();
  return 1;
}

static gboolean mu_RoutesWidget_destroy (GtkWidget*w, mu_RoutesWidget*widget)
{
  widget->hide ();
  return 0;
}

static void mu_RoutesWidget_close (GtkWidget*w, mu_RoutesWidget*widget)
{
  widget->hide ();
}

static void mu_RoutesWidget_run (GtkWidget*w, mu_RoutesWidget*widget)
{
  widget->createThread ((mu_ThreadableFunction)&mu_RoutesWidget::run, (void*)0);
}

mu_RoutesWidget::mu_RoutesWidget ()
{
  GtkWidget*windowBox, *hbox, *button, *label;
  widget = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  insist (widget);

  routesFileSelection = gtk_file_selection_new ("Routes File");
  gtk_signal_connect (GTK_OBJECT (routesFileSelection), "delete_event", GTK_SIGNAL_FUNC (mu_RoutesWidget_routesFileSelection_delete), this);
  gtk_signal_connect (GTK_OBJECT (routesFileSelection), "destroy", GTK_SIGNAL_FUNC (mu_RoutesWidget_routesFileSelection_destroy), this);
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (routesFileSelection)->ok_button), "clicked",
		      GTK_SIGNAL_FUNC (mu_RoutesWidget_routesFileSelectionOKButton_clicked), this);
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (routesFileSelection)->cancel_button), "clicked",
		      GTK_SIGNAL_FUNC (mu_RoutesWidget_routesFileSelectionCancelButton_clicked), this);

  
  gtk_window_set_title (GTK_WINDOW (widget), MU_TITLE " - Routes");
  gtk_window_set_policy (GTK_WINDOW (widget), 1, 1, 0);

  windowBox = gtk_vbox_new (0, MU_SPACING);
  insist (windowBox);
  gtk_widget_show (windowBox);
  
  gtk_container_add (GTK_CONTAINER (widget), windowBox);

  gtk_signal_connect (GTK_OBJECT (widget), "delete_event", GTK_SIGNAL_FUNC (mu_RoutesWidget_delete), this);
  gtk_signal_connect (GTK_OBJECT (widget), "destroy", GTK_SIGNAL_FUNC (mu_RoutesWidget_destroy), this);

  hbox = gtk_hbox_new (0, MU_SPACING);
  insist (hbox);
  gtk_box_pack_start (GTK_BOX (windowBox), hbox, 0, 0, 0);
  gtk_widget_show (hbox);

  label = gtk_label_new ("Routes File");
  insist (label);
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (hbox), label, 0, 0, MU_SPACING);

  routesFileEntry = gtk_entry_new ();
  gtk_widget_show (routesFileEntry);
  gtk_box_pack_start (GTK_BOX (hbox), routesFileEntry, 1, 1, MU_SPACING);

  button = gtk_button_new_with_label (" ... ");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_RoutesWidget_routesFileButton_clicked), this);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, 0, 0, MU_SPACING);

  hbox = gtk_hbox_new (0, MU_SPACING);
  insist (hbox);
  gtk_box_pack_start (GTK_BOX (windowBox), hbox, 0, 0, 0);
  gtk_widget_show (hbox);

  button = gtk_button_new_with_label ("Read");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_RoutesWidget_run), this);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, 0, 0, MU_SPACING);

  colorToggle = gtk_check_button_new_with_label ("Show Colors");
  gtk_signal_connect (GTK_OBJECT (colorToggle), "toggled", GTK_SIGNAL_FUNC (mu_RoutesWidget_toggled), this);
  gtk_widget_show (colorToggle);
  gtk_box_pack_start (GTK_BOX (hbox), colorToggle, 0, 0, MU_SPACING);

  button = gtk_button_new_with_label ("Close");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (mu_RoutesWidget_close), this);
  gtk_widget_show (button);
  gtk_box_pack_start (GTK_BOX (hbox), button, 0, 0, MU_SPACING);

  char s [300];

  mu_ExploreWidget::setFilename (s, mu_workingDirectory, "mute.routes");
  gtk_entry_set_text (GTK_ENTRY (routesFileEntry), s);
  gtk_file_selection_set_filename (GTK_FILE_SELECTION (routesFileSelection), s);

  routeTable = 0;
  
  exception:;
}

void mu_RoutesWidget::run (void*)
{
  insist (this);
  insist (mu_mainWidget && mu_mainWidget->exploreWidget);
  
  if (routeTable)
  {
    delete routeTable;
    routeTable = 0;
  }
  
  routeTable = new mt_RouteTable;
  insist (routeTable);
  
  routeTable->fromFile (gtk_entry_get_text (GTK_ENTRY (routesFileEntry)), &mu_mainWidget->exploreWidget->graph);
  exception:;
}

void mu_RoutesWidget::setRoutesFile ()
{
  insist (this && routesFileSelection && routesFileEntry);
  gtk_entry_set_text (GTK_ENTRY (routesFileEntry), gtk_file_selection_get_filename (GTK_FILE_SELECTION (routesFileSelection)));
  exception: return;
}

mt_Route*mu_RoutesWidget::getRoute (int from, int to)
{
  insist (this);
  return routeTable && routeTable->getRoute (from, to, 0, &route) ? &route : 0;
  exception: return 0;  
}

int mu_RoutesWidget::toggled ()
{
  insist (this);
  insist (mu_mainWidget && mu_mainWidget->cluster);

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (colorToggle)))
    return mu_mainWidget->cluster->drawRoutes ();
  else
    return mu_mainWidget->cluster->clearLinkColors ();

  exception: return 0;
}
