/*
  mu_Threadable.h
  thread hook.
  finucane@myri.com (David Finucane)
*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <sys/time.h>

#include "mu_Threadable.h"
#include "mu_Widget.h"

#include "insist.h"


/* extern "C" void*mu_ThreadableCreateHook (void*p); */

static GList*threadables = 0;

static void mu_ThreadablePrintline (mu_Threadable*threadable, char*s)
{
  threadable->printLine (s);
}

void mu_printLine (char*s)
{
  g_list_foreach (threadables, (GFunc) mu_ThreadablePrintline, s);
}

static void*mu_ThreadableCreateHook (void*p)
{
  mu_Threadable*threadable = (mu_Threadable*)p;
  insist (threadable->function);
  
  threadable->startThread ();
  
  exception: return 0;
}

mu_Threadable::mu_Threadable ()
{
  state = CREATED;
  thread = 0;
  threadText = 0;
  pthread_mutex_init (&mutex, 0);
  threadables = g_list_append (threadables, this);
}

mu_Threadable::~mu_Threadable ()
{
  threadables = g_list_remove (threadables, this);
  pthread_mutex_destroy (&mutex);
}

int mu_Threadable::isStopped ()
{
  insist (this);
  return state == STOPPED;
  exception: return 0;
}

int mu_Threadable::createThread (mu_ThreadableFunction function, void*data)
{
  insist (this);
  insist (function);
  
  if (state == RUNNING)
    return 1;

  state = RUNNING;
  this->function = function;
  this->data = data;

  pthread_mutex_lock (&mutex);
  pthread_create (&thread, 0, mu_ThreadableCreateHook, (void*) this);
  pthread_mutex_unlock (&mutex);

  return 1;
  exception: return 0;
}

void mu_Threadable::startThread ()
{
  GdkWindow*w = 0;
  GtkWidget*p;
  insist (this);

  if (threadText)
  {
    for (p = threadText; p->parent; p = p->parent);
    
    w = p->window;
    insist (w);
    
    GdkCursor*c = gdk_cursor_new (GDK_WATCH);
    insist (c);

    mu_threads_enter ();
    
    gdk_window_set_cursor (w, c);
    gdk_cursor_destroy (c);
    gtk_text_set_point (GTK_TEXT (threadText), 0);
    gtk_text_forward_delete (GTK_TEXT (threadText), gtk_text_get_length (GTK_TEXT (threadText)));

    mu_threads_leave ();
  }
  
  if (function)
    (this->*function) (data);
  state = STOPPED;

  if (w)
  {
    mu_threads_enter ();
    gdk_window_set_cursor (w, 0);
    mu_threads_leave ();
  }
  exception:;
}

int mu_Threadable::stopThread (int wait)
{
  state = STOPPED;
  if (wait)
    join ();
  
  return 1;
}
  

void mu_Threadable::printLine (char*s)
{
  insist (this && s);

  if (pthread_equal (thread, pthread_self ()))
  {    
    if (threadText)
    {
      gdk_threads_enter ();
      gtk_text_insert (GTK_TEXT (threadText), 0, 0, 0, s, -1);
      gtk_text_insert (GTK_TEXT (threadText), 0, 0, 0, "\n", -1);
      gdk_threads_leave ();
    }
    else
      printf ("%s\n", s);
  }
  
  exception:;
}

void mu_Threadable::sleep (int seconds)
{
  fd_set mask;
  struct timeval timeout;
  timeout.tv_sec = seconds;
  timeout.tv_usec = 0;
  
  FD_ZERO (&mask);
  
  select (0, &mask, 0, 0, &timeout);
}

int mu_Threadable::join ()
{
  void*r = 0;
  
  insist (this);
  if (thread)
    pthread_join (thread, &r);
  thread = 0;

  return (int) r;
  exception: return 0;
}

