#include "insist.h"
#include "mu_MainWidget.h"
#include "mu_ExploreWidget.h"
#include "mt_Component.h"

static char*mu_dot = ".";

char*mu_workingDirectory = "";
char*mu_firmwareDirectory = "";

mu_MainWidget*mu_mainWidget = 0;
int mu_eventMask [2] = {-1, -1};
int mu_eventHistory = MU_DEFAULT_HISTORY;
int mu_maxThreads = MU_DEFAULT_THREADS;
int mu_logEvents = 0;
pthread_t mu_mainThread = 0;

extern void mu_printLine (char*s);

static struct poptOption mute_options [] =
{
  {
    "working-directory",
    'w',
    POPT_ARG_STRING,
    &mu_workingDirectory,
    0,
    "path to working directory",
    "WORKING_DIRECTORY"
  },
  {
    "firmware-directory",
    'f',
    POPT_ARG_STRING,
    &mu_firmwareDirectory,
    0,
    "path to firmware directory",
    "FIRMWARE_DIRECTORY"
  },
  {
    "max-threads",
    't',
    POPT_ARG_INT,
    &mu_maxThreads,
    0,
    "max thread count",
    "MAX_THREADS"
  },
  {
    "event-history",
    'h',
    POPT_ARG_INT,
    &mu_eventHistory,
    0,
    "event history length",
    "EVENT_HISTORY"
  },
  {
    "log-events",
    'l',
    POPT_ARG_NONE,
    &mu_logEvents,
    0,
    "log events",
    "LOG_EVENTS"
  },

  {
    0,0,0,0,0,0
  }
};

int main (int argc, char*argv [])
{
  poptContext context;
  int r;  
  mt_Component::initialize (mu_printLine);
  m3_Component::initialize (mu_printLine);
  mu_workingDirectory = 0;
  mu_firmwareDirectory = 0;
  
  g_thread_init (0);
  
  bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
  textdomain (PACKAGE);

  gnome_init_with_popt_table (PACKAGE, VERSION, argc, argv, mute_options, 0, &context);
  poptFreeContext (context);

  if (!mu_firmwareDirectory)
    mu_firmwareDirectory = mu_dot;
  if (!mu_workingDirectory)
    mu_workingDirectory = mu_dot;
  if (mu_maxThreads < MU_MIN_THREADS || mu_maxThreads > MU_MAX_THREADS)
    mu_maxThreads = MU_DEFAULT_THREADS;
  
  mu_mainWidget = new mu_MainWidget (PACKAGE);
  insist (mu_mainWidget);
  
  mu_mainWidget->show ();
  mu_mainWidget->exploreWidget->show ();

  mu_mainThread = pthread_self ();
  
  gdk_threads_enter ();
  gtk_main ();
  gdk_threads_leave ();
  
  exception: return 0;
}

void mu_threads_enter ()
{
  if (pthread_self () != mu_mainThread)
    gdk_threads_enter ();
}


void mu_threads_leave ()
{
  if (pthread_self () != mu_mainThread)
    gdk_threads_leave ();
}


void mu_refresh ()
{
  if (pthread_self () == mu_mainThread)
  {
    while (gtk_events_pending ())
      gtk_main_iteration ();
  }
}
